/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2001,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: IpcJmsAdminConnection.java,v 1.17 2003/08/16 09:45:46 tanderson Exp $
 *
 * Date         Author  Changes
 * $Date        jimm    Created
 */
package org.exolab.jms.administration.mipc;

import java.io.IOException;
import java.util.Vector;

import javax.jms.JMSException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.ipc.IpcIfc;
import org.exolab.core.mipc.MultiplexConnection;
import org.exolab.core.mipc.MultiplexConnectionIfc;
import org.exolab.core.mipc.ObjectChannel;
import org.exolab.jms.administration.AdminConnection;
import org.exolab.jms.administration.JmsAdminServerIfc;


/**
 * This class is repsonsible for opening a TCP connection to the server
 * and passing and packing all requests for delivery.
 *
 * @version     $Revision: 1.17 $ $Date: 2003/08/16 09:45:46 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 */
public class IpcJmsAdminConnection implements JmsAdminServerIfc,
    AdminConnection {

    /**
     * This is a reference to the server connection.
     */
    private MultiplexConnectionIfc _mc = null;

    /**
     * The server ipc channel
     */
    private IpcIfc _connection = null;

    /**
     * The logger
     */
    private static final Log _log =
        LogFactory.getLog(IpcJmsAdminConnection.class);


    /**
     * The server host address
     */
    private String _host;

    /**
     * The port number the server is listening to
     */
    private int _port;

    /**
     * The connection identifier, allocated by the server
     */
    private String _connectionId = null;


    /**
     * Construct a new <code>IpcJmsAdminConnection</code>,
     * without establishing a connection
     *
     * @param host the server host
     * @param port the server port
     */
    protected IpcJmsAdminConnection(String host, int port) {
        _host = host;
        _port = port;
    }

    /**
     * Construct a new <code>IpcJmsAdminConnection</code>,
     * establishing a connection to the server
     *
     * @param host the server host
     * @param port the server port
     * @param username the admin user name
     * @param password the admin password
     * @throws JMSEXception if a connection cannot be established
     */
    public IpcJmsAdminConnection(String host, int port, String username,
                                 String password) throws JMSException {
        this(host, port);
        connect(username, password);
    }

    // implementation of JmsAdminServerIfc.addDurableConsumer
    public boolean addDurableConsumer(String topic, String name)
        throws JMSException {
        Vector v = pack("addDurableConsumer", 2);
        v.add(topic);
        v.add(name);
        send(v);
        return checkReply("addDurableConsumer");
    }

    // implementation of JmsAdminServerIfc.removeDurableConsumer
    public boolean removeDurableConsumer(String name) throws JMSException {
        Vector v = pack("removeDurableConsumer", 1);
        v.add(name);
        send(v);
        return checkReply("removeDurableConsumer");
    }

    // implementation of JmsAdminServerIfc.getDurableConsumers
    public Vector getDurableConsumers(String topic) throws JMSException {
        Vector v = pack("getDurableConsumers", 1);
        v.add(topic);
        send(v);
        return (Vector) getReply("getDurableConsumers");
    }

    // implementation of JmsAdminServerIfc.durableConsumerExists
    public boolean durableConsumerExists(String name) throws JMSException {
        Vector v = pack("durableConsumerExists", 1);
        v.add(name);
        send(v);
        return checkReply("durableConsumerExists");
    }

    /**
     * Check to see if the given consumer is currently connected
     *
     * @param       name the name of the onsumer.
     * @return      <code>true</code> if the consumer is connected
     * @throws      JMSException
     */
    public boolean isConnected(String name) throws JMSException {
        Vector v = pack("isConnected", 1);
        v.add(name);
        send(v);
        return checkReply("isConnected");
    }

    // implementation of JmsAdminServerIfc.removeDurableConsumer
    public boolean unregisterConsumer(String name) throws JMSException {
        Vector v = pack("unregisterConsumer", 1);
        v.add(name);
        send(v);
        return checkReply("unregisterConsumer");
    }

    // implementation of JmsAdminServerIfc.addDestination
    public boolean addDestination(String destination, Boolean queue)
        throws JMSException {
        Vector v = pack("addDestination", 2);
        v.add(destination);
        v.add(queue);
        send(v);
        return checkReply("addDestination");
    }

    // implementation of JmsAdminServerIfc.removeDestination
    public boolean removeDestination(String name) throws JMSException {
        Vector v = pack("removeDestination", 1);
        v.add(name);
        send(v);
        return checkReply("removeDestination");
    }

    // implementation of JmsAdminServerIfc.destinationExists
    public boolean destinationExists(String name) throws JMSException {
        Vector v = pack("destinationExists", 1);
        v.add(name);
        send(v);
        return checkReply("destinationExists");
    }

    // implementation of JmsAdminServerIfc.getAllDestinations
    public Vector getAllDestinations() throws JMSException {
        Vector v = pack("getAllDestinations", 0);
        send(v);
        return (Vector) getReply("getAllDestinations");
    }

    // implementation of JmsAdminServerIfc.getDurableConsumerMessageCount
    public int getDurableConsumerMessageCount(String topic, String name)
        throws JMSException {
        Vector v = pack("getDurableConsumerMessageCount", 2);
        v.add(topic);
        v.add(name);
        send(v);
        Integer result = (Integer) getReply("getDurableConsumerMessageCount");
        return result.intValue();
    }

    // implementation of JmsAdminServerIfc.getDurableConsumerMessageCount
    public int getQueueMessageCount(String queue) throws JMSException {
        Vector v = pack("getQueueMessageCount", 1);
        v.add(queue);
        send(v);
        return ((Integer) getReply("getQueueMessageCount")).intValue();
    }

    // implementation of JmsAdminServerIfc.purgeMessages
    public int purgeMessages() throws JMSException {
        Vector v = pack("purgeMessages", 0);
        send(v);
        return ((Integer) getReply("purgeMessages")).intValue();
    }

    // implementation of JmsAdminServerIfc.stopServer
    public void stopServer() throws JMSException {
        Vector v = pack("stopServer", 0);
        send(v);
        checkReply("stopServer");
    }

    // implementation of JmsAdminServerIfc.close
    public void close() {
        if (_connection != null) {
            try {
                _connection.close();
                _mc.finish();
            } catch (Exception exception) {
                _log.debug("Exception during close", exception);
            }
            _connection = null;
            _mc = null;
        }
    }

    // implementation of JmsAdminServerIfc.addUser
    public boolean addUser(String username, String password)
        throws JMSException {
        Vector v = pack("addUser", 2);
        v.add(username);
        v.add(password);
        send(v);
        return checkReply("addUser");
    }

    // implementation of JmsAdminServerIfc.getAllUsers
    public Vector getAllUsers() throws JMSException {
        Vector v = pack("getAllUsers", 0);
        send(v);
        return (Vector) getReply("getAllUsers");
    }

    // implementation of JmsAdminServerIfc.removeUser
    public boolean removeUser(String username) throws JMSException {
        Vector v = pack("removeUser", 1);
        v.add(username);
        send(v);
        return checkReply("removeUser");
    }

    // implementation of JmsAdminServerIfc.changePassword
    public boolean changePassword(String username, String password)
        throws JMSException {
        Vector v = pack("changePassword", 2);
        v.add(username);
        v.add(password);
        send(v);
        return checkReply("changePassword");
    }

    /**
     * Returns the server host
     *
     * @return the server host
     */
    protected String getHost() {
        return _host;
    }

    /**
     * Returns the server port
     *
     * @return the server port
     */
    protected int getPort() {
        return _port;
    }

    /**
     * Returns the underlying connection
     */
    protected IpcIfc getConnection() {
        return _connection;
    }

    /**
     * Sets the underlying connection
     *
     * @param connection the underlying connection
     */
    protected void setConnection(IpcIfc connection) {
        _connection = connection;
    }

    /**
     * Opens a connection to the server.
     *
     * @throws JMSException if an error occurs
     */
    protected void openConnection() throws JMSException {
        try {
            _mc = createClientConnection(_host, _port);
            _connection = new ObjectChannel("server", _mc);
            ((Thread) _mc).start();
        } catch (Exception exception) {
            raise(exception);
        }
    }

    /**
     * Opens an authenticated connection to the server
     *
     * @param username the user's name
     * @param password the user's password
     * @throws JMSException if the connection cannot be established
     */
    protected void connect(String username, String password)
        throws JMSException {
        openConnection();

        Vector v = pack("createConnection", 2);
        v.add(username);
        v.add(password);
        send(v);

        _connectionId = (String) getReply("createConnection");
    }

    /**
     * Create an MIPC client connection
     *
     * @param host the host of the server
     * @param port the port number to use
     * @return a new connection
     * @throws IOException if the server fails to initialise the ip service
     */
    protected MultiplexConnectionIfc createClientConnection(
        String host, int port) throws IOException {
        return new MultiplexConnection(host, port);
    }

    /**
     * Pack all the data that is required by the server in a vector.
     * Set the size of the vector to be exactly the right size for efficiency.
     *
     * @param method the function to activate on the server
     * @param num the number of additional items to pack
     * @return a vector containing all the data
     *
     */
    private Vector pack(String method, int num) {
        Vector v = new Vector(3 + num);
        v.add("org.exolab.jms.server.mipc.IpcJmsAdminConnection");
        v.add(method);
        v.add(_connectionId);
        return v;
    }

    /**
     * A convenience method to send a packed command to the server.
     * @param v the vector to be sendt
     * @throws JMSException for any error
     */
    private void send(Vector v) throws JMSException {
        try {
            _connection.send(v);
        } catch (Exception exception) {
            raise(exception);
        }
    }

    /**
     * A convenience method to check the success of operations which return
     * a true on sucess.
     *
     * @param method the requested server function.
     * @return true if the operation was successfull
     * @throws JMSException for any error
     */
    private boolean checkReply(String method) throws JMSException {
        boolean reply = false;

        try {
            Vector v = (Vector) _connection.receive();
            if (v != null) {
                Boolean b = (Boolean) v.get(0);
                if (!b.booleanValue()) {
                    if (v.get(1) instanceof JMSException) {
                        throw (JMSException) v.get(1);
                    } else {
                        throw new JMSException("Operation " + method
                            + " failed:\n" + v.get(1));
                    }
                }
            } else {
                throw new JMSException("Unknown connection error for "
                    + method);
            }
            reply = ((Boolean) v.get(1)).booleanValue();
        } catch (Exception exception) {
            raise(exception);
        }
        return reply;
    }

    /**
     * A convenience method to check the success of operations which return
     * a value
     *
     * @param method the requested server function
     * @return the data returned by the server
     * @throws JMSException On any failure.
     */
    private Object getReply(String method) throws JMSException {
        Object reply = null;

        try {
            Vector v = (Vector) _connection.receive();
            if (v != null) {
                Boolean b = (Boolean) v.get(0);
                if (!b.booleanValue()) {
                    if (v.get(1) instanceof JMSException) {
                        throw (JMSException) v.get(1);
                    } else {
                        throw new JMSException("Operation " + method
                            + " failed:\n" + v.get(1));
                    }
                } else {
                    reply = v.get(1);
                }
            } else {
                throw new JMSException("Unknown connection error for "
                    + method);
            }
        } catch (Exception exception) {
            raise(exception);
        }
        return reply;
    }

    private void raise(Exception exception) throws JMSException {
        if (exception instanceof JMSException) {
            throw (JMSException) exception;
        } else {
            JMSException error = new JMSException(exception.getMessage());
            error.setLinkedException(exception);
            throw error;
        }
    }

} //-- IpcJmsAdminConnection
