/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 */

package org.exolab.jms.authentication;


import org.exolab.core.service.BasicService;
import org.exolab.core.service.ServiceException;


/**
 * This is the active authentication component within the JMS server.
 *
 * @version     $Revision: 1.3 $ $Date: 2003/08/07 13:32:48 $
 * @author      <a href="mailto:knut@lerpold">Knut Lerpold</a>
 */
public class AuthenticationMgr extends BasicService {

    /**
     * The user manager
     */
    private UserManager _users;

    /**
     * The service name of the authentication manager
     */
    private static final String AM_SERVICE_NAME = "AuthenticationManager";

    /**
     * The singleton instance of the authentication manager.
     */
    private static volatile AuthenticationMgr _instance;


    /**
     * Construct a new <code>AuthenticationManager</code>
     *
     * @throws ServiceException if the service cannot be initialised
     */
    private AuthenticationMgr() throws ServiceException {
        super(AM_SERVICE_NAME);
        _users = new UserManager();
    }

    /**
     * Create and return an instance of the singleton.
     *
     * @return the singleton instance
     * @throws ServiceException if the service cannot be created
     */
    public static AuthenticationMgr createInstance() throws ServiceException {
        _instance = new AuthenticationMgr();
        return _instance;
    }

    /**
     * Return an instance to the AuthenticationMgr singleton. This method
     * assumes that the singleton has already been created with a call to
     * {@link #createInstance}
     *
     * @return the singleton instance
     */
    public static AuthenticationMgr instance() {
        return _instance;
    }

    // implement BasicService.run
    public void run() {
        // do nothing
    }

    // override BasicService.stop
    public void stop() throws ServiceException {
        // destroy the user manager.
        _users.destroy();

        // clear the static reference
        _instance = null;
    }

    /**
     * Create a user.
     *
     * @param user the user to create
     * @return <code>true</code> if the user is created
     * otherwise <code>false</code>
     */
    public boolean addUser(User user) {
        return _users.createUser(user);
    }

    /**
     * Remove this user
     *
     * @param user the user to remove
     * @return <code>true</code> if the user is removed
     * otherwise <code>false</code>
     */
    public boolean removeUser(User user) {
        return _users.deleteUser(user);
    }

    /**
     * Gets a user.
     *
     * @param user the user
     * @return User
     */
    public User getUser(User user) {
        return _users.getUser(user);
    }

    /**
     * Update a user.
     *
     * @param user the user to update
     * @return <code>true</code> if the password is updated
     * otherwise <code>false</code>
     */
    public boolean updateUser(User user) {
        return _users.updateUser(user);
    }

    /**
     * Validate the password for the specified user.
     *
     * @param username the user's name
     * @param password the password to check
     * @return <code>true</code> if the username and password exist, otherwise
     * <code>false</code>
     */
    public boolean validateUser(String username, String password) {
        return _users.validateUser(username, password);
    }

}
