/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: JmsSessionStubIfc.java,v 1.20 2003/08/17 01:32:21 tanderson Exp $
 *
 * Date         Author  Changes
 * 04/12/2000    jima    Created
 */
package org.exolab.jms.client;

import java.util.Vector;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;


/**
 * This is the interface that session stubs must implement in order to
 * support remote invocations. This level of indirection will enable us to
 * support different ORB environments. The only restriction is that the
 * stubs must support this interface and that it must have a default
 * constructor.
 *
 * @version     $Revision: 1.20 $ $Date: 2003/08/17 01:32:21 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 */
public interface JmsSessionStubIfc {

    /**
     * Return the client id associated with this session. 
     *
     * @return the client id associated with this session
     * @throws JMSException for any JMS error
     */
    String getClientId() throws JMSException;

    /**
     * Return the session identity. Session identities are unique
     * within the context of a server.
     *
     * @return the session identity
     * @throws JMSException for any JMS error
     */
    String getSessionId() throws JMSException;

    /**
     * This method is called before the call to <code>close</code>, so that the
     * stub can do some local clean up
     *
     * @throws JMSException for any JMS error
     */
    void beforeClose() throws JMSException;

    /**
     * Close and release any resource allocated to this session.
     *
     * @throws JMSException for any JMS error
     */
    void close() throws JMSException;

    /**
     * Acknowledge the following message If this method does not complete then
     * throw JMSException.
     *
     * @param clientId the identity ofthe client
     * @param messageId the message identity to ack
     * @throws JMSException for any JMS error
     */
    void acknowledgeMessage(long clientId, String messageId)
        throws JMSException;

    /**
     * Send the specified message to the server. If there is any problem
     * then throw the JMSException exception
     *
     * @param message the message to send
     * @throws JMSException for any JMS error
     */
    void sendMessage(Message message) throws JMSException;

    /**
     * Send the specified messages to the server. If there is any problem
     * then throw the JMSException exception
     *
     * @param messages the messages to send
     * @throws JMSException for any JMS error
     */
    void sendMessages(Vector messages) throws JMSException;

    /**
     * Return the next message for the specified client. The client id
     * maps to a consumer on the server side. The caller can also specify
     * how long to wait if no messages are currently available. If the caller
     * specifies 0 then the call will return immediately if there are no
     * messages available. If the caller specified -1 then the call will
     * block until a message becomes available.
     *
     * @param clientId the client identity
     * @param wait the number of ms to wait. -1 means wait indefinitely.
     * @return the next message or null
     * @throws JMSException for any JMS error
     */
    Message receiveMessage(long clientId, long wait) throws JMSException;

    /**
     * Return a collection of messages from the specified client upto the
     * nominated count. This method may return less than count messages
     * but it will never return more than count messages
     *
     * @param client the client identity
     * @param count max messages to return
     * @return collection of MessageImpl objects
     * @throws JMSException for any JMS error
     */
    Vector receiveMessages(long clientId, int count) throws JMSException;

    /**
     * Create a queue with the specified name. If the queue already exists
     * then simply return a reference to it. If the queue does not exist
     * then create it.
     *
     * @param queue the queue to create
     * @throws JMSException for any JMS error
     */
    void createQueue(JmsQueue queue) throws JMSException;

    /**
     * Create a topic with the specified name. If the topic already exists
     * then simply return a reference to it. If the topic does not exist
     * then create it.
     *
     * @param topic the topic to create
     * @throws JMSException for any JMS error
     */
    void createTopic(JmsTopic topic) throws JMSException;

    /**
     * Create a receiver endpoint for this session. A reciever is a message
     * consumer specific to the queue message model. The receiver is associated
     * with a queue.
     * <p>
     * You cannot create more than one receiver for the same destination
     *
     * @param queue the receiver destination
     * @param clientId the session allocated identifier of this consumer
     * @param selector the message selector. This may be null.
     * @return the unique consumer identifier
     * @throws JMSException for any JMS error
     */
    void createReceiver(JmsQueue queue, long clientId, String selector)
        throws JMSException;

    /**
     * Create a sender endpoint for this session. A sender is a message
     * publisher specific to the queue message model. The sender is associated
     * with a queue.
     * <p>
     * You cannot create more than one receiver for the same destination
     *
     * @param queue the receiver destination
     * @throws JMSException for any JMS error
     */
    void createSender(JmsQueue queue) throws JMSException;

    /**
     * Create a queue browser for this session. This allows clients to browse
     * a queue without removing any messages.
     * <p>
     *
     * You cannot create more than one queue browser for the same queue
     * in a single session.
     *
     * @param queue the queue to browse
     * @param clientId the client identity
     * @param selector the message selector. This may be null
     * @throws JMSException for any JMS error
     */
    void createBrowser(JmsQueue queue, long clientId, String selector)
        throws JMSException;

    /**
     * Delete the receiver with the corresponding client id.
     *
     * @param clientId the identity of the receiver to delete
     * @throws JMSException for any JMS error
     */
    void deleteReceiver(long clientId) throws JMSException;

    /**
     * Delete the queue browser associated with the corresponding client id.
     *
     * @param clientId the id of the browser
     * @throws JMSException for any JMS error
     */
    void deleteBrowser(long clientId) throws JMSException;

    /**
     * Create a subscriber endpoint for this session. A subscriber is a message
     * consumer specific to the topic message model. The subscriber is 
     * associated with a topic. The name is used to identify the consumer and 
     * can be set to null
     * <p>
     * You cannot create more than one subscriber for the same destination
     *
     * @param topic subscriber destination
     * @param name name of the consumer associated with the subscriber. 
     * This may be null.
     * @param clientId the session allocated identifier of this consumer
     * @param selector message selector. This may be null.
     * @param noLocal inhibit consuming messages on same connection.
     * @throws JMSException for any JMS error
     */
    void createSubscriber(JmsTopic topic, String name, long clientId,
                          String selector, boolean noLocal)
        throws JMSException;

    /**
     * Create a publisher endpoint for this session. A publisher is a message
     * publisher specific to the topic message model. The publisher is 
     * associated with a topic.
     * <p>
     * You cannot create more than one publisher for the same destination
     *
     * @param topic receiver destination
     * @throws JMSException for any JMS error
     */
    void createPublisher(JmsTopic topic) throws JMSException;

    /**
     * Delete the subscriber associated with the sepcified identity.
     *
     * @param clientid the client identity
     * @throws JMSException for any JMS error
     */
    void deleteSubscriber(long clientId) throws JMSException;

    /**
     * Unsubscribe a durable subscription
     *
     * @param name the name used to identify the subscription
     * @throws JMSException for any JMS error
     */
    void unsubscribe(String name) throws JMSException;

    /**
     * Stop message delivery to this session
     *
     * @throws JMSException for any JMS error
     */
    void stopMessageDelivery() throws JMSException;

    /**
     * Start message delivery to this session
     *
     * @throws JMSException for any JMS error
     */
    void startMessageDelivery() throws JMSException;

    /**
     * Set the listener for this session. The listener is an object that
     * implements MessageListener and is called back whenever a message
     * for the session is present
     *
     * @param listener the message listener
     */
    void setMessageListener(JmsMessageListener listener);

    /**
     * Enable or disable asynchronous message delivery for a particular
     * consumer
     *
     * @param clientId the id of the client to check
     * @param id  the message id of the last delivered message
     * @param enable true to enable; false to disable
     * @throws JMSException for any JMS error
     */
    void enableAsynchronousDelivery(long clientId, String id, boolean enable)
        throws JMSException;

    /**
     * Recover the session. This means all unacknowledged messages are
     * resent with the redelivery flag set
     *
     * @throws JMSException if the session cannot be recovered
     */
    void recover() throws JMSException;

    /**
     * Commit the session which will send all the published messages and
     * acknowledge all received messages
     *
     * @throws JMSException if the session cannot be committed
     */
    void commit() throws JMSException;

    /**
     * Rollback the session, which will not acknowledge any of the sent
     * messages
     *
     * @throws JMSException if the session cannot be rolled back
     */
    void rollback() throws JMSException;

    /**
     * Commits an XA transaction that is in progress.
     *
     * @param xid the xa transaction identity
     * @param onePhase true if it is a one phase commit
     * @throws XAException if there is a problem completing the call
     */
    void commit(Xid xid, boolean onePhase) throws XAException;

    /**
     * Ends the work performed on behalf of a transaction branch. The resource
     * manager disassociates the XA resource from the transaction branch
     * specified and let the transaction be completedCommits an XA transaction
     * that is in progress.
     *
     * @param xid the xa transaction identity
     * @param flags one of TMSUCCESS, TMFAIL, or TMSUSPEND
     * @throws XAException if there is a problem completing the call
     */
    void end(Xid xid, int flags) throws XAException;

    /**
     * Tell the resource manager to forget about a heuristically completed
     * transaction branch.
     *
     * @param xid the xa transaction identity
     * @throws XAException if there is a problem completing the call
     */
    void forget(Xid xid) throws XAException;

    /**
     * Return the transaction timeout for this instance of the resource
     * manager.
     *
     * @return the timeout in seconds
     * @throws XAException if there is a problem completing the call
     */
    int getTransactionTimeout() throws XAException;

    /**
     * Ask the resource manager to prepare for a transaction commit of the
     *transaction specified in xid
     *
     * @param xid the xa transaction identity
     * @return XA_RDONLY or XA_OK
     * @throws XAException if there is a problem completing the call
     */
    int prepare(Xid xid) throws XAException;

    /**
     * Obtain a list of prepared transaction branches from a resource manager.
     * The transaction manager calls this method during recovery to obtain the
     * list of transaction branches that are currently in prepared or
     * heuristically completed states.
     *
     * @param flag One of TMSTARTRSCAN, TMENDRSCAN, TMNOFLAGS. TMNOFLAGS
     * @return the set of Xids to recover
     * @throws XAException - if there is a problem completing the call
     */
    Xid[] recover(int flag) throws XAException;

    /**
     * Inform the resource manager to roll back work done on behalf of a
     * transaction branch
     *
     * @param xid the xa transaction identity
     * @throws XAException if there is a problem completing the call
     */
    void rollback(Xid xid) throws XAException;

    /**
     * Set the current transaction timeout value for this XAResource instance.
     *
     * @param seconds timeout in seconds
     * @return if the new transaction timeout was accepted
     * @throws XAException if there is a problem completing the call
     */
    boolean setTransactionTimeout(int seconds) throws XAException;

    /**
     * Start work on behalf of a transaction branch specified in xid If TMJOIN
     * is specified, the start is for joining a transaction previously seen by
     * the resource manager
     *
     * @param xid the xa transaction identity
     * @param flags One of TMNOFLAGS, TMJOIN, or TMRESUME
     * @throws XAException if there is a problem completing the call
     */
    void start(Xid xid, int flags) throws XAException;

    /**
     * Return the identity of the associated resource manager.
     *
     * @return the identity of the resource manager
     * @throws XAException if there is a problem completing the call
     */
    String getResourceManagerId() throws XAException;
}
