/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: JmsTemporaryTopic.java,v 1.7 2003/08/07 13:32:50 tanderson Exp $
 *
 * Date         Author  Changes
 * 3/21/2000    jima    Created
 */
package org.exolab.jms.client;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import javax.jms.JMSException;
import javax.jms.TemporaryTopic;
import javax.jms.Topic;
import javax.naming.Reference;
import javax.naming.Referenceable;

import org.exolab.jms.util.UUID;


/**
 * A temporary topic is created by a client through a session and has
 * a lifetime of the session's connection.
 *
 * @version     $Revision: 1.7 $ $Date: 2003/08/07 13:32:50 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 **/
public class JmsTemporaryTopic
    extends JmsTopic
    implements TemporaryTopic, JmsTemporaryDestination, Externalizable, Referenceable {

    /**
     * Used for serialization
     */
    static final long serialVersionUID = 1;

    /**
     * Contains the id to the connection that created this temporary
     * destination. Note that the actual connection object is not
     * streamed across, only the identifier.
     */
    private String _connectionId = null;

    /**
     * Reference to the connection that owns this deztination.
     */
    transient private JmsConnection _connection = null;


    /**
     * Need a default constructor for the serialization
     */
    public JmsTemporaryTopic() {
        super(TEMP_TOPIC_PREFIX + UUID.next());
    }

    // implementation of TemporaryTopic.delete
    public void delete()
        throws JMSException {
        // unregisterthe temporary topic from the owning destination
        _connection.deleteTemporaryDestination(this);
    }

    // implementation of JmsTemporaryDestination.setOwningConnection
    public void setOwningConnection(JmsConnection connection) {
        _connection = connection;
        _connectionId = _connection.getConnectionId();
    }

    // implementation of JmsTemporaryDestination.getOwningConnection
    public JmsConnection getOwningConnection() {
        return _connection;
    }

    // implementation of JmsTemporaryDestination.getConnectionId
    public String getConnectionId() {
        return _connectionId;
    }

    // implementation of JmsTemporaryDestination.validForConnection
    public boolean validForConnection(JmsConnection connection) {
        boolean result = false;

        if ((connection != null) &&
            (connection.getConnectionId().equals(_connectionId))) {
            result = true;
        }

        return result;
    }

    // implementation of Referenceable.getReference
    public Reference getReference() {
        // should never be called for temp destination
        return null;
    }

    // implementation of Externalizable.writeExternal
    public void writeExternal(ObjectOutput stream)
        throws IOException {
        stream.writeLong(serialVersionUID);
        stream.writeObject(_connectionId);
        super.writeExternal(stream);
    }

    // implementation of Externalizable.writeExternal
    public void readExternal(ObjectInput stream)
        throws IOException, ClassNotFoundException {
        long version = stream.readLong();
        if (version == serialVersionUID) {
            _connectionId = (String) stream.readObject();
            super.readExternal(stream);
        } else {
            throw new IOException("JmsTemporaryTopic with version " +
                version + " is not supported.");
        }
    }
}
