/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: HttpJmsSessionStub.java,v 1.16 2003/08/25 03:35:45 tanderson Exp $
 *
 * Date             Author      Changes
 * 12 Oct 2001      mourikis    Created
 */
package org.exolab.jms.client.http;

import java.io.Serializable;
import java.net.InetAddress;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Vector;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.http.HttpClient;
import org.exolab.core.ipc.IpcIfc;
import org.exolab.core.ipc.NotifierIfc;
import org.exolab.core.ipc.Server;
import org.exolab.jms.client.JmsMessageListener;
import org.exolab.jms.client.JmsQueue;
import org.exolab.jms.client.JmsSessionStubIfc;
import org.exolab.jms.client.JmsTopic;
import org.exolab.jms.jndi.JndiConstants;
import org.exolab.jms.message.MessageImpl;


/**
 * The client side stub implementing the JmsServerSession. All session
 * requests are passed on to the server.
 *
 * @version     $Revision: 1.16 $ $Date: 2003/08/25 03:35:45 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.core.http.HttpClient
 */
public class HttpJmsSessionStub implements JmsSessionStubIfc, NotifierIfc {

    /**
     * A reference to the client connection
     */
    private IpcIfc _connection = null;

    /**
     * The client connection id
     */
    private String _clientId;

    /**
     * The destination connection id
     */
    private String _connectionId;

    /**
     * The session id
     */
    private String _sessionId;

    /**
     * The tcp connection for receiving messages
     */
    private Server _msgReceiver = null;

    /**
     *  The listener
     */
    private JmsMessageListener _listener = null;

    /**
     * The url of the client's servlet. This gets passed to the server to
     * allow the server to connect back via client's servlet.
     * This host and port of the URL default to that of the server's web
     * This can be overriden by specifying the system property
     * <em>JndiConstants.HTTP_CLIENT_URL_PROPERTY</em>
     */
    private String _url;

    /**
     * The host of the machine running the client. Defaults to the local
     * host address. This can be overriden by specifying the system property
     * <em>JndiConstants.HTTP_CLIENT_SERVER_ADDRESS_PROPERTY</em>
     * which is useful if the client machine has multiple NICS.
     */
    private String _host = "localhost";

    /**
     * The client servlet
     */
    private static final String CLIENT_SERVLET = "openjms/OpenJMSClient";

    /**
     * The logger
     */
    private static final Log _log =
        LogFactory.getLog(HttpJmsSessionStub.class);


    /**
     * A new session has been established with these ids.
     *
     * @param connection  The http connection to the server.
     * @param clientId This clients unique id.
     * @param connectionId This objects connection identifier.
     * @param sessionId The unique session id for this object.
     */
    public HttpJmsSessionStub(IpcIfc connection, String clientId,
                              String connectionId, String sessionId) {
        _connection = connection;
        _clientId = clientId;
        _connectionId = connectionId;
        _sessionId = sessionId;
        String str = null;

        String clientHost = System.getProperty(
            JndiConstants.HTTP_CLIENT_SERVER_ADDRESS_PROPERTY);
        if (clientHost != null) {
            _host = clientHost;
        }
        if (_host.equals("localhost")) {
            try {
                _host = InetAddress.getLocalHost().getHostAddress();
            } catch (UnknownHostException ignore) {
            }
        }

        String clientWebServerURL = System.getProperty(
            JndiConstants.HTTP_CLIENT_URL_PROPERTY);

        if (clientWebServerURL == null) {
            // default to the host and port of the server's webserver
            URL url = ((HttpClient) _connection).getURL();
            clientWebServerURL = url.getProtocol() + "://" + url.getHost() +
                ":" + url.getPort() + "/";
        }

        if (!clientWebServerURL.endsWith("/")) {
            clientWebServerURL += "/";
        }
        _url = clientWebServerURL + CLIENT_SERVLET;
    }

    /**
     * Get the client Id
     *
     * @return the client id
     */
    public String getClientId() {
        return _clientId;
    }

    /**
     * Get the sessionId
     *
     * @return the id of this session
     */
    public String getSessionId() {
        return _sessionId;
    }

    // implementation of JmsSessionStubIfc.beforeClose
    public void beforeClose() throws JMSException {
    }

    /**
     * Close this session.
     *
     * @throws JMSException if the close fails
     */
    public void close() throws JMSException {
        Vector v = pack("close", 0);
        synchronized (_connection) {
            send(v, true);
            checkReply("close");
        }

        _listener = null;
        if (_msgReceiver != null) {
            if (_log.isDebugEnabled()) {
                _log.debug("Stopping receiver on port="
                           + _msgReceiver.getPort());
            }
            _msgReceiver.stop();
            _msgReceiver = null;
        }
    }

    /**
     * Acknowledge a message
     *
     * @param clientId the identity of the client
     * @param messageId the message identity to ack
     * @throws JMSException if the message can't be acknowledged
     */
    public void acknowledgeMessage(long clientId, String messageId)
        throws JMSException {

        Vector v = pack("acknowledgeMessage", 2);
        //Send back the ack to the consumer endpoint that forwarded the
        // message.
        v.add(new Long(clientId));
        v.add(messageId);
        synchronized (_connection) {
            send(v, true);
            checkReply("acknowledgeMessage");
        }
    }

    // implementation of JmsSessionStubIfc.sendMessage
    public void sendMessage(Message message) throws JMSException {
        Vector v = pack("sendMessage", 1);
        v.add(message);
        synchronized (_connection) {
            if (((MessageImpl) message).isPersistent()) {
                send(v, true);
                checkReply("sendMessage");
            } else {
                send(v, false);
            }

        }
    }

    // implementation of JmsSessionStubIfc.sendMessages
    public void sendMessages(Vector messages) throws JMSException {
        Vector v = pack("sendMessages", 1);
        v.add(messages);
        synchronized (_connection) {
            send(v, true);
            // always check for a reply
            checkReply("sendMessages");
        }
    }

    // implementation of JmsSessionStubIfc.receiveMessage
    public Message receiveMessage(long clientId, long wait)
        throws JMSException {

        Message message = null;

        Vector v = pack("receiveMessage", 2);
        v.add(new Long(clientId));
        v.add(new Long(wait));
        synchronized (_connection) {
            send(v, true);
            Vector reply = checkReply("receiveMessage");
            Boolean result = (Boolean) reply.get(0);

            // check that the call completed before
            // extracting the message
            if (result.booleanValue()) {
                message = (Message) reply.get(1);
            }
        }

        return message;
    }

    // implementation of JmsSessionStubIfc.receiveMessages
    public Vector receiveMessages(long clientId, int count)
        throws JMSException {

        Vector messages = null;

        Vector v = pack("receiveMessages", 2);
        v.add(new Long(clientId));
        v.add(new Integer(count));
        synchronized (_connection) {
            send(v, true);
            Vector reply = checkReply("receiveMessages");
            Boolean result = (Boolean) reply.get(0);

            // check that the call completed before
            // extracting the message
            if (result.booleanValue()) {
                messages = (Vector) reply.get(1);
            }
        }

        return messages;
    }

    /**
     * Create a new queue
     *
     * @param topic the queue to create
     * @throws JMSException if the queue can't be created
     */
    public void createQueue(JmsQueue queue) throws JMSException {
        Vector v = pack("createQueue", 1);
        v.add(queue);
        synchronized (_connection) {
            send(v, true);
            checkReply("createQueue");
        }
    }

    /**
     * Create a new topic
     *
     * @param topic the topic to create
     * @throws JMSException if the topic can't be created
     */
    public void createTopic(JmsTopic topic) throws JMSException {
        Vector v = pack("createTopic", 1);
        v.add(topic);
        synchronized (_connection) {
            send(v, true);
            checkReply("createTopic");
        }
    }

    /**
     * Create a receiver. Get the IP address of the machine the consumer runs
     * on, and the port it is listening to, and pass this to the server, so
     * it can make a new dedicated connection for sending all messages to
     * this client.
     *
     * @param queue the queue to listen to
     * @param clientId the session allocated identifier
     * @param selector the selector to filter messages (may be null)
     * @throws JMSException if the receiver cannot be created
     */
    public void createReceiver(JmsQueue queue, long clientId, String selector)
        throws JMSException {

        startReceiver();
        Vector v = pack("createReceiver", 6);
        v.add(queue);
        v.add(new Long(clientId));
        v.add(selector);
        v.add(_host);
        v.add(String.valueOf(_msgReceiver.getPort()));
        v.add(_url);
        synchronized (_connection) {
            send(v, true);
            checkReply("createReceiver");
        }
    }

    /**
     * Create a queue sender
     *
     * @param queue the queue to send messages to
     * @throws JMSException if the sender cannot be created
     */
    public void createSender(JmsQueue queue) throws JMSException {
        Vector v = pack("createSender", 1);
        v.add(queue);
        synchronized (_connection) {
            send(v, true);
            checkReply("createSender");
        }
    }

    /**
     * Create a queue browser for this session. This allows clients to browse
     * a queue without removing any messages.
     *
     * @param queue the queue to browse
     * @param clientId the identity of the client
     * @param selector the message selector. May be null
     * @throws JMSException if the browser can't be created
     */
    public void createBrowser(JmsQueue queue, long clientId, String selector)
        throws JMSException {

        startReceiver();
        Vector v = pack("createBrowser", 6);
        v.add(queue);
        v.add(new Long(clientId));
        v.add(selector);
        v.add(_host);
        v.add(String.valueOf(_msgReceiver.getPort()));
        v.add(_url);
        synchronized (_connection) {
            send(v, true);
            checkReply("createBrowser");
        }
    }

    /**
     * Delete the receiver for this queue.
     *
     * @param clientId the id of the client to delete
     * @throws JMSException if the receiver cannot be deleted
     */
    public void deleteReceiver(long clientId) throws JMSException {
        Vector v = pack("deleteReceiver", 1);
        v.add(new Long(clientId));
        synchronized (_connection) {
            send(v, true);
            checkReply("deleteReceiver");
        }
    }

    /**
     * Delete the queue browser associated with the specified queue from
     * the session.
     *
     * @param clientId the identity of the browser
     * @throws JMSException if the browser cannot be deleted
     */
    public void deleteBrowser(long clientId) throws JMSException {
        Vector v = pack("deleteBrowser", 1);
        v.add(new Long(clientId));
        synchronized (_connection) {
            send(v, true);
            checkReply("deleteBrowser");
        }
    }

    /**
     * Create a new topic subscriber
     *
     * @param topic the topic to subscribe to
     * @param name the subscribers name
     * @param client the client identity
     * @param selector the selector to filter messages (may be null)
     * @throws JMSException if the topic subscriber can't be created
     */
    public void createSubscriber(JmsTopic topic, String name, long clientId,
                                 String selector, boolean noLocal)
        throws JMSException {

        startReceiver();
        Vector v = pack("createSubscriber", 8);
        v.add(topic);
        v.add(name);
        v.add(new Long(clientId));
        v.add(selector);
        v.add(new Boolean(noLocal));
        v.add(_host);
        v.add(String.valueOf(_msgReceiver.getPort()));
        v.add(_url);
        synchronized (_connection) {
            send(v, true);
            checkReply("createSubscriber");
        }
    }

    /**
     * Create a new topic publisher
     *
     * @param topic the topic to publish to
     * @throws JMSException if the publisher can't be created
     */
    public void createPublisher(JmsTopic topic) throws JMSException {
        Vector v = pack("createPublisher", 1);
        v.add(topic);
        synchronized (_connection) {
            send(v, true);
            checkReply("createPublisher");
        }
    }

    /**
     * Unsubscribe a durable subscription
     *
     * @param name the name used to identify the subscription
     * @throws JMSException if the subscription cannot be removed
     */
    public void unsubscribe(String name) throws JMSException {
        Vector v = pack("unsubscribe", 1);
        v.add(name);
        synchronized (_connection) {
            send(v, true);
            checkReply("unsubscribe");
        }
    }

    /**
     * Delete the subscriber for this topic
     *
     * @param clientId the client identity
     * @throws JMSException for any error
     */
    public void deleteSubscriber(long clientId) throws JMSException {
        Vector v = pack("deleteSubscriber", 1);
        v.add(new Long(clientId));
        synchronized (_connection) {
            send(v, true);
            checkReply("deleteSubscriber");
        }
    }

    /**
     * Stop message delivery for this session.
     *
     * @throws JMSException for any error
     */
    public void stopMessageDelivery() throws JMSException {
        Vector v = pack("stopMessageDelivery", 0);
        synchronized (_connection) {
            send(v, true);
            checkReply("stopMessageDelivery");
        }
    }

    /**
     * Start message delivery for this session.
     *
     * @throws JMSException for any error
     */
    public void startMessageDelivery() throws JMSException {
        Vector v = pack("startMessageDelivery", 0);
        synchronized (_connection) {
            send(v, true);
            checkReply("startMessageDelivery");
        }
    }

    // implementation of JmsSessionStubIfc.recover
    public void recover() throws JMSException {
        Vector v = pack("recover", 0);
        synchronized (_connection) {
            send(v, true);
            checkReply("recover");
        }
    }

    // implementation of JmsSessionStubIfc.commit
    public void commit() throws JMSException {
        Vector v = pack("commit", 0);
        synchronized (_connection) {
            send(v, true);
            checkReply("commit");
        }
    }

    // implementation of JmsSessionStubIfc.rollback
    public void rollback() throws JMSException {
        Vector v = pack("rollback", 0);
        synchronized (_connection) {
            send(v, true);
            checkReply("rollback");
        }
    }

    // implementation of JmsSessionStubIfc.commit
    public void commit(Xid xid, boolean onePhase) throws XAException {
        try {
            Vector v = pack("xa_commit", 2);
            v.add(xid);
            v.add(new Boolean(onePhase));
            synchronized (_connection) {
                send(v, true);
                checkReply("xa_commit");
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to commit session " +
                exception);
        }
    }

    // implementation of JmsSessionStubIfc.end
    public void end(Xid xid, int flags) throws XAException {
        try {
            Vector v = pack("xa_end", 2);
            v.add(xid);
            v.add(new Integer(flags));
            synchronized (_connection) {
                send(v, true);
                checkReply("xa_end");
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to commit session " +
                exception);
        }
    }

    // implementation of JmsSessionStubIfc.forget
    public void forget(Xid xid) throws XAException {
        try {
            Vector v = pack("xa_forget", 1);
            v.add(xid);
            synchronized (_connection) {
                send(v, true);
                checkReply("xa_forget");
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to commit session " +
                exception);
        }
    }

    // implementation of JmsSessionStubIfc.getTransactionTimeout
    public int getTransactionTimeout() throws XAException {
        int timeout = 0;

        try {
            Vector v = pack("xa_getTransactionTimeout", 0);
            synchronized (_connection) {
                send(v, true);
                Vector reply = checkReply("xa_getTransactionTimeout");
                Boolean result = (Boolean) reply.get(0);

                // check that the call completed before
                // extracting the message
                if (result.booleanValue()) {
                    timeout = ((Integer) reply.get(1)).intValue();
                }
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to getTransactionTimeout session " +
                exception);
        }

        return timeout;
    }

    // implementation of JmsSessionStubIfc.prepare
    public int prepare(Xid xid) throws XAException {
        int value = 0;

        try {
            Vector v = pack("xa_prepare", 1);
            v.add(xid);
            synchronized (_connection) {
                send(v, true);
                Vector reply = checkReply("xa_prepare");
                Boolean result = (Boolean) reply.get(0);

                // check that the call completed before
                // extracting the message
                if (result.booleanValue()) {
                    value = ((Integer) reply.get(1)).intValue();
                }
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to prepare session " +
                exception);
        }

        return value;
    }

    // implementation of JmsSessionStubIfc.recover
    public Xid[] recover(int flag) throws XAException {
        Xid[] xids = new Xid[0];

        try {
            Vector v = pack("xa_recover", 1);
            v.add(new Integer(flag));
            synchronized (_connection) {
                send(v, true);
                Vector reply = checkReply("xa_recover");
                Boolean result = (Boolean) reply.get(0);

                // check that the call completed before
                // extracting the message
                if (result.booleanValue()) {
                    xids = (Xid[]) reply.get(1);
                }
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to recover session " +
                exception);
        }

        return xids;
    }

    // implementation of JmsSessionStubIfc.rollback
    public void rollback(Xid xid) throws XAException {
        try {
            Vector v = pack("xa_rollback", 1);
            v.add(xid);
            synchronized (_connection) {
                send(v, true);
                checkReply("xa_rollback");
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to rollback session " +
                exception);
        }
    }

    // implementation of JmsSessionStubIfc.setTransactionTimeout
    public boolean setTransactionTimeout(int seconds) throws XAException {
        boolean value = false;

        try {
            Vector v = pack("xa_setTransactionTimeout", 1);
            v.add(new Integer(seconds));
            synchronized (_connection) {
                send(v, true);
                Vector reply = checkReply("xa_setTransactionTimeout");
                Boolean result = (Boolean) reply.get(0);

                // check that the call completed before
                // extracting the message
                if (result.booleanValue()) {
                    value = ((Boolean) reply.get(1)).booleanValue();
                }
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to setTransactionTimeout " +
                exception);
        }

        return value;
    }

    // implementation of JmsSessionStubIfc.start
    public void start(Xid xid, int flags) throws XAException {
        try {
            Vector v = pack("xa_start", 2);
            v.add(xid);
            v.add(new Integer(flags));
            synchronized (_connection) {
                send(v, true);
                checkReply("xa_start");
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to start session " +
                exception);
        }
    }

    // implementation of JmsSessionStubIfc.getResourceManagerId
    public String getResourceManagerId() throws XAException {
        String rid = null;

        try {
            Vector v = pack("xa_getResourceManagerId", 0);
            synchronized (_connection) {
                send(v, true);
                Vector reply = checkReply("xa_getResourceManagerId");
                Boolean result = (Boolean) reply.get(0);

                // check that the call completed before
                // extracting the message
                if (result.booleanValue()) {
                    rid = (String) reply.get(1);
                }
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to getResourceManagerId " +
                exception);
        }

        return rid;
    }

    /**
     * Set a message listener to be called when new Messages arrive from the
     * server.
     *
     * @param listener A reference to the client listener.
     */
    public void setMessageListener(JmsMessageListener listener) {
        _listener = listener;
    }

    // implementation of JmsSessionStubIfc.enableAsynchronousDelivery
    public void enableAsynchronousDelivery(long clientId, String id,
                                           boolean enable)
        throws JMSException {

        Vector v = pack("enableAsynchronousDelivery", 3);
        v.add(new Long(clientId));
        v.add(id);
        v.add(new Boolean(enable));
        synchronized (_connection) {
            send(v, true);
            checkReply("enableAsynchronousDelivery");
        }
    }

    /**
     * Stop the message receiver thread.
     */
    public synchronized void stopReceiver() {
        _listener = null;
        if (_msgReceiver != null) {
            _msgReceiver.stop();
            _msgReceiver = null;
        }
    }

    /**
     * Pack all the data that is required by the server in a vector.
     * Set the size of the vector to be exactly the right size for efficiency.
     *
     * @param method The function to activate on the server.
     * @param numParams The number of paramaters this method will require.
     * @return Vector The vector containing all the data.
     */
    private Vector pack(String method, int numParams) {
        Vector v = new Vector(5 + numParams);
        v.add("org.exolab.jms.server.http.HttpJmsSessionConnection");
        packCommon(v, method);
        return v;
    }

    /**
     * Pack the common data required by all connection types.
     *
     * @param v The vector to add the common items to.
     * @param method The function to activate on the server.
     */
    private void packCommon(Vector v, String method) {
        v.add(method);
        v.add(_clientId);
        v.add(_connectionId);
        v.add(_sessionId);
    }

    /**
     * A convenience method to check the success of operations which return
     * a true on sucess.
     *
     * @param method The requested server function.
     * @throws JMSException for any error.
     */
    private Vector checkReply(String method) throws JMSException {
        Vector v = null;
        try {
            v = (Vector) _connection.receive();
        } catch (Exception err) {
            // rethrow as a JMSException
            throw new JMSException("Operation " + method + " failed: " + err);
        }

        if (v != null) {
            Boolean b = (Boolean) v.get(0);
            if (!b.booleanValue()) {
                if (v.get(1) instanceof JMSException) {
                    throw (JMSException) v.get(1);
                } else {
                    throw new JMSException("Operation " + method +
                        " failed: " + v.get(1));
                }
            }
        } else {
            throw new JMSException("Unknown connection error for " + method);
        }

        return v;
    }

    /**
     * A message has been received.
     *
     *
     * @param ob The data received,
     * @param id The connection id this data is received from
     * @return any requested result, or null, nothing is sent back to the
     * client
     */
    public Serializable notify(Object ob, String id) {
        Vector v = (Vector) ob;
        Vector reply = new Vector();

        if (_listener != null) {
            if (v.size() == 2) {
                if (v.get(1) instanceof Message) {
                    _listener.onMessage((Message) v.get(1));
                } else if (v.get(1) instanceof Vector) {
                    _listener.onMessages((Vector) v.get(1));
                } else if (v.get(1) instanceof Long) {
                    _listener.onMessageAvailable(((Long) v.get(1)).longValue());
                }
            } else {
                // if size == 1, then this is just a ping.
            }
        }
        return reply;
    }

    /**
     * A client has disconnected. Notify the caller.
     *
     * @param The unique identifier of this connection.
     */
    public void disconnection(String id) {
    }


    /**
     * A convenience method to send a packed command to the server.
     *
     * @throws JMSException for any error.
     */
    private void send(Vector v, boolean replyExpected) throws JMSException {
        try {
            if (replyExpected) {
                _connection.send(v);
            } else {
                ((HttpClient) _connection).sendWithoutResponse(v);
            }
        } catch (Exception err) {
            // rethrow as a JMSException
            throw new JMSException("Operation Failed" + err);
        }
    }

    /**
     * Start a message receiver thread to receive messages
     *
     * @throws JMSException If the IpcConnection cannot be created
     */
    private synchronized void startReceiver() throws JMSException {
        try {
            if (_msgReceiver == null) {
                _msgReceiver = new Server(this);
                if (_log.isDebugEnabled()) {
                    _log.debug("Starting receiver on port="
                               + _msgReceiver.getPort());
                }
                new Thread(_msgReceiver).start();
            }
        } catch (Exception err) {
            throw new JMSException(err.getMessage());
        }
    }

} // --HttpJmsSessionStub
