/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: IpcJmsConnectionStub.java,v 1.8 2003/08/07 13:32:52 tanderson Exp $
 *
 * Date         Author  Changes
 * $Date	    jimm    Created
 */


package org.exolab.jms.client.mipc;

import java.util.Vector;

import javax.jms.JMSException;

import org.exolab.core.ipc.IpcIfc;
import org.exolab.jms.client.JmsConnectionStubIfc;
import org.exolab.jms.client.JmsSessionStubIfc;


/**
 * Used to create OpenJmsSessions to the server. It gets the new Session Id
 * assigned by the server and returns a IpcJmsSessionStub.
 *
 * @version     $Revision: 1.8 $ $Date: 2003/08/07 13:32:52 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.jms.client.mipc.IpcJmsServerStub
 * @see			org.exolab.jms.client.mipc.IpcJmsSessionStub
 * @see 		org.exolab.core.ipc.Client
 *
 **/


public class IpcJmsConnectionStub implements JmsConnectionStubIfc {

    /**
     * The owner of this instance
     */
    private IpcJmsServerStub _server;

    /**
     * The connection to the server
     */
    private IpcIfc _connection = null;

    /**
     * The client connection id
     */
    private String _clientId;

    /**
     * The destination connection id
     */
    private String _connectionId;

    /**
     * The listener for dispatching messages to sessions
     */
    private IpcJmsMessageListener _listener = null;

    /**
     * Construct a new <code>IpcJmsConnectionStub</code>
     *
     * @param server the owner of this instance
     * @param connection the connection to the server.
     * @param clientId this client's unique id.
     * @param connectionId this object's connection identifier.
     * @param listener the listener for managing session MessageListeners
     */
    public IpcJmsConnectionStub(IpcJmsServerStub server,
                                IpcIfc connection,
                                String clientId,
                                String connectionId,
                                IpcJmsMessageListener listener) {
        _server = server;
        _connection = connection;
        _clientId = clientId;
        _connectionId = connectionId;
        _listener = listener;
    }

    /**
     * Send a session connection message. Create a new IpcJmsSessionStub
     * if this createSession request was successfull.
     *
     * @param ackMode the ack mode for the session
     * @param transacted <code>true</code> if the session is transacted
     * @return JmsSessionStubIfc the session interface.
     * @throws JMSException on failure to create a session.
     */
    public JmsSessionStubIfc createSession(int ackMode, boolean transacted)
        throws JMSException {
        JmsSessionStubIfc stub = null;

        try {
            Vector v = pack("createSession", 2);
            synchronized (_connection) {
                v.add(new Integer(ackMode));
                v.add(new Boolean(transacted));
                _connection.send(v);
                v = (Vector) _connection.receive();
            }

            if (v != null) {
                Boolean b = (Boolean) v.get(0);
                if (b.booleanValue() == true) {
                    Vector state = (Vector) v.get(1);
                    String sessionId = (String) state.get(0);
                    stub = new IpcJmsSessionStub(
                        _connection, _clientId, _connectionId, sessionId,
                        _listener);
                } else {
                    throw new JMSException("Failed to create session: "
                        + (String) v.get(1));
                }
            } else {
                throw new JMSException("Unknown connection error for "
                    + "createSession");
            }
        } catch (JMSException exception) {
            throw exception;
        } catch (Exception exception) {
            // rethrow as a JMSException
            throw new JMSException("Failed to create session: " + exception);
        }

        return stub;
    }

    // implementation of JmsConnectionStubIfc.close
    public void close() throws JMSException {
        try {
            Vector v = pack("close", 0);
            synchronized (_connection) {
                _connection.send(v);
                v = (Vector) _connection.receive();
            }

            if (v != null) {
                Boolean b = (Boolean) v.get(0);
                if (!b.booleanValue()) {
                    throw new JMSException("Failed to close connection: " +
                        (String) v.get(1));
                }
            } else {
                throw new JMSException("Unknown connection error for " +
                    "close Connection");
            }
        } catch (JMSException exception) {
            throw exception;
        } catch (Exception exception) {
            // rethrow as a JMSException
            throw new JMSException("Failed to close connection: " + exception);
        } finally {
            _server.closed(this);
        }
    }

    // implementation of JmsConnectionStubIfc.getConnectionId
    public String getConnectionId() throws JMSException {
        return _connectionId;
    }

    // implementation of JmsConnectionStubIfc.destroy
    public void destroy() {
        _connection = null;
    }

    /**
     * Pack all the data that is required by the server in a vector.
     * Set the size of the vector to be exactly the right size for efficiency.
     *
     * @param method The function to activate on the server.
     * @param numParams The number of paramaters this method will require.
     * @return Vector The vector containing all the data.
     *
     */
    private Vector pack(String method, int numParams) {
        Vector v = new Vector(4 + numParams);
        v.add("org.exolab.jms.server.mipc.IpcJmsServerConnection");
        v.add(method);
        v.add(_clientId);
        v.add(_connectionId);
        return v;
    }

} //-- IpcConnectionStub
