/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: IpcJmsSessionStub.java,v 1.20 2003/08/25 03:30:33 tanderson Exp $
 *
 * Date         Author  Changes
 * $Date        jimm    Created
 */
package org.exolab.jms.client.mipc;

import java.util.Vector;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.ipc.IpcIfc;
import org.exolab.core.mipc.ObjectChannel;
import org.exolab.jms.client.JmsMessageListener;
import org.exolab.jms.client.JmsQueue;
import org.exolab.jms.client.JmsSessionStubIfc;
import org.exolab.jms.client.JmsTopic;
import org.exolab.jms.message.MessageImpl;


/**
 * The client side stub implementing the JmsServerSession. All session
 * requests are passed on to the server. This class also an IPC server
 * connection when a listener os subscriber is set up.
 *
 * <P>Note: There is only one receive connection per client, all
 * JmsMessages for any queue/topic that the client is interested in
 * receiving are multiplexed on this single connection. The connection
 * is not bi-directional, that is no replies are sent back using this
 * connection. It is assumed that the underlying IPC protocol will confirm
 * delivery to the client. After that it is the clients responsibility to
 * ensure correct message processing. Durable messages are acked on a
 * separate port, reliable messages, are not.
 *
 * @version     $Revision: 1.20 $ $Date: 2003/08/25 03:30:33 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.jms.server.mipc.IpcJmsSessionConnection
 * @see         org.exolab.core.ipc.Server
 */
public class IpcJmsSessionStub implements JmsSessionStubIfc {

    /**
     * The client connection
     */
    private IpcIfc _connection = null;

    /**
     * The client connection id
     */
    private String _clientId;

    /**
     * The destination connection id
     */
    private String _connectionId;

    /**
     * The session id
     */
    private String _sessionId;

    /**
     * The message dispatcher
     */
    private IpcJmsMessageListener _listener;

    /**
     * The logger
     */
    private static final Log _log = LogFactory.getLog(IpcJmsSessionStub.class);


    /**
     * A new session has been established with these ids.
     *
     * @param connection the connection to the server.
     * @param clientId this clients unique id.
     * @param connectionId this objects connection identifier.
     * @param sessionId the unique session id for this object.
     * @param listener the message dispatcher
     */
    public IpcJmsSessionStub(IpcIfc connection, String clientId,
                             String connectionId, String sessionId,
                             IpcJmsMessageListener listener) {
        _connection = connection;
        _clientId = clientId;
        _connectionId = connectionId;
        _sessionId = sessionId;
        _listener = listener;
    }


    /**
     * Get the client Id
     *
     * @return         String  The client id
     * @exception      JMSException On error
     *
     */
    public String getClientId() throws JMSException {
        return _clientId;
    }


    /**
     * Get the sessionId
     *
     * @return         String  The session id of this session
     * @exception      JMSException On error
     *
     */
    public String getSessionId() throws JMSException {
        return _sessionId;
    }

    // implementation of JmsSessionStubIfc.beforeClose
    public void beforeClose()
        throws JMSException {
    }

    /**
     * Close this session.
     *
     * @exception    JMSException On error
     *
     */
    public void close() throws JMSException {
        Vector v = pack("close", 0);
        synchronized (_connection) {
            send(v);
            checkReply("close");
        }

        _listener.closeSession(_sessionId);
    }


    /**
     * Extract the destination and messageId for the message and send back
     * an ack.
     *
     * @param       clientId        the identity ofthe client
     * @param       messageId       the message identity to ack
     * @exception   JMSException
     */
    public void acknowledgeMessage(long clientId, String messageId)
        throws JMSException {
        Vector v = pack("acknowledgeMessage", 2);
        //Send back the ack to the consumer endpoint that forwarded the
        // message.
        v.add(new Long(clientId));
        v.add(messageId);
        synchronized (_connection) {
            send(v);
            checkReply("acknowledgeMessage");
        }
    }


    // implementation of JmsSessionStubIfc.sendMessage
    public void sendMessage(Message message) throws JMSException {
        Vector v = pack("sendMessage", 1);
        v.add(message);
        synchronized (_connection) {
            send(v);
            if (((MessageImpl) message).isPersistent()) {
                checkReply("sendMessage");
            }
        }
    }

    // implementation of JmsSessionStubIfc.sendMessages
    public void sendMessages(Vector messages) throws JMSException {
        Vector v = pack("sendMessages", 1);
        v.add(messages);
        synchronized (_connection) {
            send(v);
            // always check for a reply
            checkReply("sendMessages");
        }
    }

    // implementation of JmsSessionStubIfc.receiveMessage
    public Message receiveMessage(long clientId, long wait)
        throws JMSException {
        Message message = null;

        Vector v = pack("receiveMessage", 2);
        v.add(new Long(clientId));
        v.add(new Long(wait));
        synchronized (_connection) {
            send(v);
            Vector reply = checkReply("receiveMessage");
            Boolean result = (Boolean) reply.get(0);

            // check that the call completed before
            // extracting the message
            if (result.booleanValue()) {
                message = (Message) reply.get(1);
            }
        }

        return message;
    }

    // implementation of JmsSessionStubIfc.receiveMessages
    public Vector receiveMessages(long clientId, int count)
        throws JMSException {
        Vector messages = null;

        Vector v = pack("receiveMessages", 2);
        v.add(new Long(clientId));
        v.add(new Integer(count));
        synchronized (_connection) {
            send(v);
            Vector reply = checkReply("receiveMessages");
            Boolean result = (Boolean) reply.get(0);

            // check that the call completed before
            // extracting the message
            if (result.booleanValue()) {
                messages = (Vector) reply.get(1);
            }
        }

        return messages;
    }

    /**
     * Create a new Queue.
     *
     * @param         queue  The queue to create.
     * @exception     JMSException On error
     *
     */
    public void createQueue(JmsQueue queue) throws JMSException {
        Vector v = pack("createQueue", 1);
        v.add(queue);
        synchronized (_connection) {
            send(v);
            checkReply("createQueue");
        }
    }


    /**
     * Create a new topic
     *
     * @param         topic  The topic to create.
     * @exception     JMSException On error
     *
     */
    public void createTopic(JmsTopic topic) throws JMSException {
        Vector v = pack("createTopic", 1);
        v.add(topic);
        synchronized (_connection) {
            send(v);
            checkReply("createTopic");
        }
    }


    /**
     * Create a receiver. Get the IP address of the machine the consumer runs
     * on, and the port it is listening too, and pass this to the server, so
     * it can make a new dedicated connection for sending all messages to
     * this client.
     *
     * @param         queue  The queue to listen to
     * @param         clientId The session allocated identifier
     * @param         selector The selector to filter messages (may be null)
     * @exception     JMSException On error
     *
     */
    public void createReceiver(JmsQueue queue, long clientId, String selector)
        throws JMSException {
        _listener.start();
        Vector v = pack("createReceiver", 5);
        v.add(queue);
        v.add(new Long(clientId));
        v.add(selector);
        v.add(((ObjectChannel) _connection).getConnection().getHost());
        v.add(String.valueOf
            (((ObjectChannel) _connection).getConnection().getPort()));
        v.add("n");
        synchronized (_connection) {
            send(v);
            checkReply("createReceiver");
        }
    }


    /**
     * Create a queue sender
     *
     * @param         queue  The queue to send messages to
     * @exception     JMSException On error
     *
     */
    public void createSender(JmsQueue queue) throws JMSException {
        Vector v = pack("createSender", 1);
        v.add(queue);
        synchronized (_connection) {
            send(v);
            checkReply("createSender");
        }
    }

    /**
     * Create a queue browser for this session. This allows clients to browse
     * a queue without removing any messages.
     * <p>
     *
     * You cannot create more than one queue browser for the same queue
     * in a single session.
     *
     * @param       queue               queue to browse
     * @param       clientId            idenity of the client
     * @param       selector            message selector. This may be null
     * @exception   JMSException
     */
    public void createBrowser(JmsQueue queue, long clientId, String selector)
        throws JMSException {
        _listener.start();
        Vector v = pack("createBrowser", 3);
        v.add(queue);
        v.add(new Long(clientId));
        v.add(selector);
        v.add(((ObjectChannel) _connection).getConnection().getHost());
        v.add(String.valueOf
            (((ObjectChannel) _connection).getConnection().getPort()));
        v.add("n");
        synchronized (_connection) {
            send(v);
            checkReply("createBrowser");
        }
    }

    /**
     * Delete the receiver for this queue.
     *
     * @param         clientId  The id of the client to delete
     * @exception     JMSException On error
     *
     */
    public void deleteReceiver(long clientId) throws JMSException {
        Vector v = pack("deleteReceiver", 1);
        v.add(new Long(clientId));
        synchronized (_connection) {
            send(v);
            checkReply("deleteReceiver");
        }
    }


    /**
     * Delete the queue browser associated with the specified queue from
     * the session.
     * If the corresponding queue does not exist or it cannot be deleted,
     * then throw a JMSException
     *
     * @param       clientId            identity of the browser
     * @exception   JMSException
     */
    public void deleteBrowser(long clientId)
        throws JMSException {
        Vector v = pack("deleteBrowser", 1);
        v.add(new Long(clientId));
        synchronized (_connection) {
            send(v);
            checkReply("deleteBrowser");
        }
    }

    /**
     * Create a new topic subscriber
     *
     * @param         topic  The topic to subscribe to
     * @param         name The subscribers name
     * @param         client The client identity
     * @param         selector The selector to filter messages (may be null)
     * @exception     JMSException On error
     *
     */
    public void createSubscriber(JmsTopic topic, String name, long clientId,
                                 String selector, boolean noLocal)
        throws JMSException {
        _listener.start();
        Vector v = pack("createSubscriber", 5);
        v.add(topic);
        v.add(name);
        v.add(new Long(clientId));
        v.add(selector);
        v.add(new Boolean(noLocal));
        v.add(((ObjectChannel) _connection).getConnection().getHost());
        v.add(String.valueOf
            (((ObjectChannel) _connection).getConnection().getPort()));
        v.add("n");
        synchronized (_connection) {
            send(v);
            checkReply("createSubscriber");
        }
    }


    /**
     * Create a new topic publisher
     *
     * @param         topic  The topic to publish to
     * @exception     JMSException On error
     *
     */
    public void createPublisher(JmsTopic topic) throws JMSException {
        Vector v = pack("createPublisher", 1);
        v.add(topic);
        synchronized (_connection) {
            send(v);
            checkReply("createPublisher");
        }
    }


    /**
     * Unsubscribe a durable subscription
     *
     * @param       name                the name used to identify the
     *                                  subscription
     * @exception   JMSException        if the subscription cannot be removed
     */
    public void unsubscribe(String name)
        throws JMSException {
        Vector v = pack("unsubscribe", 1);
        v.add(name);
        synchronized (_connection) {
            send(v);
            checkReply("unsubscribe");
        }
    }

    /**
     * Delete the subscriber for this topic
     *
     * @param         clientId - the client identity
     * @exception     JMSException On error
     *
     */
    public void deleteSubscriber(long clientId) throws JMSException {
        Vector v = pack("deleteSubscriber", 1);
        v.add(new Long(clientId));
        synchronized (_connection) {
            send(v);
            checkReply("deleteSubscriber");
        }
    }

    /**
     * Stop message delivery for this session.
     *
     * @exception     JMSException On error
     *
     */
    public void stopMessageDelivery() throws JMSException {
        Vector v = pack("stopMessageDelivery", 0);
        synchronized (_connection) {
            send(v);
            checkReply("stopMessageDelivery");
        }
    }


    /**
     * Start message delivery for this session.
     *
     * @exception     JMSException On error
     *
     */
    public void startMessageDelivery() throws JMSException {
        Vector v = pack("startMessageDelivery", 0);
        synchronized (_connection) {
            send(v);
            checkReply("startMessageDelivery");
        }
    }

    // implementation of JmsSessionStubIfc.recover
    public void recover() throws JMSException {
        Vector v = pack("recover", 0);
        synchronized (_connection) {
            send(v);
            checkReply("recover");
        }
    }

    // implementation of JmsSessionStubIfc.commit
    public void commit() throws JMSException {
        Vector v = pack("commit", 0);
        synchronized (_connection) {
            send(v);
            checkReply("commit");
        }
    }

    // implementation of JmsSessionStubIfc.rollback
    public void rollback() throws JMSException {
        Vector v = pack("rollback", 0);
        synchronized (_connection) {
            send(v);
            checkReply("rollback");
        }
    }

    // implementation of JmsSessionStubIfc.commit
    public void commit(Xid xid, boolean onePhase)
        throws XAException {
        try {
            Vector v = pack("xa_commit", 2);
            v.add(xid);
            v.add(new Boolean(onePhase));
            synchronized (_connection) {
                send(v);
                checkReply("xa_commit");
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to commit session " +
                exception);
        }
    }

    // implementation of JmsSessionStubIfc.end
    public void end(Xid xid, int flags)
        throws XAException {
        try {
            Vector v = pack("xa_end", 2);
            v.add(xid);
            v.add(new Integer(flags));
            synchronized (_connection) {
                send(v);
                checkReply("xa_end");
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to commit session " +
                exception);
        }
    }

    // implementation of JmsSessionStubIfc.forget
    public void forget(Xid xid)
        throws XAException {
        try {
            Vector v = pack("xa_forget", 1);
            v.add(xid);
            synchronized (_connection) {
                send(v);
                checkReply("xa_forget");
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to commit session " +
                exception);
        }
    }

    // implementation of JmsSessionStubIfc.getResourceManagerId
    public String getResourceManagerId() throws XAException {
        String rid = null;

        try {
            Vector v = pack("xa_getResourceManagerId", 0);
            synchronized (_connection) {
                send(v);
                Vector reply = checkReply("xa_getResourceManagerId");
                Boolean result = (Boolean) reply.get(0);

                // check that the call completed before
                // extracting the message
                if (result.booleanValue()) {
                    rid = (String) reply.get(1);
                }
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to getResourceManagerId session " +
                exception);
        }

        return rid;
    }

    // implementation of JmsSessionStubIfc.getTransactionTimeout
    public int getTransactionTimeout()
        throws XAException {
        int timeout = 0;

        try {
            Vector v = pack("xa_getTransactionTimeout", 0);
            synchronized (_connection) {
                send(v);
                Vector reply = checkReply("xa_getTransactionTimeout");
                Boolean result = (Boolean) reply.get(0);

                // check that the call completed before
                // extracting the message
                if (result.booleanValue()) {
                    timeout = ((Integer) reply.get(1)).intValue();
                }
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to getTransactionTimeout session " +
                exception);
        }

        return timeout;
    }

    // implementation of JmsSessionStubIfc.prepare
    public int prepare(Xid xid)
        throws XAException {
        int value = 0;

        try {
            Vector v = pack("xa_prepare", 1);
            v.add(xid);
            synchronized (_connection) {
                send(v);
                Vector reply = checkReply("xa_prepare");
                Boolean result = (Boolean) reply.get(0);

                // check that the call completed before
                // extracting the message
                if (result.booleanValue()) {
                    value = ((Integer) reply.get(1)).intValue();
                }
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to prepare session " +
                exception);
        }

        return value;
    }

    // implementation of JmsSessionStubIfc.recover
    public Xid[] recover(int flag)
        throws XAException {
        Xid[] xids = new Xid[0];

        try {
            Vector v = pack("xa_recover", 1);
            v.add(new Integer(flag));
            synchronized (_connection) {
                send(v);
                Vector reply = checkReply("xa_recover");
                Boolean result = (Boolean) reply.get(0);

                // check that the call completed before
                // extracting the message
                if (result.booleanValue()) {
                    xids = (Xid[]) reply.get(1);
                }
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to recover session " +
                exception);
        }

        return xids;
    }

    // implementation of JmsSessionStubIfc.rollback
    public void rollback(Xid xid)
        throws XAException {
        try {
            Vector v = pack("xa_rollback", 1);
            v.add(xid);
            synchronized (_connection) {
                send(v);
                checkReply("xa_rollback");
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to rollback session " +
                exception);
        }
    }

    // implementation of JmsSessionStubIfc.setTransactionTimeout
    public boolean setTransactionTimeout(int seconds)
        throws XAException {
        boolean value = false;

        try {
            Vector v = pack("xa_setTransactionTimeout", 1);
            v.add(new Integer(seconds));
            synchronized (_connection) {
                send(v);
                Vector reply = checkReply("xa_setTransactionTimeout");
                Boolean result = (Boolean) reply.get(0);

                // check that the call completed before
                // extracting the message
                if (result.booleanValue()) {
                    value = ((Boolean) reply.get(1)).booleanValue();
                }
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to setTransactionTimeout " +
                exception);
        }

        return value;
    }

    // implementation of JmsSessionStubIfc.start
    public void start(Xid xid, int flags)
        throws XAException {
        try {
            Vector v = pack("xa_start", 2);
            v.add(xid);
            v.add(new Integer(flags));
            synchronized (_connection) {
                send(v);
                checkReply("xa_start");
            }
        } catch (JMSException exception) {
            // rethrow as a XAException
            throw new XAException("Failed to start session " +
                exception);
        }
    }

    /**
     * Set a message listener to be called when new Messages arrive from the
     * server.
     *
     * @param listener A reference to the client listener.
     */
    public void setMessageListener(JmsMessageListener listener) {
        _listener.setListener(_sessionId, listener);
    }

    // implementation of JmsSessionStubIfc.enableAsynchronousDelivery
    public void enableAsynchronousDelivery(long clientId, String id,
                                           boolean enable)
        throws JMSException {
        Vector v = pack("enableAsynchronousDelivery", 3);
        v.add(new Long(clientId));
        v.add(id);
        v.add(new Boolean(enable));
        synchronized (_connection) {
            send(v);
            checkReply("enableAsynchronousDelivery");
        }
    }

    /**
     * Pack all the data that is required by the server in a vector.
     * Set the size of the vector to be exactly the right size for efficiency.
     *
     * @param method The function to activate on the server.
     * @param numParams The number of paramaters this method will require.
     * @return Vector The vector containing all the data.
     *
     */
    private Vector pack(String method, int numParams) {
        Vector v = new Vector(5 + numParams);
        v.add("org.exolab.jms.server.mipc.IpcJmsSessionConnection");
        v.add(method);
        v.add(_clientId);
        v.add(_connectionId);
        v.add(_sessionId);
        return v;
    }

    /**
     * A convenience method to check the success of operations which return
     * a true on sucess.
     *
     * @param method The requested server function.
     * @exception JMSException On any failure.
     *
     */
    private Vector checkReply(String method) throws JMSException {
        Vector v = null;
        try {
            v = (Vector) _connection.receive();
        } catch (Exception err) {
            // rethrow as a JMSException
            throw new JMSException("Operation " + method + " failed:\n" + err);
        }

        if (v != null) {
            Boolean b = (Boolean) v.get(0);
            if (!b.booleanValue()) {
                if (v.get(1) instanceof JMSException) {
                    throw (JMSException) v.get(1);
                } else {
                    throw new JMSException("Operation " + method +
                        " failed:\n" + v.get(1));
                }
            }
        } else {
            throw new JMSException("Unknown connection error for " + method);
        }

        return v;
    }

    /**
     * A convenience method to send a packed command to the server.
     *
     * @throws JMSException for any failure.
     */
    private void send(Vector v) throws JMSException {
        try {
            _connection.send(v);
        } catch (Exception err) {
            // rethrow as a JMSException
            throw new JMSException("Operation Failed" + err);
        }
    }

} //-- IpcJmsSessionStub
