/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2001,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: IpcJndiInitialContextFactory.java,v 1.22.2.1 2004/05/01 06:10:04 tanderson Exp $
 */

package org.exolab.jms.jndi.mipc;

import java.util.Hashtable;

import javax.naming.ConfigurationException;
import javax.naming.Context;
import javax.naming.NameParser;
import javax.naming.NamingException;
import javax.naming.ServiceUnavailableException;
import javax.naming.spi.InitialContextFactory;

import org.apache.avalon.excalibur.naming.DefaultNamespace;
import org.apache.avalon.excalibur.naming.RemoteContext;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.ipc.Client;
import org.exolab.jms.jndi.JndiConstants;


/**
 * This is the client side helper class that retrieves the root context
 * of the JNDI server. This class implements the InitialContextFactory
 * interface.
 *
 * @version     $Revision: 1.22.2.1 $ $Date: 2004/05/01 06:10:04 $
 * @author      <a href="mailto:mourikis@intalio.com">Jim Mourikis</a>
 * @see         org.exolab.jms.server.mipc.IpcJndiServer
 */
public class IpcJndiInitialContextFactory implements InitialContextFactory {

    /**
     * The allowed scheme names for the provider URL
     */
    private static final String TCP_SCHEME = "tcp://";
    private static final String TCPS_SCHEME = "tcps://";

    /**
     * The url scheme separator
     */
    private final static String SCHEME_SEPARATOR = "://";

    /**
     * The host/port separator
     */
    private final static String HOST_PORT_SEPARATOR = ":";

    /**
     * The end separator
     */
    private final static String URL_TERMINATOR = "/";

    /**
     * The server host address
     */
    private String _host;

    /**
     * The port number the server is on
     */
    private int _port;

    /**
     * The logger
     */
    private static final Log _log =
        LogFactory.getLog(IpcJndiInitialContextFactory.class);


    /**
     * Default constructor
     */
    public IpcJndiInitialContextFactory() {
    }

    /**
     * Creates an initial context for beginning name resolution, based on
     * the {@link Context#PROVIDER_URL} attribute.
     *
     * @param environment the environment specifying information to be used in
     * the creation of the initial context.
     * @return an initial context
     * @throws NamingException if the initial context cannot be created
     */
    public Context getInitialContext(Hashtable environment)
        throws NamingException {
        Context context = null;

        String url = (String) environment.get(Context.PROVIDER_URL);
        if (url == null || url.length() == 0) {
            url = getDeprecatedEnvironment(environment);
        }
        parseURL(url);

        Client connection = openConnection();
        NameServiceProxy provider = new NameServiceProxy(connection);
        NameParser parser;
        try {
            parser = provider.getNameParser();
        } catch (NamingException exception) {
            throw exception;
        } catch (Exception exception) {
            NamingException error = new ServiceUnavailableException(
                exception.getMessage());
            error.setRootCause(exception);
            throw error;
        }
        DefaultNamespace namespace = new DefaultNamespace(parser);
        Hashtable env = new Hashtable(environment);
        env.put(RemoteContext.NAMING_PROVIDER, provider);
        env.put(RemoteContext.NAMESPACE, namespace);
        return new RemoteContext(env, parser.parse(""));
    }

    private void parseURL(String url) throws NamingException {
        if (url.startsWith("mipc://")) {
            // TODO: remove this when removing getDeprecatedEnvironment
            _log.warn(getClass().getName() +
                ": using deprecated Context.PROVIDER_URL=" + url);
            url = TCP_SCHEME + url.substring(7);
        } else if (!url.startsWith(TCP_SCHEME) &&
            !url.startsWith(TCPS_SCHEME)) {
            throw new ConfigurationException(
                "URL=" + url + " has an invalid scheme");
        }

        int schemeIndex = url.indexOf("://");
        int hostIndex = url.indexOf(HOST_PORT_SEPARATOR, schemeIndex +
            SCHEME_SEPARATOR.length());
        if (hostIndex == -1) {
            throw new ConfigurationException("URL=" + url + " is invalid");
        }

        int endIndex = url.indexOf(URL_TERMINATOR, hostIndex +
            HOST_PORT_SEPARATOR.length());
        if (endIndex != -1 && endIndex != (url.length() - 1)) {
            // url can have trailing '/', but nothing after it
            throw new ConfigurationException("URL=" + url + " is invalid");
        }

        // get the server address
        _host = url.substring(schemeIndex + SCHEME_SEPARATOR.length(),
            hostIndex);

        // get the port address
        try {
            int start = hostIndex + HOST_PORT_SEPARATOR.length();
            String port = (endIndex == -1) ? url.substring(start) :
                url.substring(start, endIndex);
            _port = Integer.parseInt(port);
        } catch (Exception exception) {
            throw new ConfigurationException("URL=" + url + " is invalid");
        }
    }

    /**
     * Get a connection to the specified host and port.
     *
     * @return the client connection to the specified host and port
     * @throws NamingException if a connection cannot be established
     */
    private Client openConnection() throws NamingException {
        Client client;
        try {
            client = new Client(_host, _port);
        } catch (Exception exception) {
            NamingException error = new NamingException(
                "Failed to connect to JNDI provider");
            error.setRootCause(exception);
            throw error;
        }
        return client;
    }

    /**
     * Returns the tcp url using the old environment properties, logging a
     * warning. This method will be removed in future releases
     */
    private String getDeprecatedEnvironment(Hashtable environment) {
        _log.warn(getClass().getName() +
            ": using deprecated environment. Use Context.PROVIDER_URL");

        StringBuffer url = new StringBuffer("tcp://");
        String host = (String) environment.get(JndiConstants.HOST_PROPERTY);
        Integer port = (Integer) environment.get(
            JndiConstants.PORT_NUMBER_PROPERTY);

        // if the host is specified then use it, otherwise use localhost
        if (host != null) {
            url.append(host);
            url.append(":");
        } else {
            url.append("localhost:");
        }

        // if the port has been specified then use it otherwise use
        // the default port
        if (port != null) {
            url.append(port.toString());
        } else {
            url.append("3035");
        }

        return url.toString();
    }

} //-- IpcJndiInitialContextFactory
