/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: BaseLease.java,v 1.6 2003/08/17 01:32:24 tanderson Exp $
 *
 * Date         Author  Changes
 * 2/29/2000    jima    Created
 */
package org.exolab.jms.lease;


import java.io.Serializable;


/**
 * This is the implementation of a non-specific lease object. It can be used to
 * lease any Objeect. .
 *
 * @version     $Revision: 1.6 $ $Date: 2003/08/17 01:32:24 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 **/
public class BaseLease
    implements LeaseIfc, Comparable, Serializable {

    /**
     * Default constructor for the Serailizable/Externalizable interface
     */
    public BaseLease() {
    }

    /**
     * Construct an instance of this class with the specified objct, and duration
     *
     * @param       object          leased object
     * @param       duration        duration of lease
     * @param       listener        will get notified on lease events
     */
    public BaseLease(Object object, long duration, LeaseEventListenerIfc listener) {

        // test for preconditions
        if ((object == null) ||
            (listener == null)) {
            throw new IllegalArgumentException("object or listener cannot be null");
        }

        leasedObject_ = object;
        duration_ = duration;
        expiryTime_ = System.currentTimeMillis() + duration;
        listener_ = listener;
    }

    // implementation for LeaseIfc.getExpiryTime()
    public long getExpiryTime() {
        return expiryTime_;
    }

    // implementation for LeaseIfc.getDuration()
    public long getDuration() {
        return duration_;
    }

    // implementation for LeaseIfc.getDuration()
    public void setDuration(long duration) {
        duration_ = duration;
        expiryTime_ = System.currentTimeMillis() + duration;
    }

    // implementation for LeaseIfc.getRemainingTime()
    public long getRemainingTime() {
        return (System.currentTimeMillis() - expiryTime_);
    }

    // implementation for LeaseIfc.getLeasedObject()
    public Object getLeasedObject() {
        return leasedObject_;
    }

    // implementation for LeaseIfc.getLeasedObjectType()
    public Class getLeasedObjectType() {
        return leasedObject_.getClass();
    }

    /**
     * Return a reference to the listener registered with this object
     *
     * @return LeaseEventListenerIfc
     */
    public LeaseEventListenerIfc getLeaseEventListener() {
        return listener_;
    }

    /**
     * Compares this object with the specified object. It returns a negative
     * integer is this object is less than the specified object; zero if this
     * object is equal to the specified object or a positive integer if this
     * object is greater than the specified object
     * <p>
     * The comparison is based on the expiration time.
     */
    public int compareTo(Object object) {
        int result = 0;

        if (object instanceof BaseLease) {
            BaseLease lease = (BaseLease) object;
            if (lease.getExpiryTime() != this.getExpiryTime()) {
                if (lease.getExpiryTime() > this.getExpiryTime()) {
                    result = -1;
                } else {
                    result = 1;
                }
            }
        }
        return result;
    }

    /**
     * Notify the listeners that this lease has expird. This method has
     * package level scope.
     */
    void notifyLeaseExpired() {
        synchronized (listener_) {
            listener_.onLeaseExpired(leasedObject_);
        }
    }

    /**
     * Return a string representation of this object
     *
     * @return      String
     */
    public String toString() {
        StringBuffer buf = new StringBuffer(leasedObject_.toString());
        buf.append(" duration = ");
        buf.append(duration_);
        buf.append(" expiryTime = ");
        buf.append(expiryTime_);

        return buf.toString();
    }


    /**
     * This is the object that is leased
     */
    protected Object leasedObject_ = null;

    /**
     * The duration of the lease in milliseconds
     */
    protected long duration_ = 0L;

    /**
     * This is the time that the lease will expire
     */
    protected long expiryTime_ = 0L;

    /**
     * The listener that will be notified when the lease expires
     */
    protected LeaseEventListenerIfc listener_ = null;
}

