/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2001,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: LeaseManager.java,v 1.22 2003/10/29 04:58:32 tanderson Exp $
 *
 * Date         Author  Changes
 * 3/01/2000    jima    Created
 */
package org.exolab.jms.lease;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.service.BasicService;
import org.exolab.core.service.ServiceException;
import org.exolab.core.service.ServiceState;
import org.exolab.core.util.OrderedQueue;
import org.exolab.jms.message.MessageHandle;


/**
 * The LeaseManager is responsible for creating and managing the lease objects.
 * The Leasemanager is a singleton. When a BaseLease object is created it is
 * added to the queue according to the duration (i.e. leases with shorter
 * durations are placed at the top of the queue.
 * <p>
 * When the lease expires the LeeaseManager calls the leasee's associated
 * listener(s).
 * <p>
 * We need to cater for persistence and non-persistent leases
 *
 * @version     $Revision: 1.22 $ $Date: 2003/10/29 04:58:32 $
 * @author      <a href="mailto:jima@intalio.com">Jim Alateras</a>
 */
public class LeaseManager extends BasicService {

    /**
     * The name of the LeaseManagerThread that scans and removes expired leases
     */
    private static final String LEASE_MANAGER_THREAD_NAME =
        "LeaseManagerReaper";

    /**
     * An ordered list of leases
     */
    private OrderedQueue _queue = null;

    /**
     * The singleton instance of the LeaseManager
     */
    private static LeaseManager _instance = new LeaseManager();

    /**
     * Helper for waiting for leases to expire
     */
    private final Object _waiter = new Object();

    /**
     * The logger
     */
    private static final Log _log = LogFactory.getLog(LeaseManager.class);


    /**
     * Create a new sorted tree set using the lease comparator as the
     * sorting functor.
     */
    protected LeaseManager() {
        super(LEASE_MANAGER_THREAD_NAME);
        _queue = new OrderedQueue(new LeaseComparator());
    }

    /**
     * Return the singleton instance of the LeaseManager
     *
     * @return LeaseManager
     */
    public static LeaseManager instance() {
        return _instance;
    }

    /**
     * Create a message lease with the specified duration and the nominated
     * listener. The listener does not to be specified but is merely a
     * convenience argument.
     * <p>
     * The object must be non-null and the duration must be greater than zero.
     *
     * @param       handle          message handle
     * @param       duration        lease duration
     * @param       listener        object that implements the LeaseEvent-
     *                              ListenerIfc
     * @return      MessageLease    creared lease object or null if no lease
     *                              is created
     */
    public MessageLease createMessageLease(MessageHandle handle,
                                           long duration,
                                           LeaseEventListenerIfc listener) {
        MessageLease result = null;

        if ((handle != null) && (duration > 0)) {
            result = new MessageLease(handle, duration, listener);
            addLease(result);
        }
        
        return result;
    }

    /**
     * Create a lease on any object for the specified duration. If a listener
     * is also provided then add it to the lease. Insert the lease in the queue
     * and return an instance of the created lease.
     * <p>
     * The object must be non-null and the duration must be greater than zero.
     *
     * @param       object          leased object
     * @param       duration        lease duration
     * @param       listener        object that implements the LeaseEvent-
     *                              ListenerIfc
     * @return      BaseLease       creared lease object or null if no lease
     *                              is created
     */
    public BaseLease createLease(Object object, long duration,
                                 LeaseEventListenerIfc listener) {
        BaseLease result = null;

        if ((object != null) && (duration > 0)) {
            result = new BaseLease(object, duration, listener);
            addLease(result);
        }

        return result;
    }

    /**
     * Remove the specified lease.
     *
     * @param       lease           lease to remove
     * @return      boolean         true if successful; false otherwise
     */
    public boolean removeLease(BaseLease lease) {
        boolean result = false;

        synchronized (_queue) {
            result = _queue.remove(lease);
        }
        if (result) {
            synchronized (_waiter) {
                _waiter.notify();
            }
        }

        return result;
    }

    /**
     * Renew the lease on the specified object
     *
     * @param object                object's whose lease we want to renew
     * @param duration              the new duration of the lease in ms
     */
    public BaseLease renewLease(BaseLease lease, long duration) {
        BaseLease newlease = null;

        if ((lease != null) && (duration > 0)) {
            synchronized (_queue) {
                // check that the lease hasn't expired yet.
                if (_queue.remove(lease)) {
                    lease.setDuration(duration);
                    _queue.add(lease);
                    newlease = lease;
                    synchronized (_waiter) {
                        _waiter.notify();
                    }
                }
            }
        }

        return newlease;
    }

    /**
     * Remove all the leases from the queue. Do not expire any of them
     */
    public void removeAll() {
        synchronized (_queue) {
            _queue.clear();
        }
    }

    /**
     * Returns the number of active leases
     *
     * @return the number of leases
     */
    public int getLeaseCount() {
        return _queue.size();
    }

    /**
     * The run method will search for expired leases, remove them from the
     * list and notify listeners
     */
    public void run() {
        while (getState() != ServiceState.STOPPED) {
            expire();

            // wait until a lease is available, or the service is terminated
            synchronized (_waiter) {
                try {
                    _waiter.wait();
                } catch (InterruptedException ignore) {
                }
            }
        }
    }

    /**
     * Stop the service.
     *
     * @throws ServiceException if the service fails to stop
     */
    public void stop() throws ServiceException {
        super.stop();
        synchronized (_waiter) {
            _waiter.notifyAll();
        }
    }

    /**
     * Expires active leases
     */
    protected void expire() {
        while (_queue.size() > 0) {
            BaseLease lease = null;
            boolean expired = false;
            synchronized (_queue) {
                lease = (BaseLease) _queue.firstElement();
                if (lease == null) {
                    continue;
                }
                
                if (lease.getExpiryTime() <= System.currentTimeMillis()) {
                    // remove from the list and notify listeners
                    _queue.removeFirstElement();
                    expired = true;
                }
            }

            if (expired) {
                lease.notifyLeaseExpired();
            } else {
                // wait until the first element in the list is
                // ready to expire
                long time = lease.getExpiryTime() - 
                    System.currentTimeMillis();
                
                if (time > 0) {
                    try {
                        synchronized (_waiter) {
                            _waiter.wait(time);
                        }
                    } catch (InterruptedException ignore) {
                    }
                }
            }
        }
    }

    /**
     * Add a lease
     *
     * @param lease the lease to add
     */
    protected void addLease(BaseLease lease) {
        synchronized (_queue) {
            _queue.add(lease);
            if (_queue.firstElement() == lease) {
                // inserted before the first element, so reset scan
                synchronized (_waiter) {
                    _waiter.notify();
                }
            }
        }
    }


} //-- LeaseManager
