/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ConsumerEndpoint.java,v 1.36 2003/09/25 11:23:13 tanderson Exp $
 *
 * Date         Author  Changes
 * 3/1/2001     jima    Created
 */
package org.exolab.jms.messagemgr;

import java.io.Serializable;
import java.sql.Connection;
import java.util.Enumeration;
import java.util.Vector;

import javax.jms.InvalidSelectorException;
import javax.jms.JMSException;
import javax.jms.Session;
import javax.transaction.TransactionManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.jms.Identifiable;
import org.exolab.jms.client.JmsDestination;
import org.exolab.jms.message.MessageHandle;
import org.exolab.jms.message.MessageImpl;
import org.exolab.jms.persistence.PersistenceException;
import org.exolab.jms.scheduler.Scheduler;
import org.exolab.jms.selector.Selector;
import org.exolab.jms.server.JmsServerSession;
import org.exolab.jms.util.UUID;


/**
 * A Consumer is a message subscriber with a unique identity
 *
 * @version     $Revision: 1.36 $ $Date: 2003/09/25 11:23:13 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @author      <a href="mailto:tma@netspace.net.au">Tim Anderson</a>
 */
abstract public class ConsumerEndpoint
    implements Serializable, Identifiable, DestinationCacheEventListener,
    Runnable {

    /**
     * The identity of the consumer
     */
    private String _id;

    /**
     * The client identity, which uniquely identifies the remote client
     * within a session. This is used to tag messages when they are
     * asynchronously delivered to the client
     */
    private long _clientId = -1;

    /**
     * The selector assoicated with this consumer. A selector is used
     * to filter messages.
     */
    protected Selector _selector = null;

    /**
     * Serializes access to the {@link #_waitingForMessage} flag. This is
     * only required when it is changed
     */
    protected final Object _waitingForMessageMonitor = new Object();

    /**
     * Used to block consumer until there is a message available
     */
    protected boolean _waitingForMessage = false;;

    /**
     * Holds the consumer's message listener. This means that messages
     * will be pushed down
     */
    protected InternalMessageListener _listener = null;

    /**
     * Maintains the maximum size of this cache
     */
    protected int _size = 1000;

    /**
     * This is the scheduler that is used to deliver messages if a consumer
     * has a registered listener
     */
    protected transient Scheduler _scheduler = null;

    /**
     * The acknowledgement mode for this endpoint.
     */
    protected transient int _ackMode = Session.AUTO_ACKNOWLEDGE;

    /**
     * The nolocal indicator, if set, inhibits consuming messages that have
     * been published on the same connection
     */
    protected transient boolean _nolocal = false;

    /**
     * Indicates whether this endpoint belongs to a transacted session
     */
    protected transient boolean _transacted = false;

    /**
     * Holds the connection id of the connection that the endpoint belongs too
     */
    protected transient int _connectionId = -1;

    /**
     * caches the session that created this endpoint
     */
    protected JmsServerSession _session = null;

    /**
     * This determines whether message delivery to the registered listener
     * is enabled or disabled.
     */
    private volatile boolean _stopped = true;

    /**
     * Identifies this endpoint as being closed
     */
    private volatile boolean _closed = false;

    /**
     * Indicates whether the this cache has been scheduled with the dispatcher
     * for asynchronous message delivery.
     */
    private boolean _scheduled = false;

    /**
     * Cache of all messages and handles for this consumer.
     */
    private MessageCache _cache = new MessageCache();

    /**
     * Synchronization helper for close() and deliverMessages()
     */
    private final Object _lock = new Object();

    /**
     * The logger
     */
    private static final Log _log = LogFactory.getLog(ConsumerEndpoint.class);


    /**
     * Construct a <code>ConsumerEndpoint</code>.</p>
     * The destination and selector determine where it will be sourcing
     * its messages from, and scheduler is used to asynchronously deliver
     * messages to the consumer.
     *
     * @param session - the owning session
     * @param clientId - uniquely identifies the remote client within session
     * @param selector - the selector attached to the consumer, if any.
     * @param scheduler - used to schedule async message delivery.
     * @throws InvalidSelectorException if the selector is not well formed
     */
    ConsumerEndpoint(JmsServerSession session, long clientId,
                     String selector, Scheduler scheduler)
        throws InvalidSelectorException {
        _id = UUID.next();
        _selector = (selector != null) ? new Selector(selector) : null;
        _clientId = clientId;
        _scheduler = scheduler;
        _session = session;
    }

    /**
     * Return the destination that this consumer is subscribed to
     *
     * @return the destination that this consumer is subscribed to
     */
    public abstract JmsDestination getDestination();

    // implementation of Identifiable.getId
    public String getId() {
        return _id;
    }

    /**
     * Returns the persistent identifier for this consumer.<p/>
     * This is the identity of the consumer which is persistent across
     * subscriptions and server restarts. <p/>
     *
     * This implementation simply returns the transient identifier, i.e,
     * {@link #getId}
     *
     * @return the persistent identifier for this consumer
     */
    public String getPersistentId() {
        return getId();
    }

    // implementation of Object.hashCode
    public int hashCode() {
        return _id.hashCode();
    }

    /**
     * Return a stringified version of the consumer
     *
     * @return String
     */
    public String toString() {
        return _id + ":" + getDestination();
    }

    /**
     * Unregister this consumer for the specified destination cache, so that it
     * will stop receiving messages from it.
     */
    public abstract void unregister();

    /**
     * Return a reference to the client identity. This is an indirect reference
     * back to the remote client, which can asynchronously receive messages
     *
     * @return identity of the client scoped to the session
     */
    public long getClientId() {
        return _clientId;
    }

    /**
     * Return the selector for this endpoint or null if one is not specified
     *
     * @return String - the endpoint's selector
     */
    public Selector getSelector() {
        return _selector;
    }

    /**
     * Set the selector for this endpoint
     *
     * @param selector - message selector as a string
     * @exception InvalidSelectorException - raised when selector is not
     *                                       well-formed
     */
    public void setSelector(String selector)
        throws InvalidSelectorException {
        _selector = (selector != null) ? new Selector(selector) : null;
    }

    /**
     * Return the ackmode for this endpoint
     *
     * @return int - acknowledgement mode
     */
    public int getAckMode() {
        return _ackMode;
    }

    /**
     * Set the ackMode for this endpoint
     *
     * @param ackmode - the new ack mode for the endpoint
     */
    public void setAckMode(int ackmode) {
        _ackMode = ackmode;
    }

    /**
     * Return the connection id that this endpoint belongs to
     *
     * @return the connection id
     */
    public int getConnectionId() {
        return _connectionId;
    }

    /**
     * Set the connection that this endpooint belongs too
     *
     * @param id - connection identity
     */
    public void setConnectionId(int id) {
        _connectionId = id;
    }

    /**
     * Return the value of the nolocal indicator
     *
     * @return boolean
     */
    public boolean getNoLocal() {
        return _nolocal;
    }

    /**
     * Set the no local indicator for this endpoint
     *
     * @param nolocal - true to inhibit messages published on this connection
     */
    public void setNoLocal(boolean nolocal) {
        _nolocal = nolocal;
    }

    /**
     * Check whether this endpoint belongs to a transacted session
     *
     * @return boolean - true if it does
     */
    public boolean getTransacted() {
        return _transacted;
    }

    /**
     * Set the state if the transacted flag for this endpoint
     *
     * @param transacted - true if it is transacted
     */
    public void setTransacted(boolean transacted) {
        _transacted = transacted;
    }

    /**
     * Set the maximum size of the cache. If there are more than this number
     * of messages in the cache the {@link CacheEvictionPolicy} is enforced
     * to remove messages.
     *
     * @param size - maximum number of messages a destination can hold
     */
    public void setMaximumSize(int size) {
        _size = size;
    }

    /**
     * Return the cache's maximum size
     *
     * @return int - size of cache
     */
    public int getMaximumSize() {
        return _size;
    }

    /**
     * Set the {@link CacheEvictionPolicy} for this object. This determines
     * how messages are removed when the cache's upper limit is reached.
     *
     * @param policy the eviction policy
     */
    public void setCacheEvictionPolicy(CacheEvictionPolicy policy) {
        // not implemented
    }

    /**
     * Return the number of unsent messages in the cache for this consumer
     *
     * @return the number of unsent messages
     */
    public int getMessageCount() {
        return _cache.getHandleCount();
    }

    /**
     * Return a reference to the session owning this endpoint
     *
     * @return JmsServerSession - the owning session
     */
    public JmsServerSession getSession() {
        return _session;
    }
   
    /**
     * Deliver messages in the cache to the consumer
     *
     * @return <code>true</code> if the endpoint should be rescheduled
     */
    public abstract boolean deliverMessages();

    /**
     * The run method is used to asynchronously deliver the messages in the
     * cache to the consumer, by invoking {@link #deliverMessages}. 
     * <p>
     * It is scheduled by the {@link Scheduler}.
     */
    public void run() {
        synchronized (_lock) {
            if (!_closed) {
                boolean reschedule = deliverMessages();
                _scheduled = false;
                if (reschedule) {
                    schedule();
                }
            }
        }
    }

    // implementation of DestinationCacheEventListener.messageAdded
    public synchronized boolean messageAdded(MessageImpl message) {
        boolean added = false;

        // create a message handle
        try {
            // if the nolocal indicator is set and the message arrived on
            // the same connection, add this consumer then mark it as
            // received, but do not add it to the queue
            if (getNoLocal()
                && message.getConnectionId() == getConnectionId()) {
                // inform them that we have processed the message
                return true;
            }

            MessageHandle handle =
                MessageHandleFactory.getHandle(this, message);

            if (!_cache.containsHandle(handle)) {
                // if the message is not already in the cache then add it
                // and flag that we have added the message to the cache
                addMessage(handle, message);
                added = true;

                schedule();
            }
        } catch (JMSException exception) {
            _log.error("Failed to add message to endpoint", exception);
        }

        return added;
    }

    // implementation of DestinationCacheEventListener.messageRemoved
    public synchronized boolean messageRemoved(MessageImpl message) {
        boolean removed = false;

        try {
            //retrieve the message handle
            MessageHandle handle = 
                MessageHandleFactory.getHandle(this, message);

            if (_cache.containsHandle(handle)) {
                // call remove regardless whether it exists
                removeMessage(handle);
                removed = true;
            }
        } catch (JMSException exception) {
            _log.error("Failed to remove message from endpoint", exception);
        }

        return removed;
    }

    // implementation of DestinationCacheEventListener.persistentMessageAdded
    public synchronized boolean persistentMessageAdded(Connection connection,
                                                       MessageImpl message)
        throws PersistenceException {

        return messageAdded(message);
    }

    // implementation of DestinationCacheEventListener.persistentMessageRemoved
    public synchronized boolean persistentMessageRemoved(Connection connection,
                                                         MessageImpl message)
        throws PersistenceException {

        return messageRemoved(message);
    }

    /**
     * Stop/start message delivery
     *
     * @param stop if <code>true</code> to stop message delivery, otherwise
     * start it
     */
    public synchronized void setStopped(boolean stop) {
        if (stop) {
            _stopped = true;
        } else {
            _stopped = false;
            // schedule message delivery if needed
            schedule();
        }
    }

    /**
     * This message will return all unacked messages to the queue and allow
     * them to be resent to the consumer with the redelivery flag on.
     */
    public synchronized void recover() {
        // default behaviour is to do nothing
    }

    /**
     * Close this endpoint.
     * <p>
     * This synchronizes with {@link #deliverMessages} before invoking
     * @link {#doClose}
     */
    public final void close() {
        _stopped = true;

        synchronized (_lock) {
            // synchronize with deliverMessages()            
            _scheduler.remove(this); // remove this, if it is scheduled
            _scheduled = false;

        }

        synchronized (this) {
            doClose();

            // clear all messages in the cache
            if (_cache != null) {
                _cache.clear();
            }

            _closed = true;
        }
    }

    /**
     * Set the message listener for this consmer. If a message listener is set
     * then messages will be scheduled to be sent to it when they are available
     * <p>
     * Each consumer cache can only have a single message listener. To remove
     * the message listener call this method with null argument
     *
     * @param listener - the message listener to add.
     */
    public synchronized void setMessageListener(
        InternalMessageListener listener) {
        _listener = listener;
        if (listener == null) {
            // remove this from the scheduler
            _scheduler.remove(this);
            _scheduled = false;
        } else {
            // scheduler for it to run
            schedule();
        }
    }

    /**
     * Return the specified message to the cache.
     *
     * @param handle - handle to return
     */
    public synchronized void returnMessage(MessageHandle handle) {
        if (_cache != null) {
            addMessage(handle);
            schedule();
        }
    }

    /**
     * Return the next message to the client. This will also mark the message as
     * sent and move it to the sent queue
     *
     * @param wait - the number of milliseconds to wait
     * @return MessageHandle - handle to the next message in the list
     */
    abstract public MessageHandle receiveMessage(long wait);

    // implementation of GarbageCollectable.collectGarbage
    public void collectGarbage(boolean aggressive) {
        if (aggressive) {
            // clear all persistent messages in the cache
            _cache.clearPersistentMessages();
            if (_log.isDebugEnabled()) {
                _log.debug("Evicted all persistent messages from dest "
                    + getDestination().getName() + " and name "
                    + getId());
            }
        }

        if (_log.isDebugEnabled()) {
            _log.debug("ENDPOINT- " + getDestination().getName() + ":"
                + getPersistentId() + " Messages: P["
                + _cache.getPersistentCount() + "] T["
                + _cache.getTransientCount() + "] Handles: ["
                + _cache.getHandleCount() + "]");
        }
    }

    /**
     * Closes the endpoint
     */
    protected abstract void doClose();

    /**
     * Schedule asynchronouse message delivery
     */
    protected void schedule() {
        if (!_stopped && !_closed && _listener != null && !_scheduled) {
            _scheduled = true;
            _scheduler.add(this);
        }
    }

    /**
     * Clear all messages in the cache, regardless of whether they are
     * persistent or non-persistent
     */
    protected void clearMessages() {
        _cache.clear();
    }

    /**
     * Check whether the vector of handles contains one or more persistent
     * handles
     *
     * @param handles - collection of {@link MessageHandle} objects
     * @return true if there is one or more persistent handles
     */
    protected boolean collectionHasPersistentHandles(Vector collection) {
        boolean result = false;
        Enumeration enum = collection.elements();

        while (enum.hasMoreElements()) {
            if (enum.nextElement() instanceof PersistentMessageHandle) {
                result = true;
                break;
            }
        }

        return result;
    }

    /**
     * Add the handle to the cache
     *
     * @param handle - the message handle to add
     */
    protected void addMessage(MessageHandle handle) {
        handle.setConsumerName(getPersistentId());
        _cache.addHandle(handle);

        // check to see whether the consumer is waiting to
        // be notified
        if (isWaitingForMessage()) {
            notifyMessageAvailable();
        }
    }

    /**
     * Cache a handle and its corresponding message
     *
     * @param handle the handle to cache
     * @param message the corresponding message to cache
     */
    protected void addMessage(MessageHandle handle, MessageImpl message) {
        handle.setConsumerName(getPersistentId());
        _cache.addMessage(handle, message);

        // check to see whether the consumer is waiting to
        // be notified
        if (isWaitingForMessage()) {
            notifyMessageAvailable();
        }
    }

    /**
     * Return the message for the specified handle
     *
     * @param handle - the handle
     * @return MessageImpl - the associated message
     */
    protected MessageImpl getMessage(MessageHandle handle) {
        return _cache.getMessage(handle);
    }

    /**
     * Remove the handle from the cache
     *
     * @param handle the handle to remove
     * @return <code>true</code> if the message was removed
     */
    protected boolean removeMessage(MessageHandle handle) {
        return _cache.removeHandle(handle);
    }

    /**
     * Determines if a message handle is already cached
     *
     * @return <code>true</code> if it is cached
     */
    protected boolean containsMessage(MessageHandle handle) {
        return _cache.containsHandle(handle);
    }

    /**
     * Return the first message handle in the cache
     *
     * @return the first message or null if cache is empty
     */
    protected MessageHandle removeFirstMessage() {
        return _cache.removeFirstHandle();
    }

    /**
     * Delete the message with the specified handle from the cache
     *
     * @param handle the handle
     */
    protected void deleteMessage(MessageHandle handle) {
        _cache.removeMessage(handle.getMessageId());
    }

    /**
     * Determines if this endpoint has been stopped
     *
     * @return <code>true</code> if this endpoint has been stopped
     */
    protected final boolean isStopped() {
        return _stopped;
    }

    /**
     * Check if the consumer is waiting for a message. If it is then
     * notify it that a message has arrived
     */
    protected void notifyMessageAvailable() {
        // if we need to notify then send out the request
        if (isWaitingForMessage()) {
            clearWaitingForMessage();

            try {
                _session.onMessageAvailable(getClientId());
            } catch (Exception exception) {
                //getLogger().logError("Error in notifyMessageAvailable " +
                //    getDestination().getName() + " " + exception.toString());
            }
        }
    }

    /**
     * Check whether the endpoint is waiting for a message
     *
     * @return boolean
     */
    protected final boolean isWaitingForMessage() {
        return _waitingForMessage;
    }

    /**
     * Set the waiting for message flag
     */
    protected final void setWaitingForMessage() {
        _waitingForMessage = true;
    }

    /**
     * Clear the waiting for message flag
     */
    protected final void clearWaitingForMessage() {
        _waitingForMessage = false;
    }

    /**
     * Helper for {@link #deliverMessages} implementations, to determines if 
     * asynchronous message delivery should be stopped
     *
     * @return <code>true</code> if asynchronous message delivery should be 
     * stopped
     */
    protected boolean stopDelivery() {
        return (_stopped || getMessageCount() == 0 || _listener == null);
    }

}
