/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: DestinationCache.java,v 1.34 2003/12/29 13:09:15 tanderson Exp $
 *
 * Date         Author  Changes
 * 3/1/2001     jima    Created
 */
package org.exolab.jms.messagemgr;

import java.sql.Connection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.transaction.TransactionManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.jms.Identifiable;
import org.exolab.jms.client.JmsDestination;
import org.exolab.jms.gc.GarbageCollectable;
import org.exolab.jms.lease.LeaseEventListenerIfc;
import org.exolab.jms.message.MessageHandle;
import org.exolab.jms.message.MessageImpl;
import org.exolab.jms.persistence.DatabaseService;
import org.exolab.jms.persistence.PersistenceException;
import org.exolab.jms.persistence.SQLHelper;
import org.exolab.jms.util.UUID;


/**
 * A DestinationCache is used to cache messages for a particular destination. 
 * <p>
 * It implements {@link MessageManagerEventListener} in order to be notified
 * of messages being added to and removed from 
 * {@link MessageMgr}
 * <p>
 * A {@link ConsumerEndpoint} registers with a {@link DestinationCache} to
 * receive messages for a particular destination.
 * <p>
 * In all instances this class doesn't deal with <code>Message</code> objects
 * directly, but instead uses their corresponding {@link MessageHandle},
 * which is far more lightweight.
 * <p>
 * A two level cache is used to facilitate quick seeding of registered
 * consumers and quick per-consumer-acknowledgment strategy. The two level
 * cache includes this, the DestinationCache, at the highest level and then
 * {@link ConsumerManager} at the lowest level.
 * <p>
 * In addition to registering {@link ConsumerEndpoint} objects the cache also
 * supports {@link DestinationCacheEventListener}s. A listener will be 
 * notified when messages are added to the cache but do not actually consume 
 * messages. A cache can have one or more registered listeners. This feature is
 * predominately used by browsers or iterators of destinations.
 * <p>
 * Clients can also become lifecycle listeners for this object to get notified
 * during initialization and shutdwon.
 * <p>
 * This cache is ordered on priority.
 *
 * @version     $Revision: 1.34 $ $Date: 2003/12/29 13:09:15 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 */
public abstract class DestinationCache
    implements MessageManagerEventListener, Identifiable,
    LeaseEventListenerIfc, GarbageCollectable {

    /**
     * The identity of this object
     */
    private String _id;

    /**
     * Create a message cache for this destination
     */
    private MessageCache _cache = new MessageCache();

    /**
     * Maintain the maximum size of this cache. When the cache reaches this
     * size then messages will be lost. We should only drop transient messages
     *  and maintain persistent messages. Remove transient messages from the
     * top of the cache and replace them with persistent messages.
     */
    private int _maximumSize = Integer.MAX_VALUE;

    /**
     * This is the list of consumers that have subscribed to this cache. It
     * hosts both durable and transient subscribers
     */
    protected List _consumers =
        Collections.synchronizedList(new LinkedList());

    /**
     * The message lease helper is used to manage leases for messages
     * cached by the destination
     */
    protected MessageLeaseHelper _leaseHelper = null;

    /**
     * The logger
     */
    private static final Log _log = LogFactory.getLog(DestinationCache.class);

    /**
     * Construct a message cache for a particular destination.
     */
    DestinationCache() {
        _id = UUID.next();
    }

    /**
     * Register this destination with the message manager. If it cannot
     * initialise then throw FailedToInitializeException
     */
    void init() throws FailedToInitializeException {
        // moved to after lease helper creation
        //MessageMgr.instance().addEventListener(getDestination(), this);

        try {
            _leaseHelper = new MessageLeaseHelper(this);
        } catch (PersistenceException exception) {
            String msg = "Failed to initialise destination cache";
            _log.error(msg, exception);
            throw new FailedToInitializeException(
                msg + ":" + exception.getMessage());
        }

        MessageMgr.instance().addEventListener(getDestination(), this);
    }

    /**
     * Register this destination with the message manager and create the
     * lease helper. The {@link MessageLeaseHelper} is initialized using the
     * specified {@link Connection}.
     * <p>
     * If there are problems with the initialization then throw
     * FailedToInitializeException
     *
     * @param connection - the connection to use
     * @throws FailedToInitializeException
     */
    void init(Connection connection)
        throws FailedToInitializeException {
        MessageMgr.instance().addEventListener(getDestination(), this);

        try {
            _leaseHelper = new MessageLeaseHelper(connection, this);
        } catch (Exception exception) {
            // rethrow
            throw new FailedToInitializeException("Error initialising " +
                exception.toString());
        }
    }

    /**
     * Set the maximum size of the cache. If there are more than this number
     * of messages in the cache the {@link CacheEvictionPolicy} is enforced
     * to remove messages.
     *
     * @param size - maximum number of messages a destination can hold
     */
    public void setMaximumSize(int size) {
        _maximumSize = size;
    }

    /**
     * Return the cache's maximum size
     *
     * @return int - size of cache
     */
    public int getMaximumSize() {
        return _maximumSize;
    }

    /**
     * Return a reference to the underlying destination
     *
     * @return JmsDestination
     */
    abstract public JmsDestination getDestination();

    /**
     * Return the string form of the destination
     *
     * @return  String
     */
    public String getDestinationByName() {
        return getDestination().getName();
    }

    /**
     * Set the {@link CacheEvictionPolicy} for this object. This determines
     * how messages are removed when the cache's upper limit is reached.
     *
     * @param policy the eviction policy
     */
    public void setCacheEvictionPolicy(CacheEvictionPolicy policy) {
        // not implemented
    }

    /**
     * Register a consumer with this cache. Part of the registration process
     * will be to seed the {@link ConsumerEndpoint} with an initial list of
     * messages that it needs to consume and then feed messages to it through
     * the specified listener object.
     * <p>
     * Messages are subsequently passed down to the consumer's through the
     * listener, as they enter the DestinationCache.
     *
     * @param consumer - message consumer for this destination
     * @return boolean - true if registered and false otherwise
     */
    public boolean registerConsumer(ConsumerEndpoint consumer) {

        boolean result = false;

        // check to see that the consumer is actually one for this
        // destination
        if (consumer.getDestination().equals(getDestination())) {
            if (!_consumers.contains(consumer)) {
                _consumers.add(consumer);
                consumer.setMaximumSize(this.getMaximumSize());
                result = true;
            }
        }

        return result;
    }

    /**
     * Remove the consumer for the list of registered consumers. If the
     * consumer does not exist then the call fails silently.
     *
     * @param consumer - consumer to remove.
     */
    public void unregisterConsumer(ConsumerEndpoint consumer) {
        if (_consumers.contains(consumer)) {
            _consumers.remove(consumer);
        } else {
        }
    }

    /**
     * Return an enmeration of all consumers attached to this cache.
     *
     * @return Iterator - list of registered consumers
     */
    public Iterator getConsumers() {
        return _consumers.iterator();
    }

    /**
     * Return the consumers as an array of {@link ConsumerEndpoint} objects
     * This is a safer way to get a list of consumers since it avoids
     * concurrent modification exceptions
     *
     * @return Object[] - an array of ConsumerEndpoint objects
     */
    Object[] getConsumersByArray() {
        return _consumers.toArray();
    }

    /**
     * This method is called when the {@link MessageMgr} adds a message
     * for this destination to the cache
     *
     * @param message - message added to cache
     */
    abstract public boolean messageAdded(JmsDestination destination,
                                         MessageImpl message);

    /**
     * This method is called when the {@link MessageMgr} removes a
     * message from the cache.
     *
     * @param message - message removed from cache
     */
    abstract public void messageRemoved(JmsDestination destination,
                                        MessageImpl message);

    /**
     * Return the number of messages currently active for this destination
     *
     * @return int - number of active messages
     */
    public int getMessageCount() {
        return _cache.getHandleCount();
    }

    /**
     * Notify the listeners that a non-persistent message has been added to the
     * cache
     *
     * @param handle - message that was added
     * @return boolean - true of at least one listener has processed itx
     */
    abstract boolean notifyOnAddMessage(MessageImpl message);

    /**
     * Notify the listeners that a non-persistent message has been removed form
     * the cache
     *
     * @param handle - message that was removed
     */
    abstract void notifyOnRemoveMessage(MessageImpl message);

    /**
     * Notify the listeners that a persistent message has been added to the
     * cache
     *
     * @param connection - the persistent connection to use
     * @param handle - message that was added
     * @return boolean - true of at least one listener has processed it
     * @throws PersistenceException - if there is a persistence related error
     */
    boolean notifyOnAddPersistentMessage(Connection connection,
                                         MessageImpl message)
        throws PersistenceException {

        //default implementation
        return true;
    }

    /**
     * Notify the listeners that a persistent message has been removed form
     * the cache
     *
     * @param connection - the persistent connection to use.
     * @param handle - message that was removed
     * @throws PersistenceException - if there is a persistence related error
     */
    void notifyOnRemovePersistentMessage(Connection connection,
                                         MessageImpl message)
        throws PersistenceException {
        //default implementation is empty
    }

    /**
     * Check whether there are any attached consumers to this cache
     *
     * @return boolean - true if there are attached consumers
     */
    abstract boolean hasActiveConsumers();

    /**
     * This method is called whenever a lease expires. It passes the
     * object that has expired.
     *
     * @param       leasedObject        reference to the leased object
     */
    public void onLeaseExpired(Object leasedObject) {
        if (leasedObject != null) {
            MessageHandle handle = (MessageHandle) leasedObject;

            // retrieve an instance of the message
            MessageImpl message = resolveExpiredMessage(handle);

            // determine whether  the message is persistent or not and take
            // the corresponding action
            if (handle instanceof PersistentMessageHandle) {
                Connection connection = null;
                try {
                    connection = DatabaseService.getConnection();
                    persistentMessageRemoved(connection, getDestination(),
                        message);
                    connection.commit();
                } catch (Exception exception) {
                    SQLHelper.rollback(connection);
                    _log.error("Failure in onLeaseExpired", exception);
                } finally {
                    SQLHelper.close(connection);
                }
            } else {
                // notify it's listeners that the non-persistent message has
                // been removed
                messageRemoved(getDestination(), message);
            }
        }
    }

    /**
     * Determines if this cache can be destroyed
     *
     * @return <code>true</code> if the cache can be destroyed, otherwise
     * <code>false</code>
     */
    public abstract boolean canDestroy();

    /**
     * Check to see if the message has a TTL. If so then set up a lease
     * for it. An expiry time of 0 means that the message never expires
     *
     * @param message - message to add
     */
    void checkMessageExpiry(MessageImpl message) {
        if (message != null) {
            _leaseHelper.addLease(message);
        }
    }

    /**
     * Destory this cache.
     */
    synchronized void destroy() {
        // clear the cache
        _cache.clear();

        // remove the consumers
        _consumers.clear();

        // unregister itself from the message manager
        MessageMgr.instance().removeEventListener(getDestination(), this);

        // remove the lease
        _leaseHelper.clear();
    }

    /**
     * Close the cache and unregister all the consumers. Notify any and all
     * DestinationCacheLifecycleListeners.
     * <p>
     * Once the cache is closed it will no longger receive messages for this
     * destination.
     */
    public void shutdown() {
        destroy();
    }

    /**
     * Insert the specified handle to the handles cache.
     *
     * @param handle - handle to add
     */
    void addMessage(MessageHandle handle) {
        handle.setConsumerName(getDestination().getName());
        _cache.addHandle(handle);
    }

    /**
     * Add the following handle and corresponding message to their respective
     * caches
     *
     * @param handle - handle to add
     * @param message - the corresponding message to add
     */
    void addMessage(MessageHandle handle, MessageImpl message) {
        handle.setConsumerName(getDestination().getName());
        _cache.addMessage(handle, message);
    }

    /**
     * Return the message for the specified handle
     *
     * @param handle - the handle
     * @return MessageImpl - the associated message
     */
    MessageImpl getMessage(MessageHandle handle) {
        return _cache.getMessage(handle);
    }

    /**
     * Remove the message handle from the cache, if it exists.
     *
     * @param handle - handle to remove
     * @return boolean - true if it was removed
     */
    boolean removeMessage(MessageHandle handle) {
        return _cache.removeHandle(handle);
    }

    /**
     * Remove and return the first message handle in the cache
     *
     * @return MessageHandle - the first handle or null if cache is empty
     */
    final MessageHandle removeFirstMessage() {
        return _cache.removeFirstHandle();
    }

    /**
     * Return the message handles in the cache as an array
     *
     * @return Object[] - array of message handles
     */
    final Object[] toMessageArray() {
        return _cache.getHandleArray();
    }

    /**
     * Delete the message with the specified handle from the cache
     *
     * @param handle - the handle
     */
    void deleteMessage(MessageHandle handle) {
        _cache.removeMessage(handle.getMessageId());
    }

    // implementation of Identifiable.getId
    public String getId() {
        return _id;
    }

    // implementation of GarbageCollectable.collectGarbage
    public void collectGarbage(boolean aggressive) {
        if (aggressive) {
            // clear all persistent messages in the cache
            _cache.clearPersistentMessages();
            if (_log.isDebugEnabled()) {
                _log.debug("Evicted all persistent messages from cache "
                    + getDestination().getName());
            }
        }

        if (_log.isDebugEnabled()) {
            _log.debug("DESTCACHE -" + getDestination().getName()
                + " Messages: P[" + _cache.getPersistentCount()
                + "] T[" + _cache.getTransientCount() + "] Handles: ["
                + _cache.getHandleCount() + "]");
        }
    }

    /**
     * Resolve an expired message through its handle
     *
     * @param handle the expired message's handle
     * @return the expired message. May be null.
     */
    protected MessageImpl resolveExpiredMessage(MessageHandle handle) {
        return handle.getMessage();
    }

}
