/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: MessageHandleFactory.java,v 1.12 2003/08/17 01:32:24 tanderson Exp $
 *
 * Date         Author  Changes
 * 3/1/2001     jima    Created
 */
package org.exolab.jms.messagemgr;

import java.sql.Connection;

import javax.jms.DeliveryMode;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.transaction.TransactionManager;

import org.exolab.jms.client.JmsDestination;
import org.exolab.jms.client.JmsQueue;
import org.exolab.jms.message.MessageHandle;
import org.exolab.jms.message.MessageImpl;
import org.exolab.jms.persistence.DatabaseService;
import org.exolab.jms.persistence.PersistenceException;


/**
 * This class defines a number of static methods that are responsible
 * for creatimg message handles
 *
 * @version     $Revision: 1.12 $ $Date: 2003/08/17 01:32:24 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 */
public class MessageHandleFactory {

    /**
     * Retrieve a {@link TransientMessageHandle} given the specified message,
     * irrespective of whether it is a persistent or non-persistent message.
     *
     * @param message the message
     * @return the message handle
     * @throws JMSException if the handle can't be created
     */
    public static MessageHandle getHandle(MessageImpl message)
        throws JMSException {

        MessageHandle handle = null;

        if (message.isPersistent()) {
            handle = new PersistentMessageHandle(message);
        } else {
            handle = new TransientMessageHandle(message);
        }

        return handle;
    }

    /**
     * Retrieve a {@link org.exolab.jms.message.MessageHandle} from the 
     * specified QueueConsumerEndpoint and message. It will create a 
     * PersistentMessageHandle if the message is persistent or a 
     * TransientMessageHandle otherwise.
     *
     * @param queue the queue destination cache
     * @param message the message
     * @return the message handle
     * @throws JMSException if the handle can't be created
     */
    public static MessageHandle getHandle(QueueDestinationCache queue,
                                          MessageImpl message)
        throws JMSException {

        MessageHandle handle = null;
        if (message.getJMSDeliveryMode() == DeliveryMode.PERSISTENT) {
            PersistentMessageHandle phandle =
                new PersistentMessageHandle(message);
            // set the consumer and destination identity
            phandle.setConsumerName(queue.getDestination().getName());
            phandle.setDestination(queue.getDestination());

            handle = phandle;
        } else {
            handle = new TransientMessageHandle(message);
        }

        return handle;
    }

    /**
     * Use the message, destination and optionally, the consumer name to
     * retrieve a message handle.
     *
     * @param destination  the message destination
     * @param name the consumer name, can be null for JmsQueue type
     * @param message the message
     * @return the message handle
     * @throws JMSException if the handle can't be created
     */
    public static MessageHandle getHandle(JmsDestination dest, String name,
                                          MessageImpl message)
        throws JMSException {

        MessageHandle handle = null;
        if (message.getJMSDeliveryMode() == DeliveryMode.PERSISTENT) {
            PersistentMessageHandle phandle =
                new PersistentMessageHandle(message);
            // set the consumer and destination identity
            phandle.setDestination(dest);
            if (dest instanceof JmsQueue) {
                phandle.setConsumerName(dest.getName());
            } else {
                phandle.setConsumerName(name);
            }
            handle = phandle;
        } else {
            handle = new TransientMessageHandle(message);
        }
        return handle;
    }

    /**
     * Retrieve a message handle for the specified {@link ConsumerEndpoint}
     * and message.
     *
     * @param consumer the consumer endpoint
     * @param message the message
     * @return the message handle
     * @throws JMSException if the handle can't be created
     */
    public static MessageHandle getHandle(ConsumerEndpoint consumer,
                                          MessageImpl message)
        throws JMSException {

        MessageHandle handle = null;
        if (message.getJMSDeliveryMode() == DeliveryMode.PERSISTENT) {
            if (consumer instanceof DurableConsumerEndpoint) {
                PersistentMessageHandle phandle =
                    new PersistentMessageHandle(message);
                DurableConsumerEndpoint durable =
                    (DurableConsumerEndpoint) consumer;

                // set the consumer and destination identity
                phandle.setConsumerName(durable.getName());
                phandle.setDestination(
                    (JmsDestination) message.getJMSDestination());

                handle = phandle;
            } else {
                handle = new TransientMessageHandle(message);
            }
        } else {
            handle = new TransientMessageHandle(message);
        }
        return handle;
    }


    /**
     * Create a transient handle for a message belonging to a queue.
     *
     * @param queue the queue destination cache
     * @param message the message
     * @return the message handle
     * @throws JMSException if the handle can't be created
     */
    public static MessageHandle createHandle(QueueDestinationCache queue,
                                             MessageImpl message)
        throws JMSException {

        return new TransientMessageHandle(message);
    }

    /**
     * Create a transient message handle for a durable consumer
     *
     * @param durable the durable consumer endpoint
     * @param message the message
     * @return the message handle
     * @throws JMSException if the handle can't be created
     */
    public static MessageHandle createHandle(DurableConsumerEndpoint durable,
                                             MessageImpl message)
        throws JMSException {

        return new TransientMessageHandle(message);
    }

    /**
     * Create a transient message handle.
     *
     * @param destination the message destination
     * @param name the consumer name, can be null for JmsQueue type
     * @param message the message
     * @return the message handle
     * @throws JMSException if the handle can't be created
     */
    public static MessageHandle createHandle(JmsDestination dest, String name,
                                             MessageImpl message)
        throws JMSException {

        return new TransientMessageHandle(message);
    }

    /**
     * Create a {@link PersistentMessageHandle} from the specified
     * {@link TransientMessageHandle}. It does not fill in all the fields of a
     * persistent message handle.
     *
     * @param handle  the transient message handle
     * @return the persistent message handle
     * @throws JMSException if the handle can't be created
     */
    public static MessageHandle createPersistentHandle(
        TransientMessageHandle handle) throws JMSException {

        PersistentMessageHandle phandle = new PersistentMessageHandle();
        phandle.setMessageId(handle.getMessageId());
        phandle.setPriority(handle.getPriority());
        phandle.setDelivered(handle.getDelivered());
        phandle.setDestination(handle.getDestination());
        phandle.setAcceptedTime(handle.getAcceptedTime());
        phandle.setExpiryTime(handle.getExpiryTime());
        phandle.setSequenceNumber(handle.getSequenceNumber());
        phandle.setClientId(handle.getClientId());

        return phandle;
    }

    /**
     * Create a persistent message handle belonging to a queue.
     *
     * @param connection connection to the database
     * @param queue queue destination cache
     * @param message the message
     * @return the message handle
     * @throws JMSException if the handle can't be created
     * @throws PersistenceException if the handle cannot be made persistent
     */
    public static MessageHandle createPersistentHandle(
        Connection connection, QueueDestinationCache queue, 
        MessageImpl message) throws JMSException, PersistenceException {

        PersistentMessageHandle handle = new PersistentMessageHandle(message);

        // fill in the persistent related particulars
        handle.setConsumerName(queue.getDestination().getName());
        handle.setDestination(queue.getDestination());

        // make the handle persistent
        DatabaseService.getAdapter().addMessageHandle(connection, handle);

        return handle;
    }

    /**
     * Create a persistent handle for a message belonging to a particular
     * durable consumer
     *
     * @param connection the connection to use.
     * @param durable durable consumer endpoint
     * @param message the message
     * @return the message handle
     * @throws JMSException if the handle can't be created
     * @throws PersistenceException if the handle cannot be made persistent
     */
    public static MessageHandle createPersistentHandle(
        Connection connection, DurableConsumerEndpoint durable, 
        MessageImpl message) throws JMSException, PersistenceException {

        PersistentMessageHandle handle = new PersistentMessageHandle(message);

        // set the consumer and destination identity
        handle.setConsumerName(durable.getName());
        handle.setDestination(durable.getDestination());

        DatabaseService.getAdapter().addMessageHandle(connection, handle);

        return handle;
    }

    /**
     * Create a persistent message handle from the destination and consumer
     * name
     *
     * @param connection the connection to use
     * @param destination the message destination
     * @param name the consumer name, can be null for JmsQueue type
     * @param message the message
     * @return the message handle
     * @throws JMSException if the handle can't be created
     * @throws PersistenceException if the handle cannot be made persistent
     */
    public static MessageHandle createPersistentHandle(Connection connection, 
                                                       JmsDestination dest, 
                                                       String name, 
                                                       MessageImpl message) 
        throws JMSException, PersistenceException {

        PersistentMessageHandle handle = new PersistentMessageHandle(message);

        // set the consumer and destination identity
        handle.setDestination(dest);
        if (dest instanceof JmsQueue) {
            handle.setConsumerName(dest.getName());
        } else {
            handle.setConsumerName(name);
        }

        DatabaseService.getAdapter().addMessageHandle(connection, handle);
        return handle;
    }

    /**
     * Destroy the specified persistent handle.
     *
     * @param connection the connection to use
     * @param handle the persistent handle to destroy
     * @throws PersistenceExcetpion if there is a persistence-related error
     */
    public static void destroyPersistentHandle(Connection connection,
                                               PersistentMessageHandle handle)
        throws PersistenceException {
        DatabaseService.getAdapter().removeMessageHandle(connection, handle);

        // notify the message manager
        MessageMgr.instance().handleDestroyed(handle);
    }

    /**
     * Destroy a persistent handle associated with a destination, consumer
     * name and message
     *
     * @param connection the connection to use
     * @param destination the destination assoicated with the message
     * @param name the name of the consumer
     * @param message the message
     * @throws JMSException if the handle can't be created
     * @throws PersistentException if there is a persistence related error
     */
    public static void destroyPersistentHandle(Connection connection,
                                               JmsDestination destination,
                                               String name,
                                               MessageImpl message)
        throws JMSException, PersistenceException {

        if (message.isPersistent()) {
            PersistentMessageHandle handle =
                (PersistentMessageHandle) getHandle(destination, name,
                    message);
            destroyPersistentHandle(connection, handle);
        }
    }

    /**
     * Update the state of the persistent handle
     *
     * @param connection the connection to use
     * @param handle the handle to update
     * @throws PersistentException if there is a persistence related error
     */
    public static void updatePersistentHandle(Connection connection,
                                              PersistentMessageHandle handle)
        throws PersistenceException {
        DatabaseService.getAdapter().updateMessageHandle(connection, handle);
    }

} //-- MessageHandleFactory

