/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2004 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: MessageQueue.java,v 1.32.2.1 2004/05/01 12:01:15 tanderson Exp $
 */
package org.exolab.jms.messagemgr;

import java.util.Comparator;
import java.util.Iterator;
import java.util.TreeSet;


/**
 * The message queue stored messages based on a comparator. The implementation
 * is based on a synchronized linked list.
 * <p>
 * We can easily improve on this implementation and this implementation is
 * not synchronized.
 *
 * @version     $Revision: 1.32.2.1 $ $Date: 2004/05/01 12:01:15 $
 * @author      <a href="mailto:jima@intalio.com">Jim Alateras</a>
 * @author      <a href="mailto:tma@netspace.net.au">Tim Anderson</a>
 */
public class MessageQueue {

    /**
     * The message queue is implemented as a linked list
     */
    private TreeSet _list = null;

    /**
     * Instantiate an instance of this class with the specified comparator
     * the comparator is used to order the elements in the queue. Elements
     * with the same order value are placed after each other.
     *
     * @param       comparator          used for ordering
     */
    public MessageQueue(Comparator comparator) {
        _list = new TreeSet(comparator);
    }

    /**
     * Add this element to the queue in the required order. It uses a
     * binary search to locate the correct position
     *
     * @param       object              object to add
     */
    public synchronized void add(Object object) {
        _list.add(object);
    }

    /**
     * Check if the specified object exists
     *
     * @param object - object to check
     * @return boolean - true if it is contained
     */
    public synchronized boolean contains(Object object) {
        return _list.contains(object);
    }

    /**
     * Check if the queue is empty
     *
     * @return boolean - true if the queue is empty
     */
    public synchronized boolean isEmpty() {
        return _list.isEmpty();
    }

    /**
     * Return all elements in the collection
     *
     * @return Object[] array of elements to return
     */
    public synchronized Object[] toArray() {
        return _list.toArray();
    }


    /**
     * Return an iterator to the list
     *
     * @return Iterator
     */
    public synchronized Iterator iterator() {
        return _list.iterator();
    }

    /**
     * Remove the object from the queue
     *
     * @param       object          object to remove
     */
    public synchronized boolean remove(Object object) {
        return _list.remove(object);
    }

    /**
     * Remove all the elements from the queue
     */
    public synchronized void clear() {
        _list.clear();
    }

    /**
     * Return the number elements in the queue
     *
     * @return      int         size of the queue
     */
    public synchronized int size() {
        return _list.size();
    }

    /**
     * Returns the first element in the queue.
     *
     * @return the first element in the queue, or <code>null</code>, if the
     * queue is empty
     */
    public synchronized Object first() {
        return (_list.size() > 0) ? _list.first() : null;
    }

    /**
     * Returns the last element in the queue.
     *
     * @return the last element in the queue, or <code>null</code>, if the
     * queue is empty
     */
    public synchronized Object last() {
        return (_list.size() > 0) ? _list.last() : null;
    }

    /**
     * Removes and returns the first element on the queue.
     *
     * @return the first element in the queue, or <code>null</code>, if the
     * queue is empty
     */
    public synchronized Object removeFirst() {
        Object object = first();
        if (object != null) {
            _list.remove(object);
        }
        return object;
    }

    /**
     * Removes and returns the last element in the queue
     *
     * @return the last element in the queue, or <code>null</code>, if the
     * queue is empty
     */
    public synchronized Object removeLast() {
        Object object = last();
        if (object != null) {
            _list.remove(object);
        }
        return object;
    }

} //-- MessageQueue
