/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PersistentMessageHandle.java,v 1.17 2003/08/07 13:33:04 tanderson Exp $
 *
 * Date         Author  Changes
 * 3/1/2001     jima    Created
 */
package org.exolab.jms.messagemgr;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.sql.Connection;

import javax.jms.DeliveryMode;
import javax.jms.JMSException;
import javax.transaction.TransactionManager;

import org.exolab.core.foundation.PersistentObject;
import org.exolab.jms.client.JmsDestination;
import org.exolab.jms.client.JmsQueue;
import org.exolab.jms.client.JmsTopic;
import org.exolab.jms.message.MessageHandle;
import org.exolab.jms.message.MessageId;
import org.exolab.jms.message.MessageImpl;
import org.exolab.jms.persistence.DatabaseService;
import org.exolab.jms.persistence.PersistenceException;


/**
 * A persistent message handle extends {@link MessageHandle} and references a
 * persistent message. These messages can be discarded from the cache and later
 * faulted in through the 'resolve' method.
 *
 * @version     $Revision: 1.17 $ $Date: 2003/08/07 13:33:04 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 **/
public class PersistentMessageHandle
    extends PersistentObject
    implements MessageHandle, Cloneable {

    /**
     * Used for serialization
     */
    static final long serialVersionUID = 1;

    /**
     * This is the message id of the underlying message
     */
    private MessageId _id = null;

    /**
     * This is the priority of the inderlying message
     */
    private int _priority;

    /**
     * A flag indicating whether an attempt was made to deliver the message
     */
    private boolean _delivered;

    /**
     * a transient attribute, to associate this handle with an endpoint
     */
    private transient long _clientId = -1;

    /*
     * The time that the underlying message was accepted by the server
     */
    private long _acceptedTime;

    /**
     * This is the sequence number assigned to it by the message manager
     * that created this handle
     */
    private long _sequenceNumber;

    /**
     * The expiry time of this message. 0 if it doesn't expire
     */
    private long _expiryTime;

    /**
     * The name of the durable consumer that this handle belongs too
     */
    private String _consumerName;

    /**
     * The destination that this handle belongs too
     */
    private JmsDestination _destination;


    /**
     * Cache the hashcode for this object
     */
    private transient int _hashCode = -1;


    /**
     * Default constructor
     */
    public PersistentMessageHandle() {
        super();
    }

    /**
     * Create a persistent handle from a message
     *
     * @param message - persistent message
     * @exception JMSException - if the object cannot be constructed
     */
    public PersistentMessageHandle(MessageImpl message)
        throws JMSException {
        super();

        if ((message != null) &&
            (message.getJMSDeliveryMode() == DeliveryMode.PERSISTENT)) {
            _id = message.getMessageId();
            _priority = message.getJMSPriority();
            _delivered = message.getJMSRedelivered();
            _destination = (JmsDestination) message.getJMSDestination();
            _acceptedTime = message.getAcceptedTime();
            _expiryTime = message.getJMSExpiration();
            _sequenceNumber = message.getSequenceNumber();
            _clientId = message.getClientId();
        } else {
            throw new JMSException(
                "Cannot create persistent handle from non-persistent message");
        }
    }

    /**
     * Set the identity of the underlying message
     *
     * @param id - the message id
     */
    public void setMessageId(MessageId id) {
        _id = id;
    }

    /**
     * Return the identity of the underlying message
     *
     * @return MessageId
     */
    public MessageId getMessageId() {
        return _id;
    }

    /**
     * Return the associated message or null if it is invalid
     *
     * @return  MessageImpl
     */
    public MessageImpl getMessage() {
        return (MessageImpl) MessageMgr.instance().getMessage(this);
    }

    /**
     * Set the messagee priority
     *
     * @param id the persistent id
     */
    public void setPriority(int priority) {
        _priority = priority;
    }

    /**
     * Return the priority of the corresponding message
     *
     * @return int
     */
    public int getPriority() {
        return _priority;
    }

    /**
     * Set the delivered flag for the underlying message. True if an attempt
     * has been made to deliver the message
     *
     * @param value - true if delivered
     */
    public void setDelivered(boolean value) {
        _delivered = value;
    }

    /**
     * Return the delivered state of the message
     *
     * @return long
     */
    public boolean getDelivered() {
        return _delivered;
    }

    /**
     * Set the time that the message was accepted by the server
     *
     * @param long - time since epoch
     */
    public void setAcceptedTime(long time) {
        _acceptedTime = time;
    }

    /**
     * Return the time that the message was accepted by the server
     *
     * @return long - time in ms since epoch
     */
    public long getAcceptedTime() {
        return _acceptedTime;
    }

    /**
     * Set the message's sequence number
     *
     * @param seq - sequence time
     */
    public void setSequenceNumber(long seq) {
        _sequenceNumber = seq;
    }

    /**
     * Return message's sequence number
     *
     * @return long - the sequence number
     */
    public long getSequenceNumber() {
        return _sequenceNumber;
    }

    /**
     * Set the message expiry time
     *
     * @long time - time in ms since epoch
     */
    public void setExpiryTime(long time) {
        _expiryTime = time;
    }

    /**
     * Return the message expiry time
     *
     * @return long - time in ms since epoch
     */
    public long getExpiryTime() {
        return _expiryTime;
    }

    /**
     * Determines if the message has expired
     *
     * @return <code>true</code> if the message has expired, otherwise
     * <code>false</code>
     */
    public boolean hasExpired() {
        return (_expiryTime != 0 && _expiryTime <= System.currentTimeMillis());
    }

    // implementation of MessageHandle.setConsumerName
    public void setConsumerName(String name) {
        _consumerName = name;
    }

    // implementation of MessageHandle.getConsumerName
    public String getConsumerName() {
        return _consumerName;
    }

    // implementation of MessageHandle.isPersistent
    public boolean isPersistent() {
        return true;
    }

    /**
     * Set the destination that owns this handle
     *
     * @param destination - the destination
     */
    public void setDestination(JmsDestination destination) {
        _destination = destination;
    }

    /**
     * Return the destination for this handle
     *
     * @return JmsDestination
     */
    public JmsDestination getDestination() {
        return _destination;
    }

    /**
     * Set the client id, that owns this handle
     *
     * @param clientId - client identity
     */
    public void setClientId(long clientId) {
        _clientId = clientId;
    }

    /**
     * Retrieve the client identity associated with this handle
     *
     * @return long
     */
    public long getClientId() {
        return _clientId;
    }

    /**
     * Override the behaviour of the base class. I don't think we want to
     * clear the handle unless there are definitely no more references.
     */
    public void clear() {
    }

    /**
     * Destroy this handle
     */
    public void destroy() {
        // notify the message manager
        MessageMgr.instance().handleDestroyed(this);

        // remove the handle from the database
        Connection connection = null;
        TransactionManager tm = null;

        try {
            connection = DatabaseService.getConnection();
            DatabaseService.getAdapter().removeMessageHandle(connection, this);
            connection.commit();
        } catch (PersistenceException exception) {
            if (connection != null) {
                try {
                    connection.rollback();
                } catch (Exception nested) {
                    // ignore
                }
            }
        } catch (Exception exception) {
            // ignore for the moment
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (Exception nested) {
                    // ignore
                }
            }
        }
    }

    /**
     * Clone the persistent message handle object
     *
     * @return copy of a PersistentMessageHandle
     * @throws CloneNotSupportedException
     */
    public Object clone() throws CloneNotSupportedException {
        PersistentMessageHandle handle =
            (PersistentMessageHandle) super.clone();

        handle._id = new MessageId(_id.getId());
        handle._priority = _priority;
        handle._delivered = _delivered;
        handle._acceptedTime = _acceptedTime;
        handle._expiryTime = _expiryTime;
        handle._consumerName = _consumerName;
        handle._sequenceNumber = _sequenceNumber;

        // clone the destination
        if (_destination instanceof JmsQueue) {
            handle._destination = new JmsQueue(_destination.getName());
        } else {
            handle._destination = new JmsTopic(_destination.getName());
        }
        handle._destination.setPersistent(true);

        return handle;
    }

    // override the definition of equals
    public boolean equals(Object object) {

        boolean result = false;

        if ((object != null) &&
            (object instanceof PersistentMessageHandle)) {

            PersistentMessageHandle mhdl = (PersistentMessageHandle) object;
            if (mhdl._id.equals(_id)) {
                result = true;
            }

            //if ((mhdl._id.equals(_id)) &&
            //    (mhdl._consumerName.equals(_consumerName))) {
            //    result = true;
            //}
        }

        return result;
    }

    // override the Object.toString()
    public String toString() {
        return ("Handle : " + _id.getId() + "@" + _consumerName + ":" +
            _sequenceNumber + ":" + _acceptedTime + ":" + _priority);
    }


    // override the definition of hashCode
    public int hashCode() {
        if (_hashCode == -1) {
            _hashCode = (_id + _consumerName).hashCode();
        }

        return _hashCode;
    }

    /**
     * This is used to set the message as delivered. This can only be called
     * by the delivery engine and should not be used anywhere else. This will
     * mark the message as being delivered and also update the database
     * <p>
     * We do not really throw an exception here....and it seems that this is
     * a very fine grained approach to dealing with this...basically translates to
     * a transaction for every message.
     */
    public void setDelivered() {
        if (!_delivered) {

            Connection connection = null;
            TransactionManager tm = null;

            try {
                connection = DatabaseService.getConnection();

                DatabaseService.getAdapter().updateMessageHandle(connection,
                    this);
                _delivered = true;

                // commit the transactions
                connection.commit();
            } catch (PersistenceException exception) {
                System.err.println("Failed in setDelivereds b/c " +
                    exception.toString());
                if (connection != null) {
                    try {
                        connection.rollback();
                    } catch (Exception nested) {
                        // ignore
                    }
                }
            } catch (Exception exception) {
                // ignore for the moment
            } finally {
                if (connection != null) {
                    try {
                        connection.close();
                    } catch (Exception nested) {
                        // ignore
                    }
                }
            }
        }
    }

    // implementation of Externalizable.writeExternal
    public void writeExternal(ObjectOutput stream)
        throws IOException {
        stream.writeLong(serialVersionUID);
        stream.writeObject(_id);
        stream.writeInt(_priority);
        stream.writeBoolean(_delivered);
        stream.writeLong(_acceptedTime);
        stream.writeLong(_sequenceNumber);
        stream.writeLong(_expiryTime);
        stream.writeObject(_consumerName);
        stream.writeObject(_destination);
        super.writeExternal(stream);
    }

    // implementation of Externalizable.writeExternal
    public void readExternal(ObjectInput stream)
        throws IOException, ClassNotFoundException {
        long version = stream.readLong();
        if (version == serialVersionUID) {
            _id = (MessageId) stream.readObject();
            _priority = stream.readInt();
            _delivered = stream.readBoolean();
            _acceptedTime = stream.readLong();
            _sequenceNumber = stream.readLong();
            _expiryTime = stream.readLong();
            _consumerName = (String) stream.readObject();
            _destination = (JmsDestination) stream.readObject();
            super.readExternal(stream);
        } else {
            throw new IOException("PersistentMessageHandle with version " +
                version + " is not supported.");
        }
    }
}

