/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2001,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * Date         Author  Changes
 * 04/07/2000   jima    Created
 */
package org.exolab.jms.server;

import java.sql.Connection;
import java.util.Enumeration;
import java.util.Vector;

import javax.jms.JMSException;
import javax.transaction.TransactionManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.service.ServiceManager;
import org.exolab.jms.authentication.AuthenticationMgr;
import org.exolab.jms.authentication.User;
import org.exolab.jms.client.JmsDestination;
import org.exolab.jms.client.JmsQueue;
import org.exolab.jms.client.JmsTopic;
import org.exolab.jms.messagemgr.ConsumerEndpoint;
import org.exolab.jms.messagemgr.ConsumerManager;
import org.exolab.jms.messagemgr.DestinationCache;
import org.exolab.jms.messagemgr.DestinationManager;
import org.exolab.jms.persistence.DatabaseService;
import org.exolab.jms.persistence.SQLHelper;
import org.exolab.jms.util.UUID;


/**
 * A connection is created for every adminclient connecting to the JmsServer.
 *
 * @version     $Revision: 1.2.2.1 $ $Date: 2004/05/04 01:26:49 $
 * @author      <a href="mailto:knut@lerpold.no">Knut Lerpold</a>
 * @see         org.exolab.jms.server.AdminConnectionManager
 */
public class AdminConnection {

    /**
     * This is the client that is responsible for this connection. It should
     * map back to a client-side entity
     */
    private String _clientId = null;

    /**
     * The connection id is unique within the context of a JMS server
     */
    private String _identifierId = null;

    /**
     * The logger
     */
    private static final Log _log =
        LogFactory.getLog(AdminConnection.class);

    /**
     * Instantiate an instance of this class for the specified client identity.
     * Generates a unique identifier for this connection.
     *
     * @param       id              client identity
     */
    AdminConnection(String id) {
        _clientId = id;
        _identifierId = UUID.next();
    }

    /**
     * Returns the client identifier
     *
     * @return String  the client identifier
     */
    public String getClientId() {
        return _clientId;
    }

    /**
     * Returns the identifier
     *
     * @return String   the identifierId
     */
    public String getIdentifierId() {
        return _identifierId;
    }

    /**
     * Close the admin connection
     */
    public void close() {
    }

    /**
     * Determine whether the client is active
     * Could e.g. disconnect the client if it has been inactive for a
     * period of time.
     *
     * for now always return true
     *
     * @return boolean true if active; false otherwise
     */
    boolean isClientActive() {
        boolean active = true;

        return active;
    }

    /**
     * Return the number of messages for a durable consumer.
     *
     * @param       topic               name of the topic
     * @param       name                consumer name
     * @return      int                 number of unsent or unacked messages
     */
    public int getDurableConsumerMessageCount(String topic, String name) {
        int count = -1;
        Connection connection = null;

        try {
            // first see if the cache is loaded in memory
            DestinationManager dmgr = DestinationManager.instance();
            ConsumerManager cmgr = ConsumerManager.instance();
            JmsDestination dest = dmgr.destinationFromString(topic);
            ConsumerEndpoint endpoint = null;
            if ((dest != null)
                && ((name != null)
                || (name.length() > 0))) {

                endpoint = cmgr.getConsumerEndpoint(name);
                if ((endpoint != null)
                    && (endpoint.getDestination().equals(dest))) {
                    // retrieve the number of handles for the endpoint, which
                    // reflects the number of messages
                    count = endpoint.getMessageCount();
                } else {
                    // there is no cache with this name stored in memory. If
                    // this is an administered destination then read the count
                    //  directly from the database.
                    if (dmgr.isAdministeredDestination(topic)) {
                        connection = DatabaseService.getConnection();
                        count = DatabaseService.getAdapter().
                            getDurableConsumerMessageCount(connection, topic, name);

                        connection.commit();
                    }
                }
            }
        } catch (Exception exception) {
            _log.error("Failed to get message count for topic=" + topic,
                exception);
            SQLHelper.rollback(connection);
        } finally {
            SQLHelper.close(connection);
        }

        return count;
    }

    /**
     * First use the destination manager to return the number of persistent and
     * non-persistent messages in a queue
     *
     * @param queue name of the queue
     * @return int - the number of messages for that destination or -1 if the
     *               destination is invalid
     */
    public int getQueueMessageCount(String queue) {
        int count = -1;
        Connection connection = null;

        try {
            // first see if the cache is loaded in memory
            DestinationManager mgr = DestinationManager.instance();
            JmsDestination dest = mgr.destinationFromString(queue);
            DestinationCache cache = null;
            if (dest != null) {
                cache = mgr.getDestinationCache(dest);

                if (cache != null) {
                    // retrieve the number of handles for the cache, which
                    // reflects the number of messages
                    count = cache.getMessageCount();
                } else {
                    // there is no cache with this name stored in memory. If
                    // this is an administered destination then read the count
                    //  directly from the database.
                    if (mgr.isAdministeredDestination(queue)) {
                        connection = DatabaseService.getConnection();
                        count = DatabaseService.getAdapter().
                            getQueueMessageCount(connection, queue);
                        connection.commit();
                    }
                }
            }
        } catch (Exception exception) {
            _log.error("Failed to get message count for queue=" + queue,
                exception);
            SQLHelper.rollback(connection);
        } finally {
            SQLHelper.close(connection);
        }

        return count;
    }

    /**
     * Add the specified durable consumer to the database
     *
     * @param       topic               name of the destination
     * @param       name                name of the consumer
     * @return      boolean             true if successful
     */
    public boolean addDurableConsumer(String topic, String name) {
        boolean result = false;
        try {
            ConsumerManager.instance().createDurableConsumer(
                new JmsTopic(topic), name);
            result = true;
        } catch (JMSException exception) {
            _log.error("Failed to add durable consumer=" + name
                + " for topic=" + topic, exception);
        }

        return result;
    }

    /**
     * Remove the consumer attached to the specified destination and with
     * the passed in name
     *
     * @param       name                name of the consumer
     * @return      boolean             true if successful
     */
    public boolean removeDurableConsumer(String name) {
        boolean result = false;
        try {
            ConsumerManager.instance().removeDurableConsumer(name);
            result = true;
        } catch (JMSException exception) {
            _log.error("Failed to remove durable consumer=" + name, exception);
        }

        return result;
    }

    /**
     * Check if the durable consumer exists.
     *
     * @param       name                name of the durable conusmer
     * @return      boolean             true if it exists and false otherwise
     */
    public boolean durableConsumerExists(String name) {
        return ConsumerManager.instance().durableConsumerExists(name);
    }

    /**
     * Return the collection of durable consumer names for a particular
     * topic destination.
     *
     * @param       topic               the topic name
     * @return      Vector              collection of strings
     */
    public Vector getDurableConsumers(String topic) {
        Enumeration iter = null;
        Vector result = new Vector();
        Connection connection = null;

        try {
            connection = DatabaseService.getConnection();

            iter = DatabaseService.getAdapter().getDurableConsumers(
                connection, topic);
            // copy the elements into the vector
            while (iter.hasMoreElements()) {
                result.addElement(iter.nextElement());
            }
            connection.commit();
        } catch (Exception exception) {
            _log.error("Failed on get durable consumers for topic=" + topic,
                exception);
            SQLHelper.rollback(connection);
        } finally {
            SQLHelper.close(connection);
        }

        return result;
    }

    /**
     * De-Activate an active persistent consumer.
     *
     * @param       name                name of the consumer
     * @return      boolean             true if successful
     */
    public boolean unregisterConsumer(String name) {
        boolean success = false;

        try {
            ConsumerManager.instance().deleteDurableConsumerEndpoint(name);
            success = true;
        } catch (JMSException exception) {
            //
        }

        return success;
    }

    /**
     * Check to see if the given consumer is currently connected to the
     * OpenJMSServer. This is only valid when in online mode.
     *
     * @param name The name of the onsumer.
     * @return boolean True if the consumer is connected.
     */
    public boolean isConnected(String name) {
        return ConsumerManager.instance().isDurableConsumerActive(name);
    }

    /**
     * Return a list of all registered destinations.
     *
     * @return      Vector     collection of strings
     */
    public Vector getAllDestinations() {
        Enumeration iter = null;
        Vector result = new Vector();
        Connection connection = null;
        TransactionManager tm = null;

        try {
            connection = DatabaseService.getConnection();

            iter = DatabaseService.getAdapter().getAllDestinations(connection);
            // copy the elements into the vector
            while (iter.hasMoreElements()) {
                result.addElement(iter.nextElement());
            }
            connection.commit();
        } catch (Exception exception) {
            _log.error("Failed to get all destinations", exception);
            SQLHelper.rollback(connection);
        } finally {
            SQLHelper.close(connection);
        }

        return result;
    }

    /**
     * Add an administered destination with the specified name
     *
     * @param       name                destination name
     * @param       queue               whether it is queue or a topic
     * @return      boolean             true if successful
     */
    public boolean addDestination(String name, Boolean queue) {

        boolean success = false;

        // create the appropriate destination object
        JmsDestination destination = (queue.booleanValue())
            ? (JmsDestination) new JmsQueue(name)
            : (JmsDestination) new JmsTopic(name);
        destination.setPersistent(true);

        // create the administered destination
        try {
            success = DestinationManager.instance().
                createAdministeredDestination(destination);
        } catch (JMSException exception) {
            _log.error("Failed to add destination=" + name, exception);
        }

        return success;
    }

    /**
     * Destroy the specified destination and all associated messsages and
     * consumers. This is a very dangerous operation to execute while there
     * are clients online
     *
     * @param       name         destination to destroy
     * @return      boolean             true if successful
     */
    public boolean removeDestination(String name) {

        boolean success = false;
        JmsDestination dest =
            DestinationManager.instance().destinationFromString(name);

        // ensure that the destination actually translates to a valid
        // object.
        if (dest != null) {
            try {
                DestinationManager.instance().deleteAdministeredDestination(dest);
                success = true;
            } catch (JMSException exception) {
                _log.error("Failed to remove destination=" + name, exception);
            }
        }

        return success;
    }

    /**
     * Check whether the specified destination exists
     *
     * @param  name - the name of the destination to check
     * @return boolean - true if it does and false otherwise
     */
    public boolean destinationExists(String name) {

        boolean exists = false;
        DestinationManager mgr = DestinationManager.instance();
        JmsDestination dest = mgr.destinationFromString(name);

        if (dest != null) {
            exists = mgr.destinationExists(dest);
        }

        return exists;
    }

    /**
     * Terminate the JMS Server. If it is running as a standalone application
     * then exit the application. It is running as an embedded application then
     * just terminate the thread
     */
    public void stopServer() {
        // todo perform neccesary clean up for a much nicer shutdown.
        // probably should inform all clients.

        _log.info("Stopping all services");
        ServiceManager.instance().removeAll();
        _log.info("Server Shutdown scheduled for 5 secs");
        Runnable r = new Runnable() {

            public void run() {
                try {
                    Thread.sleep(5000);
                    System.exit(0);
                } catch (Exception err) {
                    //
                }
            }
        };
        Thread t = new Thread(r);
        t.start();
    }

    /**
     * Purge all processed messages from the database
     *
     * @return      int         number of messages purged
     * @deprecated  no replacement
     */
    public int purgeMessages() {
        return 0;
    }

    /**
     * Add a user with the specified name
     *
     * @param username    the users name
     * @param password    the users password
     * @return <code>true</code> if the user is added
     * otherwise <code>false</code>
     */
    public boolean addUser(String username, String password) {
        return AuthenticationMgr.instance().addUser(
            new User(username, password));
    }

    /**
     * Change password for the specified user
     *
     * @param username    the users name
     * @param password    the users password
     * @return <code>true</code> if the password is changed
     * otherwise <code>false</code>
     */
    public boolean changePassword(String username, String password) {
        return AuthenticationMgr.instance().updateUser(
            new User(username, password));
    }

    /**
     * Remove the specified user
     *
     * @param username    the users name
     * @return <code>true</code> if the user is removed
     * otherwise <code>false</code>
     */
    public boolean removeUser(String username) {
        return AuthenticationMgr.instance().removeUser(
            new User(username, null));
    }

    /**
     * Return a list of all registered users.
     *
     * @return Vector of users
     */
    public Vector getAllUsers() {
        Enumeration iter = null;
        Vector result = new Vector();
        Connection connection = null;

        try {
            connection = DatabaseService.getConnection();

            iter = DatabaseService.getAdapter().getAllUsers(
                connection);
            // copy the elements into the vector
            while (iter.hasMoreElements()) {
                result.addElement(iter.nextElement());
            }
            connection.commit();
        } catch (Exception exception) {
            _log.error("Failed on get all users", exception);
            SQLHelper.rollback(connection);
        } finally {
            SQLHelper.close(connection);
        }

        return result;
    }

}
