/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: IpcJmsSessionConnection.java,v 1.28 2003/08/30 07:58:51 tanderson Exp $
 *
 * Date         Author  Changes
 * $Date	    jimm    Created
 */
package org.exolab.jms.server.mipc;

import java.io.Serializable;
import java.util.Vector;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.ipc.NotifierIfc;
import org.exolab.core.mipc.MultiplexConnectionIfc;
import org.exolab.jms.client.JmsQueue;
import org.exolab.jms.client.JmsTopic;
import org.exolab.jms.message.MessageImpl;
import org.exolab.jms.server.JmsServerConnection;
import org.exolab.jms.server.JmsServerConnectionManager;
import org.exolab.jms.server.JmsServerSession;


/**
 * This is the server side receiver for JmsSession requests. All requests are
 * unpacked and passed on to the appropriate JmsServerSession object.
 *
 * @version     $Revision: 1.28 $ $Date: 2003/08/30 07:58:51 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.jms.server.mipc.IpcJmsReceiver
 * @see	       	org.exolab.jms.server.JmsServerConnection
 * @see	       	org.exolab.jms.server.JmsServerConnectionManager
 * @see         org.exolab.jms.server.JmsServerSession
 */
public class IpcJmsSessionConnection implements NotifierIfc {

    /**
     * The server
     */
    private IpcJmsServer _server;

    /**
     * The logger
     */
    private static final Log _log =
        LogFactory.getLog(IpcJmsSessionConnection.class);


    /**
     * Construct a new <code>IpcJmsSessionConnection</code>
     *
     * @param server the server instance
     */
    public IpcJmsSessionConnection(IpcJmsServer server) {
        _server = server;
    }

    /**
     * A new request has been received.
     * Carry out the request, and pass back any relevent data.
     *
     * @param ob The data received
     * @param id The unique IPC id of the connection
     * @param The unique identifier of this connection.
     * @return Object Return any requested result. This must never be null.
     *
     */
    public Serializable notify(Object ob, String id) {
        Vector v = (Vector) ob;
        String func = (String) v.get(1);
        JmsServerSession session = getSession(id, v);
        Serializable result = null;
        if (session == null) {
            JMSException error = new JMSException(
                "Failed to process request " + func
                + ": session not found");
            result = pack(Boolean.FALSE, error);
        } else {
            if (func.equals("close")) {
                result = close(session, getConnection(id));
            } else if (func.equals("acknowledgeMessage")) {
                result = acknowledgeMessage(session, (Long) v.get(5),
                                            (String) v.get(6));
            } else if (func.equals("sendMessage")) {
                result = sendMessage(session, (Message) v.get(5));
            } else if (func.equals("sendMessages")) {
                result = sendMessages(session, (Vector) v.get(5));
            } else if (func.equals("receiveMessage")) {
                result = receiveMessage(session, (Long) v.get(5),
                                        (Long) v.get(6));
            } else if (func.equals("receiveMessages")) {
                result = receiveMessages(session, (Long) v.get(5),
                                         (Integer) v.get(6));
            } else if (func.equals("createQueue")) {
                result = createQueue(session, (JmsQueue) v.get(5));
            } else if (func.equals("createTopic")) {
                result = createTopic(session, (JmsTopic) v.get(5));
            } else if (func.equals("createReceiver")) {
                result = createReceiver(
                    session, (JmsQueue) v.get(5), (Long) v.get(6),
                    (String) v.get(7), getConnection(id), (String) v.get(8),
                    (String) v.get(9), (String) v.get(10));
            } else if (func.equals("createSender")) {
                result = createSender(session, (JmsQueue) v.get(5));
            } else if (func.equals("createBrowser")) {
                result = createBrowser(
                    session, (JmsQueue) v.get(5), (Long) v.get(6),
                    (String) v.get(7), getConnection(id), (String) v.get(8),
                    (String) v.get(9), (String) v.get(10));
            } else if (func.equals("deleteReceiver")) {
                result = deleteReceiver(session, (Long) v.get(5));
            } else if (func.equals("deleteSender")) {
                result = deleteSender(session, (Long) v.get(5));
            } else if (func.equals("deleteBrowser")) {
                result = deleteBrowser(session, (Long) v.get(5));
            } else if (func.equals("createSubscriber")) {
                result = createSubscriber(
                    session, (JmsTopic) v.get(5), (String) v.get(6),
                    (Long) v.get(7), (String) v.get(8), (Boolean) v.get(9),
                    getConnection(id), (String) v.get(10), (String) v.get(11),
                    (String) v.get(12));
            } else if (func.equals("createPublisher")) {
                result = createPublisher(session, (JmsTopic) v.get(5));
            } else if (func.equals("deleteSubscriber")) {
                result = deleteSubscriber(session, (Long) v.get(5));
            } else if (func.equals("deletePublisher")) {
                result = deletePublisher(session, (JmsTopic) v.get(5));
            } else if (func.equals("unsubscribe")) {
                result = unsubscribe(session, (String) v.get(5));
            } else if (func.equals("stopMessageDelivery")) {
                result = stopMessageDelivery(session);
            } else if (func.equals("startMessageDelivery")) {
                result = startMessageDelivery(session);
            } else if (func.equals("recover")) {
                result = recover(session);
            } else if (func.equals("commit")) {
                result = commit(session);
            } else if (func.equals("rollback")) {
                result = rollback(session);
            } else if (func.equals("xa_commit")) {
                result = XACommit(session, (Xid) v.get(5), (Boolean) v.get(6));
            } else if (func.equals("xa_end")) {
                result = XAEnd(session, (Xid) v.get(5), (Integer) v.get(6));
            } else if (func.equals("xa_forget")) {
                result = XAForget(session, (Xid) v.get(5));
            } else if (func.equals("xa_getTransactionTimeout")) {
                result = XAGetTransactionTimeout(session);
            } else if (func.equals("xa_recover")) {
                result = XARecover(session, (Integer) v.get(5));
            } else if (func.equals("xa_rollback")) {
                result = XARollback(session, (Xid) v.get(5));
            } else if (func.equals("xa_setTransactionTimeout")) {
                result = XASetTransactionTimeout(session, (Integer) v.get(5));
            } else if (func.equals("xa_start")) {
                result = XAStart(session, (Xid) v.get(5), (Integer) v.get(6));
            } else if (func.equals("xa_prepare")) {
                result = XAPrepare(session, (Xid) v.get(5));
            } else if (func.equals("xa_getResourceManagerId")) {
                result = XAGetResourceManagerId(session);
            } else if (func.equals("enableAsynchronousDelivery")) {
                result = enableAsynchronousDelivery(
                    session, (Long) v.get(5), (String) v.get(6),
                    (Boolean) v.get(7));
            } else {
                JMSException error = new JMSException(
                    "Unknown request received: " + func);
                result = pack(Boolean.FALSE, error);
            }
        }

        return result;
    }

    /**
     * The connection has been broken.              
     *
     * @param id The unique identifier of this connection.
     */
    public void disconnection(String id) {
    }

    /**
     * A convenience routine to get the session.
     *
     * @param id The unique IPC id of the connection
     * @param v The vector containing the packed data from the client.
     * @return JmsServerSession The required session or null if not found.
     *
     */
    protected JmsServerSession getSession(String id, Vector v) {
        // Lookup the connection using the unique connection id
        JmsServerConnection connection =
            JmsServerConnectionManager.instance().getConnection
            ((String) v.get(2));

        JmsServerSession session = null;

        if (connection != null) {
            session = connection.getSession((String) v.get(4));
        } else if (_log.isDebugEnabled()) {
            _log.debug("Failed to locate connection=" + v.get(2));
        }
        if (session == null && _log.isDebugEnabled()) {
            _log.debug("Failed to locate session=" + v.get(4));
        }

        return session;
    }

    /**
     * A convenience routine to get the MutliplexConnection.
     *
     * @param id The unique IPC id of the connection
     * @return MultiplexConnection The connection associated with the ipc channel.
     *
     */
    protected MultiplexConnectionIfc getConnection(String id) {
        IpcServerChannel channel = IpcServerChannel.getServerChannel(id);
        MultiplexConnectionIfc mc = null;
        if (channel != null) {
            mc = channel.getConnection();
        }
        return mc;
    }

    /**
     * A close session request has been received.
     *
     * @param session The session to close
     * @return Vector The result of the request.
     *
     */
    protected Vector close(JmsServerSession session,
                           MultiplexConnectionIfc connection) {
        try {
            _server.removeConnection(session, connection);
            session.close();
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Acknowledge a JmsMessage
     *
     * @param session - the session the request is for
     * @param clientId - the identity of the client that sent it
     * @param id The id of the message
     * @return Vector The result of the request.
     *
     */
    protected Vector acknowledgeMessage(JmsServerSession session,
                                        Long clientId, String id) {
        try {
            session.acknowledgeMessage(clientId.longValue(), id);
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * A JmsMessage has been sent.
     *
     * @param session The session the request is for.
     * @param message The message to process.
     * @return Vector The result of the request.
     *
     */
    protected Vector sendMessage(JmsServerSession session, Message message) {
        boolean persistent = ((MessageImpl) message).isPersistent();

        try {
            session.sendMessage(message);
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        if (persistent) {
            return pack(Boolean.TRUE, null);
        }
        // non-persistent messages do not need to send back anything.
        return null;
    }

    /**
     * A collection of messages have been sent
     *
     * @param session - the session the request is for.
     * @param messages - the messages to process.
     * @return Vector - the result of the request.
     *
     */
    protected Vector sendMessages(JmsServerSession session, Vector messages) {
        try {
            session.sendMessages(messages);
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Receive a message from the provider
     *
     * @param session The session the request is for.
     * @param clientId The client identity
     * @param wait How long to wait
     * @return Vector The result of the request.
     *
     */
    protected Vector receiveMessage(JmsServerSession session, Long clientId,
                                    Long wait) {
        Message message = null;
        try {
            message = session.receiveMessage(clientId.longValue(),
                                             wait.longValue());
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, message);
    }


    /**
     * Receive upto count messages from the endpoint
     *
     * @param session - the session the request is for.
     * @param clientId - the client identity
     * @param count - max number of messages to receive
     * @return Vector - the result of the request.
     *
     */
    protected Vector receiveMessages(JmsServerSession session, Long clientId,
                                     Integer count) {
        Vector messages = null;
        try {
            messages = session.receiveMessages(clientId.longValue(),
                                               count.intValue());
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }
        
        return pack(Boolean.TRUE, messages);
    }


    /**
     * Create a new Queue.
     *
     * @param session The session the request is for.
     * @param queue The queue to create
     * @return Vector The result of the request.
     *
     */
    protected Vector createQueue(JmsServerSession session, JmsQueue queue) {
        try {
            session.createQueue(queue);
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Create a new topic
     *
     * @param session The session the request is for.
     * @param topic The topic to create
     * @return Vector The result of the request.
     *
     */
    protected Vector createTopic(JmsServerSession session, JmsTopic topic) {
        try {
            session.createTopic(topic);
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Create a new receiver
     *
     * @param session The session the request is for.
     * @param queue The queue to create the reciver for
     * @param consumerName The unique name of this consumer,
     * only valid for persitent messages
     * @param selector The selector to filter messages. This may be null.
     * @param connection The MultiplexConnection to the machine the consumer is on
     * @param host The host the client is running on. Only used for http.
     * @param port The port the client is listening on. Only used for http
     * @param url The url for the clients web server. Only used for http
     * @return Vector The result of the request.
     *
     */
    protected Vector createReceiver(
        JmsServerSession session, JmsQueue queue,
         Long consumerId, String selector, MultiplexConnectionIfc connection,
         String host, String port, String url) {

        try {
            session.createReceiver(queue, consumerId.longValue(), selector);
            _server.addConnection(session, connection);
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }


    /**
     * Create a new sender for the given queue.
     *
     * @param session The session the request is for.
     * @param queue The queue the sender is sending to
     * @return Vector The result of the request.
     *
     */
    protected Vector createSender(JmsServerSession session, JmsQueue queue) {
        try {
            session.createSender(queue);
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }
        return pack(Boolean.TRUE, null);
    }


    /**
     * Create a new queue browser for the specified session and queue.
     *
     * @param session             session that the request is for
     * @param queue               queue to browse
     * @param clientId            the client identity
     * @param selector            message selector. May be null
     * @param connection          the connection to the remote machine
     * @param host The host the client is running on. Only used for http.
     * @param port The port the client is listening on. Only used for http
     * @param url The url for the clients web server. Only used for http
     * @return      Vector              result of the request
     *
     */
    protected Vector createBrowser(JmsServerSession session,
                                   JmsQueue queue, Long clientId, 
                                   String selector,
                                   MultiplexConnectionIfc connection, 
                                   String host, String port,
                                   String url) {

        try {
            session.createBrowser(queue, clientId.longValue(), selector);
            _server.addConnection(session, connection);
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }
        return pack(Boolean.TRUE, null);
    }

    /**
     * Delete the receiver for this queue.
     *
     * @param session  - the session the request is for.
     * @param clientId -  the client id
     * @return Vector The result of the request.
     *
     */
    protected Vector deleteReceiver(JmsServerSession session, Long clientId) {
        try {
            session.deleteReceiver(clientId.longValue());
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Delete the sender for this queue.
     *
     * @param session The session the request is for.
     * @param clientId The identity of the client
     * @return Vector The result of the request.
     */
    protected Vector deleteSender(JmsServerSession session, Long clientId) {
        try {
            session.deleteSender(clientId.longValue());
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Delete the queue browser for the specified queue.
     *
     * @param       session             session that the request is for
     * @param       clientId            the identity of the browser
     * @return      Vector              The result of the request.
     */
    protected Vector deleteBrowser(JmsServerSession session, Long clientId) {
        try {
            session.deleteBrowser(clientId.longValue());
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Create a new subscriber, and connect back to the client through the
     * MultiplexConnection.
     *
     * @param session The session the request is for.
     * @param topic The topic the subscriber is subscribing on
     * @param name The unique name of this subscriber,
     * only valid for persitent messages
     * @param selector The selector to filter messages. This may be null.
     * @param connection The MultiplexConnection to the machine the consumer is on
     * @param host The host the client is running on. Only used for http.
     * @param port The port the client is listening on. Only used for http
     * @param url The url for the clients web server. Only used for http
     * @return Vector The result of the request.
     */
    protected Vector createSubscriber(
        JmsServerSession session, JmsTopic topic, String name, Long clientId,
        String selector, Boolean noLocal, MultiplexConnectionIfc connection,
        String host, String port, String url) {
        try {
            session.createSubscriber(topic, name, clientId.longValue(),
                                     selector, noLocal.booleanValue());
            _server.addConnection(session, connection);
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }


    /**
     * Create a new publisher for the given topic.
     *
     * @param session The session the request is for.
     * @param topic The topic the publisher is publishing to
     * @return Vector The result of the request.
     */
    protected Vector createPublisher(JmsServerSession session,
                                     JmsTopic topic) {
        try {
            session.createPublisher(topic);
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Delete a subscriber for the given topic
     *
     * @param session The session the request is for.
     * @param clientId - The client identity of the subscriber to delete
     * @return Vector The result of the request.
     */
    protected Vector deleteSubscriber(JmsServerSession session,
                                      Long clientId) {
        try {
            session.deleteSubscriber(clientId.longValue());
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Delete a publisher for the given topic.
     *
     * @param session The session the request is for.
     * @param topic The topic the publisher is publishing to
     * @return Vector The result of the request.
     *
     */
    protected Vector deletePublisher(JmsServerSession session,
                                     JmsTopic topic) {
        try {
            session.deletePublisher(topic);
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Unsubscribe a durable subscription
     *
     * @param       session             the session the request is for
     * @param       name                the name used to identify the
     *                                  subscription
     * @return      Vector              the result of the request.
     */
    public Vector unsubscribe(JmsServerSession session, String name) {
        try {
            session.unsubscribe(name);
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }
        return pack(Boolean.TRUE, null);
    }

    /**
     * Stop message delivery for this session.
     *
     * @param session The session the request is for.
     * @return Vector The result of the request.
     *
     */
    protected Vector stopMessageDelivery(JmsServerSession session) {
        try {
            session.stopMessageDelivery();
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }


    /**
     * Enable or disable asynchronous message delivery.
     *
     * @param session - the session the request is for.
     * @param clientId - consumer identity
     * @param enable - true to enable
     * @return Vector - the result of the request.
     *
     */
    protected Vector enableAsynchronousDelivery(JmsServerSession session,
                                                Long clientId, String id,
                                                Boolean enable) {
        try {
            session.enableAsynchronousDelivery(
                clientId.longValue(), id, enable.booleanValue());
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Start message delivery for this session.
     *
     * @param session The session the request is for.
     * @return Vector The result of the request.
     *
     */
    protected Vector startMessageDelivery(JmsServerSession session) {
        try {
            session.startMessageDelivery();
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * recover the session
     *
     * @param session The session the request is for.
     * @return Vector The result of the request.
     */
    protected Vector recover(JmsServerSession session) {
        try {
            session.recover();
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * commit the session
     *
     * @param session The session the request is for.
     * @return Vector The result of the request.
     */
    protected Vector commit(JmsServerSession session) {
        try {
            session.commit();
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * rollback the session
     *
     * @param session The session the request is for.
     * @return Vector The result of the request.
     */
    protected Vector rollback(JmsServerSession session) {
        try {
            session.rollback();
        } catch (JMSException exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Commits an XA transaction that is in progress.
     *
     * @param xid - the xa transaction identity
     * @param onePhase - treu if it is a one phase commit
     * @return Vector
     */
    protected Vector XACommit(JmsServerSession session, Xid xid,
                              Boolean onePhase) {
        try {
            session.commit(xid, onePhase.booleanValue());
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Ends the work performed on behalf of a transaction branch. The resource
     * manager disassociates the XA resource from the transaction branch
     * specified and let the transaction be completedCommits an XA transaction
     * that is in progress.
     *
     * @param xid - the xa transaction identity
     * @param flags - one of TMSUCCESS, TMFAIL, or TMSUSPEND
     * @return Vector
     */
    protected Vector XAEnd(JmsServerSession session, Xid xid, Integer flags) {
        try {
            session.end(xid, flags.intValue());
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Tell the resource manager to forget about a heuristically completed
     * transaction branch.
     *
     * @param xid - the xa transaction identity
     * @return Vector
     */
    protected Vector XAForget(JmsServerSession session, Xid xid) {
        try {
            session.forget(xid);
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Return the transaction timeout for this instance of the resource
     * manager.
     *
     * @return int - the timeout in seconds
     * @return Vector
     */
    protected Vector XAGetTransactionTimeout(JmsServerSession session) {
        int seconds = 0;
        try {
            seconds = session.getTransactionTimeout();
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, new Integer(seconds));
    }

    /**
     * Ask the resource manager to prepare for a transaction commit of the
     * transaction specified in xid
     *
     * @param xid - the xa transaction identity
     * @return Vector
     */
    protected Vector XAPrepare(JmsServerSession session, Xid xid) {
        try {
            session.prepare(xid);
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Obtain a list of prepared transaction branches from a resource manager.
     * The transaction manager calls this method during recovery to obtain the
     * list of transaction branches that are currently in prepared or
     * heuristically completed states.
     *
     * @param flag - One of TMSTARTRSCAN, TMENDRSCAN, TMNOFLAGS. TMNOFLAGS
     * @param Xid[] - the set of Xids to recover
     * @return Vector
     */
    public Vector XARecover(JmsServerSession session, Integer flag) {
        Xid[] xids = null;
        try {
            xids = session.recover(flag.intValue());
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, xids);
    }

    /**
     * Inform the resource manager to roll back work done on behalf of a
     * transaction branch
     *
     * @param xid - the xa transaction identity
     * @author Vector
     */
    protected Vector XARollback(JmsServerSession session, Xid xid) {
        try {
            session.rollback(xid);
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Set the current transaction timeout value for this XAResource instance.
     *
     * @param seconds - timeout in seconds
     * @return boolean - true if the new transaction timeout was accepted
     * @return Vector
     */
    protected Vector XASetTransactionTimeout(JmsServerSession session,
                                             Integer seconds) {
        boolean result = false;
        try {
            result = session.setTransactionTimeout(seconds.intValue());
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, new Boolean(result));
    }

    /**
     * Start work on behalf of a transaction branch specified in xid If TMJOIN
     * is specified, the start is for joining a transaction previously seen by
     * the resource manager
     *
     * @param xid - the xa transaction identity
     * @param flags - One of TMNOFLAGS, TMJOIN, or TMRESUME
     * @return Vector
     */
    protected Vector XAStart(JmsServerSession session, Xid xid,
                             Integer flags) {
        try {
            session.start(xid, flags.intValue());
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, null);
    }

    /**
     * Return the identity of the resource manager.
     *
     * @param session - the session identity
     * @return Vector
     */
    protected Vector XAGetResourceManagerId(JmsServerSession session) {
        String id = null;
        try {
            id = session.getResourceManagerId();
        } catch (Exception exception) {
            return pack(Boolean.FALSE, exception);
        }

        return pack(Boolean.TRUE, id);
    }

    /**
     * Pack all the data that is required by the server in a vector.
     * Set the size of the vector to be exactly the right size for efficiency.
     *
     * @param success Boolean indicating success or failure of request.
     * @param ob The Object being returned.
     * @return Vector The vector containing all the data.
     *
     */
    protected Vector pack(Boolean success, Object ob) {
        Vector v = new Vector(2);
        v.add(success);
        v.add(ob);
        return v;
    }

}

