/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: IpcJmsSessionSender.java,v 1.10 2003/08/07 13:33:09 tanderson Exp $
 *
 * Date         Author  Changes
 * $Date	    jimm    Created
 */


package org.exolab.jms.server.mipc;

import java.util.Vector;

import javax.jms.Message;

import org.exolab.core.mipc.ObjectChannel;
import org.exolab.jms.client.JmsMessageListener;
import org.exolab.jms.message.MessageImpl;
import org.exolab.jms.scheduler.Scheduler;
import org.exolab.jms.server.ClientDisconnectionException;
import org.exolab.jms.server.JmsServerSession;


/**
 * This class contains the ipc connection to a receiver or subscriber
 * for passing the JmsMessages.
 *
 * If the connection is detected as closed, the Session will also be shutdown.
 *
 * @version     $Revision: 1.10 $ $Date: 2003/08/07 13:33:09 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.jms.server.mipc.IpcJmsSessionConnection
 * @see		org.exolab.jms.server.JmsServerSession
 * @see		org.exolab.core.ipc.Client
 *
 */
public class IpcJmsSessionSender implements JmsMessageListener {

    /**
     * The client connection
     */
    private ObjectChannel _client;

    /**
     * The session
     */
    private JmsServerSession _session = null;


    /**
     * Create a connection to the client listener for sending JmsMessages.
     *
     * @param client The connection to the client.
     * @param session The session this connection belongs to.
     *
     */
    public IpcJmsSessionSender(ObjectChannel client,
                               JmsServerSession session) {
        _client = client;
        _session = session;
        session.setMessageListener(this);
    }

    /**
     * Send a JmsMessage to a listener.
     *
     * @param message The message to send.
     */
    public synchronized void onMessage(Message message) {
        try {
            if (_client != null) {
                Vector v = new Vector(2);
                v.add(_session.getSessionId());
                v.add((MessageImpl) message);

                // artw: this was not synchronized before.  In the old code,
                // it was "safe" to call into Client by multi-threads.  Now
                // with the Mutliplex stuff, each channel reader/writer
                // has state in it during the receive/read and write/send
                // calls.  I could synchronize the inside of the channel
                // but decided to do it right here for now.
                _client.send(v);
                Vector reply = (Vector) _client.receive();
            }
        } catch (Exception err) {
            throw new ClientDisconnectionException(err.getMessage());
        }
    }

    /**
     * Send the collection of messages to the client. This is used for async
     * message delivery.
     *
     * @param messages - collection of MessageImpl objects
     */
    public synchronized void onMessages(Vector messages) {
        try {
            if (_client != null) {
                Vector v = new Vector(2);
                v.add(_session.getSessionId());
                v.add(messages);
                // artw: this was not synchronized before.  In the old code,
                // it was "safe" to call into Client by multi-threads.  Now
                // with the Mutliplex stuff, each channel reader/writer
                // has state in it during the receive/read and write/send
                // calls.  I could synchronize the inside of the channel
                // but decided to do it right here for now.
                _client.send(v);
                Vector reply = (Vector) _client.receive();
            }
        } catch (Exception err) {
            throw new ClientDisconnectionException(err.getMessage());
        }
    }

    /**
     * Notify the specified client that a message avaiable.
     *
     * @param clinet - the client identity
     */
    public synchronized void onMessageAvailable(long clientId) {
        // Workaround for bug 777419 - Deadlock in the tcp connector
        Scheduler.instance().add(new Notifier(clientId));
    }

    /**
     * Remove the callback from the JmsServerSession.
     */
    public synchronized void close() {
        if (_session != null) {
            _session.setMessageListener(null);
            _session = null;
        }
        _client = null;
    }

    private class Notifier implements Runnable {

        private long _clientId;

        public Notifier(long clientId) {
            _clientId = clientId;
        }

        public void run() {
            synchronized (IpcJmsSessionSender.this) {
                try {
                    if (_client != null) {
                        Vector v = new Vector(2);
                        v.add(_session.getSessionId());
                        v.add(new Long(_clientId));
                        _client.send(v);
                        Vector reply = (Vector) _client.receive();
                    }
                } catch (Exception exception) {
                    exception.printStackTrace();
                }
            }
        }
    }

} //-- IpcJmsSessionSender
