/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: RemoteJmsServerSessionIfc.java,v 1.18 2003/08/07 13:33:10 tanderson Exp $
 *
 * Date         Author  Changes
 * 04/07/2000   jima    Created
 * 05/09/2000	jima	Changed acknowledgeMessage signature to take
 * 			destination name instead of client identity
 */
package org.exolab.jms.server.rmi;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.util.Vector;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

import org.exolab.jms.client.JmsQueue;
import org.exolab.jms.client.JmsTopic;
import org.exolab.jms.client.rmi.RemoteJmsMessageListenerIfc;


/**
 * This is an RMI wrapper for the JmsServerSession object. It allows remote
 * access to this object.
 * <p>
 * For more semantic information on this object refer to org.exolab.jms.server.
 * JmsServerSession.
 * <p>
 * All the methods throw RemoteException which is used to indicate an RMI
 * specific error.
 *
 * @version     $Revision: 1.18 $ $Date: 2003/08/07 13:33:10 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         org.exolab.jms.server.JmsServerSession
 * @see         org.exolab.jms.server.rmi.RemoteJmsServerConnectionIfc
 **/
public interface RemoteJmsServerSessionIfc
    extends Remote {

    /**
     * Return a reference to the client id.
     *
     * @return      String      client id
     * @exception   RemoteException
     */
    public String getClientId()
        throws RemoteException;

    /**
     * Return a reference to the session id
     *
     * @return      String      session id
     * @exception   RemoteException
     */
    public String getSessionId()
        throws JMSException, RemoteException;

    /**
     * Close and release any resource allocated to this session. Throw the
     * JMSException exception is a problem is encountered
     *
     * @exception   JMSException
     * @exception   RemoteException
     */
    public void close()
        throws JMSException, RemoteException;

    /**
     * Acknowledge that the message with the following id has been processed
     *
     * @param       clientId            the client identity
     * @param       messageId           id of message to ack
     * @exception   JMSException        if method does not complete
     * @exception   RemoteException
     */
    public void acknowledgeMessage(long clientId, String messageId)
        throws JMSException, RemoteException;

    /**
     * Send the specified message to the server. If there is any problem
     * then throw the JMSException exception
     *
     * @param       message             message to send
     * @exception   JMSException
     * @exception   RemoteException
     */
    public void sendMessage(Message message)
        throws JMSException, RemoteException;

    /**
     * Send the specified messages to the server. If there is any problem
     * then throw the JMSException exception
     *
     * @param       messages            messages to send
     * @exception   JMSException
     * @exception   RemoteException
     */
    public void sendMessages(Vector messages)
        throws JMSException, RemoteException;

    /**
     * Return the next message for the specified client. The <code>wait</code>
     * parameter indicates how long many milliseconds to wait for a message
     * before returning. If <code>wait</code> is 0 then do not wait at all. If
     * <code>wait</code> is -1 then wait indefinitely for the next message
     *
     * @param       clientId            the client identity
     * @param       wait                number of ms to wait
     * @return      Message             the next message or null
     * @exception   JMSException        if there is an app level problem
     * @exception   RemoteException     if there is a RMI based exception.
     */
    public Message receiveMessage(long clientId, long wait)
        throws JMSException, RemoteException;

    /**
     * Return upto count messages from the endpoint with the specified client
     * identity.
     *
     * @param       clientId            the client identity
     * @param       count               max number of messages to receive
     * @return      Message             the next message or null
     * @exception   JMSException        if there is an app level problem
     * @exception   RemoteException     if there is a RMI based exception.
     */
    public Vector receiveMessages(long clientId, int count)
        throws JMSException, RemoteException;

    /**
     * Create a queue with the specified name. If the queue already exists
     * then simply return a reference to it. If the queue does not exist
     * then create it. If it cannot create the queue then throw the
     * JMSException exception
     *
     * @param       queue               queue to create
     * @exception   JMSException
     * @exception   RemoteException
     */
    public void createQueue(JmsQueue queue)
        throws JMSException, RemoteException;

    /**
     * Create a topic with the specified name. If the topic already exists
     * then simply return a reference to it. If the topic does not exist
     * then create it. If it cannot create the topic then throw the
     * JMSException exception
     *
     * @param       topic               topic to create
     * @exception   JMSException
     * @exception   RemoteException
     */
    public void createTopic(JmsTopic topic)
        throws JMSException, RemoteException;

    /**
     * Create a receiver endpoint for this session. A reciever is a message
     * consumer specific to the queue message model. The receiver is
     * associated with a queue.
     * <p>
     * You cannot create more than one receiver for the same destination
     *
     * @param       queue               receiver destination
     * @param       clientId            the session allocated identifier of
     *                                  this consumer
     * @param       selector            the selector to filter messages.
     *                                  This may be null.
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public void createReceiver(JmsQueue queue, long clientId, String selector)
        throws JMSException, RemoteException;

    /**
     * Create a sender endpoint for this session. A sender is a message
     * publisher specific to the queue message model. The sender is associated
     * with a queue.
     * <p>
     *
     * You cannot create more than one receiver for the same destination
     *
     * @param       queue               receiver destination
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public void createSender(JmsQueue queue)
        throws JMSException, RemoteException;

    /**
     * Delete the receiver with the specified identity. If the receiver does
     * not exist or cannot be deleted throw JMSException
     *
     * @param       queue
     * @exception   JMSException.
     * @exception   RemoteException
     */

    /**
     * Create a queue browser for this session. This allows clients to browse
     * a queue without removing any messages.
     * <p>
     *
     * You cannot create more than one queue browser for the same queue
     * in a single session.
     *
     * @param       queue               queue to browse
     * @param       clientId            identity of the client
     * @param       selector            message selector. This may be null
     * @exception   JMSException
     * @exception   RemoteException
     */
    public void createBrowser(JmsQueue queue, long clientId, String selector)
        throws JMSException, RemoteException;

    /**
     * Delete the specified receiver
     *
     * @param       clientId            the identity of the receiver
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public void deleteReceiver(long clientId)
        throws JMSException, RemoteException;

    /**
     * Delete the sender for the specified queue. If the sender does not
     * exist or the sender cannot be deleted then throw JMSException
     *
     * @param       clientId             the unique client identity
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public void deleteSender(long clientId)
        throws JMSException, RemoteException;

    /**
     * Delete the queue browser associated with the specified queue from
     * the session.
     * If the corresponding queue does not exist or it cannot be deleted,
     * then throw a JMSException.
     *
     * @param       clientId - the client identity
     * @exception   JMSException
     * @exception   RemoteException
     */
    public void deleteBrowser(long clientId)
        throws JMSException, RemoteException;

    /**
     * Create a subscriber endpoint for this session. A subscriber is a message
     * consumer specific to the topic message model. The subscriber is
     * associated with a topic. The consumer name is used to identify an
     * endpoint for persistent delivery. If we are using non-persistent
     * delivery then this can be set to null
     * <p>
     * You cannot create more than one subscriber for the same destination
     *
     * @param       topic               subscriber destination
     * @param       name                consumer name
     * @param       clientId            the session allocated identifier of
     *                                  this consumer
     * @param       selector            the selector to filter messages.
     *                                  This may be null.
     * @param       noLocal             inhibit consuming messages on same
     *                                  connection.
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public void createSubscriber(JmsTopic topic, String name, long clientId,
                                 String selector, boolean noLocal)
        throws JMSException, RemoteException;

    /**
     * Create a publisher endpoint for this session. A publisher is a message
     * publisher specific to the topic message model. The publisher is
     * associated with a topic. Return the unique identity of the publisher
     * <p>
     * You cannot create more than one publisher for the same destination
     *
     * @param       topic               receiver destination
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public void createPublisher(JmsTopic topic)
        throws JMSException, RemoteException;

    /**
     * Delete the subscriber for the specified topic. If this subscriber
     * does not exist or the cannot be deleted then throw JMSException.
     *
     * @param       clientId            the client identity
     * @exception   JMSException.
     */
    public void deleteSubscriber(long clientId)
        throws JMSException, RemoteException;

    /**
     * Delete the publisher for the specified topic. If the publisher does
     * not exist or cannot be deleted then throw JMSException
     *
     * @param       topic               topic name
     * @exception   JMSException.
     * @exception   RemoteException.
     */
    public void deletePublisher(JmsTopic topic)
        throws JMSException, RemoteException;

    /**
     * Unsubscribe a durable subscription
     *
     * @param       name                the name used to identify the
     *                                  subscription
     * @exception   JMSException        if the subscription cannot be removed
     * @exception   RemoteException
     */
    public void unsubscribe(String name)
        throws JMSException, RemoteException;

    /**
     * Stop message delivery to this session. If there are any problems
     * completing the request then throw the JMSException exception
     *
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public void stopMessageDelivery()
        throws JMSException, RemoteException;

    /**
     * Start message delivery to this session. If there are any problems
     * completing this request then throw the JMSException exception
     *
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public void startMessageDelivery()
        throws JMSException, RemoteException;

    /**
     * Recover the session. If there are any problems completing the request
     * then throw the JMSException exception
     *
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public void recover()
        throws JMSException, RemoteException;

    /**
     * Commit the session. If there are any problems completing the request
     * then throw the JMSException exception
     *
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public void commit()
        throws JMSException, RemoteException;

    /**
     * Rollback the session. If there are any problems completing the request
     * then throw the JMSException exception
     *
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public void rollback()
        throws JMSException, RemoteException;

    /**
     * Commits an XA transaction that is in progress.
     *
     * @param xid - the xa transaction identity
     * @param onePhase - treu if it is a one phase commit
     * @throws XAException - if there is a problem completing the call
     * @throws RemoteException - communication related error
     */
    public void commit(Xid xid, boolean onePhase)
        throws XAException, RemoteException;

    /**
     * Ends the work performed on behalf of a transaction branch. The resource
     * manager disassociates the XA resource from the transaction branch
     * specified and let the transaction be completedCommits an XA transaction
     * that is in progress.
     *
     * @param xid - the xa transaction identity
     * @param flags - one of TMSUCCESS, TMFAIL, or TMSUSPEND
     * @throws XAException - if there is a problem completing the call
     * @throws RemoteException - communication related error
     */
    public void end(Xid xid, int flags)
        throws XAException, RemoteException;

    /**
     * Tell the resource manager to forget about a heuristically completed
     * transaction branch.
     *
     * @param xid - the xa transaction identity
     * @throws XAException - if there is a problem completing the call
     * @throws RemoteException - communication related error
     */
    public void forget(Xid xid)
        throws XAException, RemoteException;

    /**
     * Return the transaction timeout for this instance of the resource
     * manager.
     *
     * @return int - the timeout in seconds
     * @throws XAException - if there is a problem completing the call
     * @throws RemoteException - communication related error
     */
    public int getTransactionTimeout()
        throws XAException, RemoteException;

    /**
     * Ask the resource manager to prepare for a transaction commit of the
     * transaction specified in xid
     *
     * @param xid - the xa transaction identity
     * @return int - XA_RDONLY or XA_OK
     * @throws XAException - if there is a problem completing the call
     * @throws RemoteException - communication related error
     */
    public int prepare(Xid xid)
        throws XAException, RemoteException;

    /**
     * Obtain a list of prepared transaction branches from a resource manager.
     * The transaction manager calls this method during recovery to obtain the
     * list of transaction branches that are currently in prepared or
     * heuristically completed states.
     *
     * @param flag - One of TMSTARTRSCAN, TMENDRSCAN, TMNOFLAGS. TMNOFLAGS
     * @param Xid[] - the set of Xids to recover
     * @throws XAException - if there is a problem completing the call
     * @throws RemoteException - communication related error
     */
    public Xid[] recover(int flag)
        throws XAException, RemoteException;

    /**
     * Inform the resource manager to roll back work done on behalf of a
     * transaction branch
     *
     * @param xid - the xa transaction identity
     * @throws XAException - if there is a problem completing the call
     * @throws RemoteException - communication related error
     */
    public void rollback(Xid xid)
        throws XAException, RemoteException;

    /**
     * Set the current transaction timeout value for this XAResource instance.
     *
     * @param seconds - timeout in seconds
     * @return boolean - true if the new transaction timeout was accepted
     * @throws XAException - if there is a problem completing the call
     * @throws RemoteException - communication related error
     */
    public boolean setTransactionTimeout(int seconds)
        throws XAException, RemoteException;

    /**
     * Start work on behalf of a transaction branch specified in xid If TMJOIN
     * is specified, the start is for joining a transaction previously seen by
     * the resource manager
     *
     * @param xid - the xa transaction identity
     * @param flags - One of TMNOFLAGS, TMJOIN, or TMRESUME
     * @throws XAException - if there is a problem completing the call
     * @throws RemoteException - communication related error
     */
    public void start(Xid xid, int flags)
        throws XAException, RemoteException;

    /**
     * Return the resource manager identity
     *
     * @return the resource manager identity
     * @throws XAException - if there is a problem completing the call
     * @throws RemoteException - communication related error
     */
    public String getResourceManagerId() throws XAException, RemoteException;

    /**
     * All server side sessions register with the consumer manager for
     * message consumption (i.e push-model). When a message arrives this
     * server-side instance will send it down to the client side stub for
     * further processing.
     * <p>
     * If the listener is null then throw JMSException
     *
     * @param       listener            listener to delivery messages too.
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public void setMessageListener(RemoteJmsMessageListenerIfc listener)
        throws JMSException, RemoteException;

    /**
     * Enable or disable asynchronous message delivery for a particular
     * consumer
     *
     * @param clientId - the id of the client to check
     * @param id - the last message asynchronously delivered
     * @param enable - true to enable; false to disable
     */
    public void enableAsynchronousDelivery(long clientId, String id,
                                           boolean enable)
        throws JMSException, RemoteException;
}
