/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: RmiJmsServerSession.java,v 1.30 2003/08/07 13:33:10 tanderson Exp $
 *
 * Date         Author  Changes
 * 04/07/2000   jima    Created
 */
package org.exolab.jms.server.rmi;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Vector;

import javax.jms.JMSException;
import javax.jms.Message;
import javax.transaction.xa.XAException;
import javax.transaction.xa.XAResource;
import javax.transaction.xa.Xid;

import org.exolab.jms.client.JmsMessageListener;
import org.exolab.jms.client.JmsQueue;
import org.exolab.jms.client.JmsTopic;
import org.exolab.jms.client.rmi.RemoteJmsMessageListenerIfc;
import org.exolab.jms.server.ClientDisconnectionException;
import org.exolab.jms.server.JmsServerSession;


/**
 * This is an implementation of the RemoteJmsServerSessionIfc which delegates
 * all requests to the JmsServerSession class. Actually, an instsnce of this
 * class cannot be created without a JmsServerSession object.
 *
 * @version     $Revision: 1.30 $ $Date: 2003/08/07 13:33:10 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         org.exolab.jms.server.JmsServerSession
 * @see         org.exolab.jms.server.rmi.RemoteJmsServerConnectionIfc
 **/
public class RmiJmsServerSession
    extends UnicastRemoteObject
    implements RemoteJmsServerSessionIfc, JmsMessageListener {

    /**
     * This instance attribute maintains a reference to the delegate. The
     * delegate must be assigned during object construction
     */
    private JmsServerSession _delegate = null;

    /**
     * This is the message listener to the remote object. This is how all the
     * messages are pushed down to the client
     */
    private RemoteJmsMessageListenerIfc _listener = null;


    /**
     * Instantiate an instance of this class given a JmsServerSession.
     * object. If the session object is null then throw the JMSException
     * exception
     *
     * @param       session
     * @exception   JMSException
     * @exception   RemoteException
     */
    public RmiJmsServerSession(JmsServerSession session)
        throws JMSException, RemoteException {
        if (session != null) {
            _delegate = session;
        } else {
            throw new JMSException(
                "Cannot create RmiJmsServerSession with null session");
        }
    }

    /**
     * Return a reference to the client id.
     *
     * @return      String      client id
     * @exception   RemoteException
     */
    public synchronized String getClientId()
        throws RemoteException {
        return _delegate.getClientId();
    }

    /**
     * Return a reference to the session id
     *
     * @return      String      session id
     * @exception   RemoteException
     */
    public synchronized String getSessionId()
        throws JMSException, RemoteException {
        return _delegate.getSessionId();
    }

    /**
     * Close and release any resource allocated to this session. Throw the
     * JMSException exception is a problem is encountered
     *
     * @exception   JMSException
     * @exception   RemoteException
     */
    public synchronized void close()
        throws JMSException, RemoteException {
        _listener = null;
        _delegate.close();
        _delegate = null;
    }

    /**
     * Acknowledge that the message id for the specified client identity. If
     * the acknowledgement faisl then throw JMSException
     *
     * @param       clientId            the client identity.
     * @param       messageId           identity of the message
     * @exception   JMSException        if method does not complete
     * @exception   RemoteException
     */
    public synchronized void acknowledgeMessage(long clientId, String messageId)
        throws JMSException, RemoteException {
        _delegate.acknowledgeMessage(clientId, messageId);
    }

    /**
     * Send the specified message to the server. If there is any problem
     * then throw the JMSException exception
     *
     * @param       message             message to send
     * @exception   JMSException
     * @exception   RemoteException
     */
    public synchronized void sendMessage(Message message)
        throws JMSException, RemoteException {
        _delegate.sendMessage(message);
    }

    /**
     * Send the specified messages to the server. If there is any problem
     * then throw the JMSException exception
     *
     * @param       messages            messages to send
     * @exception   JMSException
     * @exception   RemoteException
     */
    public synchronized void sendMessages(Vector messages)
        throws JMSException, RemoteException {
        _delegate.sendMessages(messages);
    }

    /**
     * Return the next message for the specified client. The <code>wait</code>
     * parameter indicates how long many milliseconds to wait for a message
     * before returning. If <code>wait</code> is 0 then do not wait at all. If
     * <code>wait</code> is -1 then wait indefinitely for the next message
     *
     * @param       clientId            the client identity
     * @param       wait                number of ms to wait
     * @return      Message             the next message or null
     * @exception   JMSException        if there is an app level problem
     * @exception   RemoteException     if there is a RMI based exception.
     */
    public Message receiveMessage(long clientId, long wait)
        throws JMSException, RemoteException {
        return _delegate.receiveMessage(clientId, wait);
    }

    // implementation of RemoteJmsServerSessionIfc.receiveMessages
    public Vector receiveMessages(long clientId, int count)
        throws JMSException, RemoteException {
        return _delegate.receiveMessages(clientId, count);
    }

    /**
     * Create a queue with the specified name. If the queue already exists
     * then simply return a reference to it. If the queue does not exist
     * then create it. If it cannot create the queue then throw the
     * JMSException exception
     *
     * @param       queue               queue to create
     * @exception   JMSException
     * @exception   RemoteException
     */
    public synchronized void createQueue(JmsQueue queue)
        throws JMSException, RemoteException {
        _delegate.createQueue(queue);
    }

    /**
     * Create a topic with the specified name. If the topic already exists
     * then simply return a reference to it. If the topic does not exist
     * then create it. If it cannot create the topic then throw the
     * JMSException exception
     *
     * @param       topic               topic to create
     * @exception   JMSException
     * @exception   RemoteException
     */
    public synchronized void createTopic(JmsTopic topic)
        throws JMSException, RemoteException {
        _delegate.createTopic(topic);
    }

    /**
     * Create a receiver endpoint for this session. A reciever is a message
     * consumer specific to the queue message model. The receiver is
     * associated with a queue.
     * <p>
     * You cannot create more than one receiver for the same destination
     *
     * @param       queue               receiver destination
     * @param       clientId            the session allocated identifier of
     *                                  this consumer
     * @param       selector            the selector to filter messages.
     *                                  This may be null.
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public synchronized void createReceiver(JmsQueue queue, long clientId, String selector)
        throws JMSException, RemoteException {
        _delegate.createReceiver(queue, clientId, selector);
    }

    /**
     * Create a sender endpoint for this session. A sender is a message
     * publisher specific to the queue message model. The sender is associated
     * with a queue.
     * <p>
     * You cannot create more than one receiver for the same destination
     *
     * @param       queue               receiver destination
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public synchronized void createSender(JmsQueue queue)
        throws JMSException, RemoteException {
        _delegate.createSender(queue);
    }

    /**
     * Create a queue browser for this session. This allows clients to browse
     * a queue without removing any messages.
     * <p>
     *
     * You cannot create more than one queue browser for the same queue
     * in a single session.
     *
     * @param       queue               queue to browse
     * @param       clientId            the client identity
     * @param       selector            message selector. This may be null
     * @exception   JMSException
     */
    public synchronized void createBrowser(JmsQueue queue, long clientId, String selector)
        throws JMSException {
        _delegate.createBrowser(queue, clientId, selector);
    }

    /**
     * Delete the receiver for the specified queue. If the receiver does
     * not exist or cannot be deleted throw JMSException
     *
     * @param       id                  identity of receiver
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public synchronized void deleteReceiver(long clientId)
        throws JMSException, RemoteException {
        _delegate.deleteReceiver(clientId);
    }

    /**
     * Delete the sender for the specified queue. If the sender does not
     * exist or the sender cannot be deleted then throw JMSException
     *
     * @param       clientId            identity of client to delete
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public synchronized void deleteSender(long clientId)
        throws JMSException, RemoteException {
        _delegate.deleteSender(clientId);
    }

    /**
     * Delete the queue browser associated with the specified queue from
     * the session.
     * If the corresponding queue does not exist or it cannot be deleted,
     * then throw a JMSException.
     *
     * @param       clientId            the identity of the client
     * @exception   JMSException
     */
    public synchronized void deleteBrowser(long clientId)
        throws JMSException {
        _delegate.deleteBrowser(clientId);
    }

    /**
     * Create a subscriber endpoint for this session. A subscriber is a message
     * consumer specific to the topic message model. The subscriber is
     * associated with a topic. The consumer name must be specfied for
     * persistent delivery but can be null otherwise
     * <p>
     * You cannot create more than one subscriber for the same destination
     *
     * @param       topic               subscriber destination
     * @param       name                name of the cnsumer; can be null
     * @param       clientId            the session allocated identifier of
     *                                  this consumer
     * @param       selector            the selector to filter messages.
     *                                  This may be null.
     * @param       noLocal             inhibit consuming messages on same
     *                                  connection.
     * @return                          the unique consumer identifier
     * @exception   JMSException
     * @exception   RemoteException
     */
    public synchronized void createSubscriber(JmsTopic topic, String name, long clientId,
                                              String selector, boolean noLocal)
        throws JMSException, RemoteException {
        _delegate.createSubscriber(topic, name, clientId, selector, noLocal);
    }

    /**
     * Create a publisher endpoint for this session. A publisher is a message
     * publisher specific to the topic message model. The publisher is
     * associated with a topic.
     * <p>
     * You cannot create more than one publisher for the same destination
     *
     * @param       topic               receiver destination
     * @exception   JMSException.
     * @exception   RemoteException
     */
    public synchronized void createPublisher(JmsTopic topic)
        throws JMSException, RemoteException {
        _delegate.createPublisher(topic);
    }

    /**
     * Delete the subscriber for the specified topic. If this subscriber
     * does not exist or the cannot be deleted then throw JMSException.
     *
     * @param       clientId            the identity of the client
     * @exception   JMSException.
     */
    public synchronized void deleteSubscriber(long clientId)
        throws JMSException, RemoteException {
        _delegate.deleteSubscriber(clientId);
    }

    /**
     * Delete the publisher for the specified topic. If the publisher does
     * not exist or cannot be deleted then throw JMSException
     *
     * @param       topic               topic object
     * @exception   JMSException.
     * @exception   RemoteException.
     */
    public synchronized void deletePublisher(JmsTopic topic)
        throws JMSException, RemoteException {
        _delegate.deletePublisher(topic);
    }

    /**
     * Unsubscribe a durable subscription
     *
     * @param       name                the name used to identify the
     *                                  subscription
     * @exception   JMSException        if the subscription cannot be removed
     */
    public synchronized void unsubscribe(String name)
        throws JMSException {
        _delegate.unsubscribe(name);
    }

    /**
     * Stop message delivery to this session. If there are any problems
     * completing the request then throw the JMSException exception
     *
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public synchronized void stopMessageDelivery()
        throws JMSException, RemoteException {
        _delegate.stopMessageDelivery();
    }

    /**
     * Start message delivery to this session. If there are any problems
     * completing this request then throw the JMSException exception
     *
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public synchronized void startMessageDelivery()
        throws JMSException, RemoteException {
        _delegate.startMessageDelivery();
    }

    /**
     * Recover the session
     *
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public synchronized void recover()
        throws JMSException, RemoteException {
        _delegate.recover();
    }

    /**
     * Commit the session
     *
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public synchronized void commit()
        throws JMSException, RemoteException {
        _delegate.commit();
    }

    /**
     * Rollback the session
     *
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public synchronized void rollback()
        throws JMSException, RemoteException {
        _delegate.rollback();
    }

    // implementation of RmiJmsServerSessionIfc.commit
    public void commit(Xid xid, boolean onePhase)
        throws XAException, RemoteException {
        _delegate.commit(xid, onePhase);
    }

    // implementation of RmiJmsServerSessionIfc.end
    public void end(Xid xid, int flags)
        throws XAException, RemoteException {
        _delegate.end(xid, flags);
    }

    // implementation of RmiJmsServerSessionIfc.forget
    public void forget(Xid xid)
        throws XAException, RemoteException {
        _delegate.forget(xid);
    }

    // implementation of RmiJmsServerSessionIfc.getTransactionTimeout
    public int getTransactionTimeout()
        throws XAException, RemoteException {
        return _delegate.getTransactionTimeout();
    }

    // implementation of RmiJmsServerSessionIfc.prepare
    public int prepare(Xid xid)
        throws XAException, RemoteException {
        return _delegate.prepare(xid);
    }

    // implementation of RmiJmsServerSessionIfc.recover
    public Xid[] recover(int flag)
        throws XAException, RemoteException {
        return _delegate.recover(flag);
    }

    // implementation of RmiJmsServerSessionIfc.rollback
    public void rollback(Xid xid)
        throws XAException, RemoteException {
        _delegate.rollback(xid);
    }

    // implementation of RmiJmsServerSessionIfc.setTransactionTimeout
    public boolean setTransactionTimeout(int seconds)
        throws XAException, RemoteException {
        return _delegate.setTransactionTimeout(seconds);
    }

    // implementation of RmiJmsServerSessionIfc.start
    public void start(Xid xid, int flags)
        throws XAException, RemoteException {
        _delegate.start(xid, flags);
    }

    // implementation of RmiJmsServerSessionIfc.getResourceManageRId
    public String getResourceManagerId() throws XAException, RemoteException {
        return _delegate.getResourceManagerId();
    }

    /**
     * All server side sessions register with the consumer manager for
     * message consumption (i.e push-model). When a message arrives this
     * server-side instance will send it down to the client side stub for
     * further processing.
     * <p>
     * If the listener is null then throw JMSException
     *
     * @param       listener            listener to delivery messages too.
     * @exception   JMSException
     * @exception   RemoteException.
     */
    public synchronized void setMessageListener(RemoteJmsMessageListenerIfc listener)
        throws JMSException, RemoteException {
        _listener = listener;
        _delegate.setMessageListener(this);
    }

    /**
     * Enable or disable asynchronous message delivery for a particular
     * consumer
     *
     * @param clientId - the id of the client to check
     * @param id - the last message asynchronously delivered to consumer
     * @param enable - true to enable; false to disable
     */
    public void enableAsynchronousDelivery(long clientId, String id,
                                           boolean enable)
        throws JMSException, RemoteException {
        _delegate.enableAsynchronousDelivery(clientId, id, enable);
    }

    // implementation of MessageListener.onMessage
    // throws unchecked exception ClientDisconnectionException
    public void onMessage(Message message) {
        if (_listener != null) {
            try {
                _listener.onMessage(message);
            } catch (RemoteException exception) {
                // rethrow
                throw new ClientDisconnectionException("Failed in onMessage " +
                    exception);
            }
        } else {
            // throw a ClientDisconnectException
            throw new ClientDisconnectionException(
                "No listener registered for this session");
        }
    }

    // implementation of MessageListener.onMessage
    // throws unchecked exception ClientDisconnectionException
    public void onMessages(Vector messages) {
        if (_listener != null) {
            try {
                _listener.onMessages(messages);
            } catch (RemoteException exception) {
                exception.printStackTrace();
                // rethrow
                throw new ClientDisconnectionException("Failed in onMessages " +
                    exception);
            }
        } else {
            // throw a ClientDisconnectException
            throw new ClientDisconnectionException(
                "No listener registered for this session");
        }
    }

    // implementation of MessageListener.onMessage
    // throws unchecked exception ClientDisconnectionException
    public void onMessageAvailable(long clientId) {
        if (_listener != null) {
            try {
                _listener.onMessageAvailable(clientId);
            } catch (RemoteException exception) {
                // rethrow
                throw new ClientDisconnectionException("Failed in onMessageAvailable " +
                    exception);
            }
        } else {
            // throw a ClientDisconnectException
            throw new ClientDisconnectionException(
                "No listener registered for this session");
        }
    }
}
