/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000-2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ThreadPoolManager.java,v 1.4 2003/08/07 13:33:10 tanderson Exp $
 *
 * Date         Author  Changes
 * 11/23/00	    jima    Created
 */
package org.exolab.jms.threads;

// core java

import java.util.HashMap;

import org.exolab.core.service.BasicService;
import org.exolab.core.service.ServiceException;
import org.exolab.core.service.ServiceState;
import org.exolab.core.threadPool.ThreadPool;


/**
 * The thread pool manager manages all the required {@link ThreadPool}
 * objects.
 * The clients can obtain a ThreadPool from the {@link ThreadPoolManager} with
 * a given number of Threads.
 * <p>
 * The manager keeps a list of all The ThreadPools it has dished out with a
 * unique client provided name. Client can then re-request the already created
 * ThreadPool or share common ThreadPools if required.
 * <p>
 * The ThreadPool manager will attempt to shutdown all ThreadPools and
 * stop all Threads when it receives a stop request.
 * <p>
 * If a client attempts to create a ThreadPool with a name that already
 * exists a ThreadPoolExistsException will be raised and the creation
 * will fail.
 *
 * @version     $Revision: 1.4 $ $Date: 2003/08/07 13:33:10 $
 * @author      <a href="mailto:mourikis@intalio.com">Jim Mourikis</a>
 **/
public class ThreadPoolManager extends BasicService {

    // The list of all allocated ThreadPools.
    private static HashMap _pools = null;

    // The one and only instance of this class.
    private static ThreadPoolManager _instance = null;

    // The thread we run in.
    private static final String THREADPOOLMGR_NAME = "ThreadPoolManager";


    /**
     * The constructor initialises the parent, and creates an empty container
     * to hold the thread pools, as they are requested.
     *
     * <P> Entry into this constructor is only through the intialise method
     * below. Note only one instance of this mgr is created.
     */
    private ThreadPoolManager() {
        super(THREADPOOLMGR_NAME);
        _pools = new HashMap();
    }

    /**
     * Return the one and only instance of the ThreadPoolManager.
     *
     * <P>Note: Initiase must be called above to initialise the pool, else
     * a null object will be returned.
     */
    static public ThreadPoolManager instance() {
        if (_instance == null) {
            _instance = new ThreadPoolManager();
        }
        return _instance;
    }

    /**
     * A client has requested a new thread pool creation with the given name.
     * if the named Pool has not already been created, then create it
     * add it to the list, and return it to the client.
     *
     * @param name The unique name given to this ThreadPool
     * @param size The maximum nuber of Threads this pool contains.
     * @throws ThreadPoolExistsException if the pool with the given name
     * already exists.
     */
    public ThreadPool createThreadPool(String name, int size)
        throws ThreadPoolExistsException {
        synchronized (_pools) {
            if (_pools.containsKey(name)) {
                throw new ThreadPoolExistsException("ThreadPool with name " +
                    name + " exists");
            }
            ThreadPool pool = new ThreadPool(name, size, true);

            _pools.put(name, pool);
            return pool;
        }
    }

    /**
     * Get the ThreadPool with the given name. if the named pool does not exist
     * throw an exception.
     *
     * @param name The unique name of the requested ThreadPool
     * @throws UnknownThreadPoolException if the pool with the given name
     * does not exist.
     */
    public ThreadPool getThreadPool(String name)
        throws UnknownThreadPoolException {
        ThreadPool pool = null;

        synchronized (_pools) {
            pool = (ThreadPool) _pools.get(name);
        }

        if (pool == null) {
            throw new UnknownThreadPoolException("ThreadPool with name " +
                name + " does not exist");
        }

        return pool;
    }

    /**
     * Attempts to shutdown all the threads in the pool and removes the given
     * ThreadPool.
     *
     * @param name The unique name of the requested ThreadPool
     * @throws  UnknownThreadPoolException if the pool with the given name
     * does not exist.
     */
    public void deleteThreadPool(String name)
        throws UnknownThreadPoolException {
        ThreadPool pool = null;

        synchronized (_pools) {
            pool = (ThreadPool) _pools.remove(name);
        }

        if (pool == null) {
            throw new UnknownThreadPoolException("ThreadPool with name " +
                name + " does not exist");
        }
        pool.stopRequestAllWorkers();
    }

    /**
     * This starts the ThreadPoolManager's thread
     */
    public synchronized void run() {
        while (getState() != ServiceState.STOPPED) {
            try {
                // wait for the sevice to terminate.
                wait();
            } catch (InterruptedException ignore) {
            }
        }
    }

    /**
     * Overide BasicService.stop. Clean up all thread pools,
     * by calling their stop methods.
     *
     * @throws ServiceException
     */
    public void stop() throws ServiceException {
        synchronized (_pools) {
            Object[] ob = _pools.values().toArray();

            for (int i = 0, j = _pools.size(); i < j; i++) {
                ((ThreadPool) ob[i]).stopRequestAllWorkers();
            }
            _pools.clear();
        }
        super.stop();
        // zero out the static reference
        _pools = null;
        _instance = null;
    }

} //-- ThreadPoolManager
