/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: TransactionManagerType.java,v 1.2 2003/08/07 13:33:12 tanderson Exp $
 *
 * Date         Author  Changes
 * 07/02/2001   jima    Created
 */
package org.exolab.jms.transaction;

import java.io.Serializable;


/**
 * This class defines all the valid transaction manager types
 *
 * @version     $Revision: 1.2 $ $Date: 2003/08/07 13:33:12 $
 * @author      <a href="mailto:jima@intalio.com">Jim Alateras</a>
 */
public final class TransactionManagerType
    implements Serializable {

    /**
     * Limit the scope of the constructor. All instances of this class are
     * created during class initialisation time.
     *
     * @param	    type            message type name
     * @param       int             associatedordinal value
     */
    private TransactionManagerType(String type, int ord) {
        _type = type;
        _ord = ord;
    }

    /**
     * Returns the ordinal value for this type
     *
     * @return      int
     */
    public int getOrd() {
        return _ord;
    }

    /**
     * Return the transaction type as a string
     *
     * @return      String
     */
    public String getType() {
        return _type;
    }

    /**
     * Returns the name of this transaction type
     *
     * @return       String
     */
    public String toString() {
        return _type;
    }

    /**
     * Return the TransactionManagerType give its ordinal value
     *
     * @param value - message ord value
     * @return TransactionManagerType
     */
    public static TransactionManagerType fromInt(int value) {
        switch (value) {
            case NULL_TM_ORD:
                return NULL;

            case TYREX_TM_ORD:
                return TYREX;

            default:
                return null;
        }
    }

    /**
     * Return true if the two specified object is ofthe same type
     * andthat their ordinal values are the same.
     *
     * @param       obj         object to compare against
     * @return      boolean     true if the two objects are equivalent
     */
    public boolean equals(Object obj) {
        boolean result = false;

        if (obj instanceof TransactionManagerType) {
            TransactionManagerType req_obj = (TransactionManagerType) obj;
            if (req_obj.getOrd() == getOrd()) {
                result = true;
            }
        }

        return result;
    }

    /**
     * Returns true if it is a Tyrex transaction manager
     *
     * @return      boolean
     */
    public boolean isTyrexTransactionManager() {
        return getOrd() == TYREX.getOrd();
    }

    /**
     * Returns true if it is a Tyrex transaction manager
     *
     * @return      boolean
     */
    public boolean isNullTransactionManager() {
        return getOrd() == NULL.getOrd();
    }


    /**
     * Oridnal values for each transaction manager type
     */
    public static final int NULL_TM_ORD = 1;
    public static final int TYREX_TM_ORD = 2;

    /**
     * Strign value for each transaction manager type
     */
    public static final String NULL_TM_STR = "null-tm";
    public static final String TYREX_TM_STR = "tyrex-tm";


    /**
     * Instatiate instances related to each of the enumerations
     */
    public static final TransactionManagerType NULL =
        new TransactionManagerType(NULL_TM_STR, NULL_TM_ORD);
    public static final TransactionManagerType TYREX =
        new TransactionManagerType(TYREX_TM_STR, TYREX_TM_ORD);


    /**
     * Holds the type of the message type. Cannot be null
     */
    private String _type;

    /**
     * Holds the ordinal value of this message type. This is unique within the
     * transaction manager type domain
     */
    private int _ord;
}


