/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: TestBug705640.java,v 1.2 2003/08/16 09:44:52 tanderson Exp $
 *
 * Date         Author      Changes
 * 21/02/2002   jima        Created
 */

package org.exolab.testharness.jms.bugs;

// junit library
import junit.framework.TestCase;

import java.util.Hashtable;
import javax.naming.Context;
import javax.naming.InitialContext;

// jms library
import javax.jms.DeliveryMode;
import javax.jms.Session;
import javax.jms.QueueConnectionFactory;
import javax.jms.QueueConnection;
import javax.jms.QueueSession;
import javax.jms.QueueReceiver;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicConnection;
import javax.jms.TopicSession;
import javax.jms.TopicPublisher;
import javax.jms.TopicSubscriber;
import javax.jms.MessageListener;
import javax.jms.Message;
import javax.jms.Queue;
import javax.jms.Topic;
import javax.jms.Session;
import javax.jms.TemporaryQueue;
import javax.jms.TemporaryTopic;
import javax.jms.QueueRequestor;
import javax.jms.QueueSender;
import javax.jms.QueueReceiver;
import javax.jms.MapMessage;
import javax.jms.QueueConnectionFactory;
import javax.jms.JMSException;
import javax.jms.IllegalStateException;

// openjms
import org.exolab.jms.administration.JmsAdminServerIfc;

// openjms testhaness
import org.exolab.testharness.jms.util.ClientHelper;
import org.exolab.testharness.jms.util.CountingListener;


/**
 * This tests for http://sourceforge.net/tracker/index.php?func=detail&aid=705640&group_id=54559&atid=474136
 *
 * @version     $Revision: 1.2 $
 * @author      <a href="mailto:jima@intalio.com">Jim Alateras</a>
 */
public class TestBug705640 extends TestCase {

    /**
     * Constructor
     */
    public TestBug705640(String name) {
        super(name);
    }

    /** 
     * Test the bug for queues
     */
    public void testQueueSimpleOrder() 
        throws Exception {
        final String name = "testQueue705640";
        final QueueConnection connection = ClientHelper.getQueueConnection();
        try {
            connection.start();
            
            QueueSession ssess = connection.createQueueSession(
                false, Session.AUTO_ACKNOWLEDGE);

            // create a queue sender and send a message
            QueueSender sender = ssess.createSender(null);
            sender.setDeliveryMode(DeliveryMode.NON_PERSISTENT);
            Queue queue = ssess.createQueue(name);

            // send 10 messages
            for (int index = 0; index < 10; index++) {
                MapMessage msg = ssess.createMapMessage();
                msg.setInt("index", index);
                sender.send(queue, msg);
            }

            QueueSession rsess = connection.createQueueSession(
                false, Session.AUTO_ACKNOWLEDGE);

            // create a queue sender and send a message
            QueueReceiver receiver = rsess.createReceiver(queue);

            // receive all ten messages
            int last_index = -1;
            for (int index = 0; index < 10; index++) {
                MapMessage msg = (MapMessage)receiver.receive();
                int value = msg.getInt("index");
                assertTrue(value == (last_index + 1));
                last_index = value;
            }
        } catch (Exception exception) {
            exception.printStackTrace();
            assertTrue("TestBug705640.testQueueSimpleOrder : " + 
                       exception.getMessage(), false);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (JMSException ignore) {
                }
            }
        }
    }

    /** 
     * Test the bug for queues setting different priorities
     */
    public void testQueuePrtyOrder() 
        throws Exception {
        final String name = "testQueue705640";
        final QueueConnection connection = ClientHelper.getQueueConnection();
        try {
            connection.start();
            
            QueueSession ssess = connection.createQueueSession(
                false, Session.AUTO_ACKNOWLEDGE);

            // create a queue sender and send a message
            QueueSender sender = ssess.createSender(null);
            sender.setDeliveryMode(DeliveryMode.NON_PERSISTENT);
            Queue queue = ssess.createQueue(name);

            // send 10 messages in increasing priority
            for (int index = 0; index < 10; index++) {
                MapMessage msg = ssess.createMapMessage();
                msg.setJMSPriority(index);
                sender.send(queue, msg);
            }

            QueueSession rsess = connection.createQueueSession(
                false, Session.AUTO_ACKNOWLEDGE);

            // create a queue sender and send a message
            QueueReceiver receiver = rsess.createReceiver(queue);

            // receive all ten messages
            boolean first = true;
            int last_priority = 0;
            for (int index = 0; index < 10; index++) {
                MapMessage msg = (MapMessage)receiver.receive();
                int priority = msg.getJMSPriority();
                if (first) {
                    last_priority = priority;
                    first = false;
                } else {
                    assertTrue(priority <= last_priority);
                }
            }
        } catch (Exception exception) {
            exception.printStackTrace();
            assertTrue("TestBug705640.testQueuePrtyOrder : " + 
                       exception.getMessage(), false);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (JMSException ignore) {
                }
            }
        }
    }

    /** 
     * Test the bug for queues using selectors
     */
    public void testQueueSelector() 
        throws Exception {
        final String name = "testQueue705640";
        final QueueConnection connection = ClientHelper.getQueueConnection();
        try {
            connection.start();
            
            QueueSession ssess = connection.createQueueSession(
                false, Session.AUTO_ACKNOWLEDGE);

            // create a queue sender and send a message
            QueueSender sender = ssess.createSender(null);
            sender.setDeliveryMode(DeliveryMode.NON_PERSISTENT);
            Queue queue = ssess.createQueue(name);

            // send 10 messages in increasing priority
            for (int index = 0; index < 10; index++) {
                MapMessage msg = ssess.createMapMessage();
                msg.setJMSPriority(index);
                sender.send(queue, msg);
            }

            QueueSession rsess = connection.createQueueSession(
                false, Session.AUTO_ACKNOWLEDGE);

            // create a queue sender and send a message
            QueueReceiver receiver = rsess.createReceiver(queue, "1=1");

            // receive all ten messages
            boolean first = true;
            int last_priority = 0;
            for (int index = 0; index < 10; index++) {
                MapMessage msg = (MapMessage)receiver.receive();
                int priority = msg.getJMSPriority();
                if (first) {
                    last_priority = priority;
                    first = false;
                } else {
                    assertTrue(priority <= last_priority);
                }
            }
        } catch (Exception exception) {
            exception.printStackTrace();
            assertTrue("TestBug705640.testQueueSelector : " + 
                       exception.getMessage(), false);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (JMSException ignore) {
                }
            }
        }
    }

    /** 
     * Test the bug for topics
     */
    public void testTopicSimpleOrder() 
    throws Exception {
        final Object waitForReceiver = new Object();
        final String name = "testTopic705640";
        final TopicConnection connection = ClientHelper.getTopicConnection();
        try {
            connection.start();

            // set up and start the consumer before the publisher
            Thread consumer = new Thread(new Runnable() {
                    // implementation of Runnable.run
                    public void run() {          
                        try {
                            TopicSession session = connection.createTopicSession(
                                false, Session.AUTO_ACKNOWLEDGE);
                            Topic topic = session.createTopic(name);
                            TopicSubscriber subscriber = session.createSubscriber(topic);

                            // receive 10 messages and ensure they are in
                            // the correct order
                            int last_index = -1;
                            for (int index = 0; index < 10; index++) {
                                MapMessage msg = (MapMessage)subscriber.receive();
                                int value = msg.getInt("index");
                                assertTrue(value == (last_index + 1));
                                last_index = value;
                            }

                            Thread.sleep(3000);
                            synchronized(waitForReceiver) {
                                waitForReceiver.notifyAll();
                            }
                        } catch (Exception exception) {
                            exception.printStackTrace();
                            assertTrue("TestBug705640.testTopic : " + 
                                       exception.getMessage(), false);
                        }
                    }
                    });

            // start the receiver and then wait before starting the 
            // publisher
            consumer.start();  
            Thread.sleep(2000);

            // set up the publisher to publish the 10 messages
            TopicSession session = connection.createTopicSession(
                false, Session.AUTO_ACKNOWLEDGE);
            Topic topic = session.createTopic(name);
            TopicPublisher publisher = session.createPublisher(topic);
            publisher.setDeliveryMode(DeliveryMode.NON_PERSISTENT);

            // send 10 messages and wait for 3 seconds
            for (int index = 0; index < 10; index++) {
                MapMessage msg = session.createMapMessage();
                msg.setInt("index", index);
                publisher.publish(msg);
            }

            try {
                synchronized(waitForReceiver) {
                    waitForReceiver.wait();
                }
            } catch (InterruptedException exception) {
                exception.printStackTrace();
                assertTrue("TestBug705640.testTopicSimpleOrder : Failed to receive the " +
                           "10 messages in the specified time frame",
                           false);
            }
        } catch (Exception exception) {
            exception.printStackTrace();
            assertTrue("TestBug705640.testTopicSimpleOrder : " + 
                       exception.getMessage(), false);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (JMSException ignore) {
                }
            }
        }
    }


    /** 
     * Test the bug for topic using priority
     */
    public void testTopicPrtyOrder() 
    throws Exception {
        final Object waitForReceiver = new Object();
        final String name = "testTopic705640";
        final TopicConnection connection = ClientHelper.getTopicConnection();
        try {
            connection.start();

            // set up and start the consumer before the publisher
            Thread consumer = new Thread(new Runnable() {
                    // implementation of Runnable.run
                    public void run() {          
                        try {
                            TopicSession session = connection.createTopicSession(
                                false, Session.AUTO_ACKNOWLEDGE);
                            Topic topic = session.createTopic(name);
                            TopicSubscriber subscriber = session.createSubscriber(topic);

                            // receive 10 messages and ensure they are in
                            // the correct order
                            boolean first = true;
                            int last_priority = 0;
                            for (int index = 0; index < 10; index++) {
                                MapMessage msg = (MapMessage)subscriber.receive();
                                int priority = msg.getJMSPriority();
                                if (first) {
                                    last_priority = priority;
                                    first = false;
                                } else {
                                    assertTrue(priority <= last_priority);
                                }
                            }

                            Thread.sleep(3000);
                            synchronized(waitForReceiver) {
                                waitForReceiver.notifyAll();
                            }
                        } catch (Exception exception) {
                            exception.printStackTrace();
                            assertTrue("TestBug705640.testTopic : " + 
                                       exception.getMessage(), false);
                        }
                    }
                    });

            // start the receiver and then wait before starting the 
            // publisher
            consumer.start();  
            Thread.sleep(2000);

            // set up the publisher to publish the 10 messages
            TopicSession session = connection.createTopicSession(
                false, Session.AUTO_ACKNOWLEDGE);
            Topic topic = session.createTopic(name);
            TopicPublisher publisher = session.createPublisher(topic);
            publisher.setDeliveryMode(DeliveryMode.NON_PERSISTENT);

            // send 10 messages and wait for 3 seconds
            for (int index = 0; index < 10; index++) {
                MapMessage msg = session.createMapMessage();
                msg.setJMSPriority(index);
                publisher.publish(msg);
            }

            try {
                synchronized(waitForReceiver) {
                    waitForReceiver.wait();
                }
            } catch (InterruptedException exception) {
                exception.printStackTrace();
                assertTrue("TestBug705640.testTopicPrtyOrder : Failed to receive the " +
                           "10 messages in the specified time frame",  false);
            }
        } catch (Exception exception) {
            exception.printStackTrace();
            assertTrue("TestBug705640.testTopicPrtyOrder : " + 
                       exception.getMessage(), false);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (JMSException ignore) {
                }
            }
        }
    }

    /** 
     * Test the bug for topic using selector
     **/
    public void testTopicSelector() 
    throws Exception {
        final Object waitForReceiver = new Object();
        final String name = "testTopic705640";
        final TopicConnection connection = ClientHelper.getTopicConnection();
        try {
            connection.start();

            // set up and start the consumer before the publisher
            Thread consumer = new Thread(new Runnable() {
                    // implementation of Runnable.run
                    public void run() {          
                        try {
                            TopicSession session = connection.createTopicSession(
                                false, Session.AUTO_ACKNOWLEDGE);
                            Topic topic = session.createTopic(name);
                            TopicSubscriber subscriber = session.createSubscriber(
                                topic, "1=1", false);

                            // receive 10 messages and ensure they are in
                            // the correct order
                            boolean first = true;
                            int last_priority = 0;
                            for (int index = 0; index < 10; index++) {
                                MapMessage msg = (MapMessage)subscriber.receive();
                                int priority = msg.getJMSPriority();
                                if (first) {
                                    last_priority = priority;
                                    first = false;
                                } else {
                                    assertTrue(priority <= last_priority);
                                }
                            }

                            Thread.sleep(3000);
                            synchronized(waitForReceiver) {
                                waitForReceiver.notifyAll();
                            }
                        } catch (Exception exception) {
                            exception.printStackTrace();
                            assertTrue("TestBug705640.testTopic : " + 
                                       exception.getMessage(), false);
                        }
                    }
                    });

            // start the receiver and then wait before starting the 
            // publisher
            consumer.start();  
            Thread.sleep(2000);

            // set up the publisher to publish the 10 messages
            TopicSession session = connection.createTopicSession(
                false, Session.AUTO_ACKNOWLEDGE);
            Topic topic = session.createTopic(name);
            TopicPublisher publisher = session.createPublisher(topic);
            publisher.setDeliveryMode(DeliveryMode.NON_PERSISTENT);

            // send 10 messages and wait for 3 seconds
            for (int index = 0; index < 10; index++) {
                MapMessage msg = session.createMapMessage();
                msg.setJMSPriority(index);
                publisher.publish(msg);
            }

            try {
                synchronized(waitForReceiver) {
                    waitForReceiver.wait();
                }
            } catch (InterruptedException exception) {
                exception.printStackTrace();
                assertTrue("TestBug705640.testTopicSelector : Failed to receive the " +
                           "10 messages in the specified time frame",  false);
            }
        } catch (Exception exception) {
            exception.printStackTrace();
            assertTrue("TestBug705640.testTopicSelector : " + 
                       exception.getMessage(), false);
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (JMSException ignore) {
                }
            }
        }
    }
}

