/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ClientHelper.java,v 1.3 2003/03/24 11:49:02 tanderson Exp $
 *
 * Date         Author  Changes
 * 09/06/2001   jima    created
 */
package org.exolab.testharness.jms.util;

import java.net.MalformedURLException;
import java.util.Hashtable;

import javax.jms.ConnectionFactory;
import javax.jms.QueueConnectionFactory;
import javax.jms.TopicConnectionFactory;
import javax.jms.Connection;
import javax.jms.QueueConnection;
import javax.jms.TopicConnection;
import javax.jms.JMSException;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.exolab.jms.administration.AdminConnectionFactory;
import org.exolab.jms.administration.JmsAdminServerIfc;
import org.exolab.jms.config.ConfigHelper;
import org.exolab.jms.config.Configuration;
import org.exolab.jms.config.ConfigurationManager;
import org.exolab.jms.config.ConnectionFactories;
import org.exolab.jms.config.Connector;
import org.exolab.jms.config.Property;
import org.exolab.jms.config.types.SchemeType;
import org.exolab.jms.client.JmsQueueConnectionFactory;
import org.exolab.jms.client.JmsTopicConnectionFactory;


/**
 * Utility class for getting connections, sessions etc to the 
 * openjms server.
 *
 * @version     $Revision: 1.3 $ $Date: 2003/03/24 11:49:02 $
 * @author      <a href="mailto:jima@intalio.com">Jim Alateras</a>
 */
public class ClientHelper {

    /**
     * Returns an initial context, using the configured properties
     *
     * @return an initial context
     * @throws IllegalStateException if {@link ConfigurationManager} hasn't
     * been initialised
     */
    public static Context getInitialContext() throws NamingException {    
        Hashtable env = new Hashtable();

        Configuration config = ConfigurationManager.getConfig();
        Property[] properties = config.getJndiConfiguration().getProperty();

        for (int i = 0; i < properties.length; ++i) {
            Property property = properties[i];
            env.put(property.getName(), property.getValue());
        }

        return new InitialContext(env);
    }

    /**
     * Returns an initial context
    /**
     * Returns a {@link QueueConnection} to the server
     *
     * @return QueueConnection  
     * @throws JMSException - if we cannot get one
     */
    public static QueueConnection getQueueConnection() throws JMSException {
        QueueConnection connection = null;
        
        try {
            Context context = getInitialContext();

            // if we can't get the root context then exit with an exception
            if (context == null) {
                throw new JMSException("Failed to get the root context");
            }
    
            // get the connection factory name
            Connector connector = ConfigurationManager.getConnector();
            ConnectionFactories factories = connector.getConnectionFactories();
            String factory_name = 
                factories.getQueueConnectionFactory(0).getName();

            // lookup the connection factory from the context
            QueueConnectionFactory factory = 
                (QueueConnectionFactory)context.lookup(factory_name);
            if (factory != null) {
                connection = factory.createQueueConnection();
            }
        } catch (JMSException exception) {
            throw exception;
        } catch (Exception exception) {
            throw new JMSException("Failed to get InitialContext");
        }
    
        if (connection == null) {
            throw new JMSException("Failed to get QueueConnection");
        }

        return connection;
    }

    /**
     * Returns a {@link TopicConnection} to the server
     *
     * @return TopicConnection  
     * @throws JMSException - if we cannot get one
     */
    public static TopicConnection getTopicConnection() throws JMSException {
        TopicConnection connection = null;

        try {
            Context context = getInitialContext();

            // if we can't get the root context then exit with an exception
            if (context == null) {
                throw new JMSException("Failed to get the root context");
            }
    
            // get the connection factory name
            Connector connector = ConfigurationManager.getConnector();
            ConnectionFactories factories = connector.getConnectionFactories();
            String factory_name = 
                factories.getTopicConnectionFactory(0).getName();
    
            // lookup the connection factory from the context
            TopicConnectionFactory factory = 
                (TopicConnectionFactory)context.lookup(factory_name);
            if (factory != null) {
                connection = factory.createTopicConnection();
            }
        } catch (JMSException exception) {
            throw exception;
        } catch (Exception exception) {
            throw new JMSException("Failed to get InitialContext");
        }

        if (connection == null) {
            throw new JMSException("Failed to get TopicConnection");
        }

        return connection;
    }

    /**
     * Return the OpenJMS administration interface specified by 
     * {@link ConfigurationManager}
     *
     * @return the OpenJMS administration interface
     * @throws JMSException if the admin class cannot be created
     */ 
    public static JmsAdminServerIfc getAdmin() throws JMSException {
        JmsAdminServerIfc result = null;
        Configuration config = ConfigurationManager.getConfig();
        SchemeType scheme = ConfigurationManager.getConnector().getScheme();
        String url = ConfigHelper.getAdminURL(scheme, config);
        try {
            result = AdminConnectionFactory.create(url);
        } catch (MalformedURLException exception) {
            throw new JMSException(exception.getMessage());
        }
        return result;
    }

    public static void setupDestination(String name, Boolean type) 
        throws JMSException {
        JmsAdminServerIfc admin = getAdmin();
        admin.removeDestination(name);
        admin.addDestination(name, type);
        admin.close();
    }

    public static void removeDestination(String name) throws JMSException {
        JmsAdminServerIfc admin = getAdmin();
        admin.removeDestination(name);
        admin.close();
    }
}
